/*->c.curver */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>


#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "bbc.h"
#include "flex.h"
#include "transform.h"
#include "xprocess.h"
#include "BezierArc.h"

#include "err.h"
#include "fsx.h"
#include "bf.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "colour.h"
#include "etc.h"
#include "xmath.h"
#include "deb.h"
#include "DrawMod.h"


#include "constants.h"
#include "str.h"

#include "object.h"

#include "viewr.h"
#include "view.h"
#include "file.h"
#include "chain.h"
#include "xcolour.h"
#include "xredraw.h"
#include "xrender.h"
#include "geometry.h"


#include "posn.h"



#include "curvelo.h"
#include "curve.h"

/*****************************************************************************/


#ifndef PROD

void debugcurve(int * p)
{
 int c;

 while(1)
 {
  c=*p++;

  switch(c)
  {
   case DRAWMOVE:
                 dprintf(0,"draw move (%d,%d)\n",p[0],p[1]);
                 p+=2;
                 break;
 
  case DRAWDRAW:
                 dprintf(0,"draw draw (%d,%d)\n",p[0],p[1]);
                 p+=2;
                 break;

  case DRAWCURVE:
                 dprintf(0,"draw curve (%d,%d) (%d,%d) (%d,%d)\n",
                        p[0],p[1],p[2],p[3],p[4],p[5]);
                 p+=6;
                 break;

  case DRAWCLOSE:
                 dprintf(0,"draw close\n");
                 break;

   case DRAWEND:
                 dprintf(0,"draw end\n");
                 return;
                 break;
  }
 }
}


void debugpath(int path,filestr * file)
{
 pathstr * pathp;
 pathp=(pathstr*)fob_find(path,file);
 debugcurve(pathp->data);
}


#endif




#define BOXPATHSIZE     17
#define ELLIPSEPATHSIZE 33
#define LINEPATHSIZE    16
#define RBOXPATHSIZE    45
#define MAXPATHSIZE     64




/* constructs the path for an ellipse */

void ellipsepath(int * p,curvestr * c,transformstr * t)
{
 bezier_arc_coord coord[16];
 bezier_arc_coord centre;
 transformstr     tr;
 transformstr     r;
 int              n;
 int              i;
 int              j;

 centre.x=centre.y=0;

 if(c->box.h>c->box.w)
 {
  bezier_arc_circle(centre,c->box.h/2,coord);
  tr_setshift(&tr,c->box.x+c->box.w/2,c->box.y+c->box.h/2);
  tr_addscale(&tr,c->box.h,c->box.w,1,1);
 }
 else
 {
  bezier_arc_circle(centre,c->box.w/2,coord);
  tr_setshift(&tr,c->box.x+c->box.w/2,c->box.y+c->box.h/2);
  tr_addscale(&tr,1,1,c->box.w,c->box.h);
 }

 if(t)
 {
  tr_multiply(&r,&tr,t);
  tr_transformset((coordstr*)coord,13,&r);
 }
 else
 {
  tr_transformset((coordstr*)coord,13,&tr);
 }

 *p++=DRAWMOVE;
 *p++=coord[0].x;
 *p++=coord[0].y;
 n=1;
 for(i=0;i<4;i++)
 {
  *p++=DRAWCURVE;
  for(j=0;j<3;j++)
  {
   *p++=coord[n].x;
   *p++=coord[n++].y;
  }
 }

 *p++=DRAWCLOSE;
 *p=DRAWEND;
}




static os_error * ellipsetopath(curvestr * c,filestr * file)
{
 os_error * err;
 int        size;
 pathstr  * path;

 size=ELLIPSEPATHSIZE*sizeof(int);

 err=fob_create(sizeof(pathstr)+size-sizeof(int),file,OPATH,(fobjptr)&path);
 if(!err)
 {
  ellipsepath(path->data,c,NULL);
  c->type=CPATH;
  c->data.path.path=path->key;
 }

 return(err);
}





/* constuct an elliptical arc from (sx,sy) to (ex,ey) in given direction */
/* assumes that you are sat at given end point                           */
/* returns number of ints written                                        */


#define ELLIPSEMIN 32


int ellipsearc(int * q,curvestr * c,int sox,int soy,
                                    int eox,int eoy,int clockwise)
{
 bezier_arc_coord coord[16];
 bezier_arc_coord centre;
 transformstr     tr;
 int              i;
 int              j;
 int              n;
 double           start;
 double           end;
 int              segment;
 int              segn;
 int              sx;
 int              sy;
 int              ex;
 int              ey;
 int            * p;


 centre.x=centre.y=0;
 segn=0;
 i=1;

 p=q;

 sx=sox-(c->box.x+c->box.w/2);
 sy=soy-(c->box.y+c->box.h/2);
 sy=-sy;

 ex=eox-(c->box.x+c->box.w/2);
 ey=eoy-(c->box.y+c->box.h/2);
 ey=-ey;

/*
 dprintf(1,"sx=%d sy=%d ex=%d ey=%d h=%d w=%d",sx,sy,ex,ey,c->box.w,c->box.h);
 */

 if(c->box.h>c->box.w)
 {
  sx=scale(sx,c->box.h,c->box.w);
  ex=scale(ex,c->box.h,c->box.w);

  start=atan2((double)sy,(double)sx);
  end=atan2((double)ey,(double)ex);

  bezier_arc_start(centre,start,end,c->box.h/2,&coord[0],&coord[3],
                                               &coord[1],&coord[2]);

  segment=1;
  while(1)
  {
   if(ABS(coord[i+2].y-coord[i-1].y)>ELLIPSEMIN)
   {
    i+=3;
    segn++;
   }

   segment=bezier_arc_segment(segment,&coord[i+2],&coord[i],&coord[i+1]);
   if(!segment) break;
  }

 /*
  dprintf(0,"ellipse arc h>w segn=%d start=%e end=%e",segn,start,end);
  */

  tr_setshift(&tr,c->box.x+c->box.w/2,c->box.y+c->box.h/2);
  tr_addscale(&tr,c->box.h,c->box.w,1,-1);
 }
 else
 {
  sy=scale(sy,c->box.w,c->box.h);
  ey=scale(ey,c->box.w,c->box.h);

  start=atan2((double)sy,(double)sx);
  end=atan2((double)ey,(double)ex);


  bezier_arc_start(centre,start,end,c->box.w/2,&coord[0],&coord[3],
                                               &coord[1],&coord[2]);

  segment=1;
  while(1)
  {
   if(ABS(coord[i+2].y-coord[i-1].y)>ELLIPSEMIN)
   {
    i+=3;
    segn++;
   }

   segment=bezier_arc_segment(segment,&coord[i+2],&coord[i],&coord[i+1]);
   if(!segment) break;
  }


/* dprintf(0,"ellipse arc h<w segn=%d start=%e end=%e",segn,start,end); */

  tr_setshift(&tr,c->box.x+c->box.w/2,c->box.y+c->box.h/2);
  tr_addscale(&tr,0,0,c->box.w,-c->box.h);
 }

 tr_transformset((coordstr*)coord,i,&tr);

 if(segn)
 {
  if(clockwise)
  {
   for(j=i-2;j>=0;j-=3)
   {
    *p++=DRAWCURVE;
    for(n=0;n<3;n++)
    {
     *p++=coord[j-n].x;
     *p++=coord[j-n].y;
    }
   }

   if(soy>eoy)
   {
    if(*(p-3)>soy) *(p-3)=soy;
    if(*(q+2)<eoy) *(q+2)=eoy;
   }
   else
   {
    if(*(p-3)<soy) *(p-3)=soy;
    if(*(q+2)>eoy) *(q+2)=eoy;
   }

   *(p-2)=sox;
   *(p-1)=soy;
  }
  else
  {
   for(j=1;j<i;j+=3)
   {
    *p++=DRAWCURVE;
    for(n=0;n<3;n++)
    {
     *p++=coord[j+n].x;
     *p++=coord[j+n].y;
    }
   }

   if(soy<eoy)
   {
    if(*(p-3)>eoy) *(p-3)=eoy;
    if(*(q+2)<soy) *(q+2)=soy;
   }
   else
   {
    if(*(p-3)<eoy) *(p-3)=eoy;
    if(*(q+2)>soy) *(q+2)=soy;
   }

   *(p-2)=eox;
   *(p-1)=eoy;
  }
 }

/*
 dprintf(0,"eox=%d eoy=%d p1=%d p2=%d p3=%d",eox,eoy,*(p-1),*(p-2),*(p-3));
 dprintf(0,"sox=%d soy=%d clock=%d",sox,soy,clockwise);
 dprintf(0,"q0=%d q1=%d q2=%d q3=%d q4=%d",*q,*(q+1),*(q+2),*(q+3),*(q+4));
 q+=5;
 dprintf(0,"q0=%d q1=%d q2=%d q3=%d q4=%d",*q,*(q+1),*(q+2),*(q+3),*(q+4));
 q+=5;
 dprintf(0,"q0=%d q1=%d q2=%d q3=%d q4=%d",*q,*(q+1),*(q+2),*(q+3),*(q+4));
*/

 return(segn*7);  /* no of ints written */
}









/* constructs the path for a box */

static void boxpath(int * p,curvestr * c)
{
 *p++=DRAWMOVE;
 *p++=c->box.x;
 *p++=c->box.y;
 *p++=DRAWDRAW;
 *p++=c->box.x+c->box.w;
 *p++=c->box.y;
 *p++=DRAWDRAW;
 *p++=c->box.x+c->box.w;
 *p++=c->box.y+c->box.h;
 *p++=DRAWDRAW;
 *p++=c->box.x;
 *p++=c->box.y+c->box.h;
 *p++=DRAWDRAW;
 *p++=c->box.x;
 *p++=c->box.y;
 *p++=DRAWCLOSE;
 *p++=DRAWEND;
}




void linepath(int * p,curvestr * c,int type)
{
 int * q;

 q=p;
 *p++=DRAWMOVE;
 *p++=c->box.x;
 *p++=c->box.y;
 *p++=DRAWDRAW;
 *p++=c->box.x+c->box.w;
 *p++=c->box.y+c->box.h;
 *p++=DRAWEND;

 if(type==CLINEX || type==CLINEXR) SWOP(q[1],q[4]);
 if(type==CLINER || type==CLINEXR) SWOP(q[2],q[5]);
}




static os_error * linetopath(curvestr * c,int type,filestr * file)
{
 os_error * err;
 int        size;
 pathstr  * path;

 size=7*sizeof(int);

 err=fob_create(sizeof(pathstr)+size-sizeof(int),file,OPATH,(fobjptr)&path);
 if(!err)
 {
  linepath(path->data,c,type);
  c->type=CPATH;
  c->data.path.path=path->key;
 }

 return(err);
}





int pathisbox(int * p,boxstr * box)
{
 int x0;
 int y0;
 int x1;
 int y1;
 int x2;
 int y2;
 int x3;
 int y3;
 int xlo;
 int xhi;
 int ylo;
 int yhi;


 if(*p++==DRAWMOVE)
 {
  x0=*p++;
  y0=*p++;
  if(*p++==DRAWDRAW)
  {
   x1=*p++;
   y1=*p++;
   if(*p++==DRAWDRAW)
   {
    x2=*p++;
    y2=*p++;
    if(*p++==DRAWDRAW)
    {
     x3=*p++;
     y3=*p++;
     if(*p++==DRAWDRAW)
     {
      if(*p++==x0 && *p++==y0)
      {
       if(*p++==DRAWCLOSE)
       {
        if((y0==y1 && (y2==y3 && y2!=y0))||
           (y0==y2 && (y1==y3 && y1!=y0))||
           (y0==y3 && (y1==y2 && y1!=y0)))
        {
         if((x0==x1 && (x2==x3 && x2!=x0))||
            (x0==x2 && (x1==x3 && x1!=x0))||
            (x0==x3 && (x1==x2 && x1!=x0)))
         {
          xlo=MIN(x0,x1);
          xlo=MIN(xlo,x2);
          xlo=MIN(xlo,x3);

          xhi=MAX(x0,x1);
          xhi=MAX(xhi,x2);
          xhi=MAX(xhi,x3);

          ylo=MIN(y0,y1);
          ylo=MIN(ylo,y2);
          ylo=MIN(ylo,y3);

          yhi=MAX(y0,y1);
          yhi=MAX(yhi,y2);
          yhi=MAX(yhi,y3);

          box->x=xlo;
          box->w=xhi-xlo;
          box->y=ylo;
          box->h=yhi-ylo;

          return(1);
         }
        }
       }
      }
     }
    }
   }
  }
 }
 return(0);
}



int pathisline(int * p,boxstr * box)
{
 int x0;
 int y0;
 int x1;
 int y1;
 int swop;

 if(*p++==DRAWMOVE)
 {
  x0=*p++;
  y0=*p++;
  if(*p++==DRAWDRAW)
  {
   x1=*p++;
   y1=*p++;
   if(*p==DRAWEND)
   {
    swop=0;

    if(x1>x0) 
    {
     box->x=x0;
     box->w=x1-x0;
    }
    else
    {
     box->x=x1;
     box->w=x0-x1;
     swop^=1;
    }

    if(y1>y0)
    {
     box->y=y0;
     box->h=y1-y0;
    }
    else
    {
     box->y=y1;
     box->h=y0-y1;
     swop^=2;
    }

    if(swop & 0x1)
    {
     if(swop & 0x2) return(CLINEXR);
     else           return(CLINEX);
    }
    else
    {
     if(swop & 0x2) return(CLINER);
     else           return(CLINE);
    }
   }
  }
 }

 return(0);
}



int curveisline(curvestr * c,filestr * file,boxstr * box)
{
 pathstr  * pathp;


 if(c->type==CLINE || c->type==CLINEX || c->type==CLINER || c->type==CLINEXR)
 {
  *box=c->box;
  return(c->type);
 }
 if(c->type==CPATH)
 {
  pathp=(pathstr*)fob_find(c->data.path.path,file);
  return(pathisline(pathp->data,box));
 }

 return(0);
}



os_error * curvetoline(curvestr * c,filestr * file)
{
 int type;

 if((type=curveisline(c,file,&c->box))!=0)
 {
  c->type=type;
  return(fob_delete(c->data.path.path,file));
 }

 return(NULL);
}



static os_error * boxtopath(curvestr * c,filestr * file)
{
 os_error * err;
 int        size;
 pathstr  * path;

 size=BOXPATHSIZE*sizeof(int);

 err=fob_create(sizeof(pathstr)+size-sizeof(int),file,OPATH,(fobjptr)&path);
 if(!err)
 {
  boxpath(path->data,c);
  c->type=CPATH;
  c->data.path.path=path->key;
 }

 return(err);
}




/* constructs the path for a round box */

static void rboxpath(int * p,curvestr * c)
{
 int              radius;
 int              xradius;
 int              yradius;
 bezier_arc_coord centre;
 coordstr         cx[4];
 transformstr     tr;

 xradius=c->data.rbox.xradius;
 if(xradius>c->box.w/2) xradius=c->box.w/2;

 yradius=c->data.rbox.yradius;
 if(yradius>c->box.h/2) yradius=c->box.h/2;

 radius=MAX(xradius,yradius);

 *p++=DRAWMOVE;
 *p++=c->box.x;
 *p++=c->box.y+yradius;

 centre.x=0;
 centre.y=0;
 bezier_arc_90(centre,PI,radius,(bezier_arc_coord*)&cx[0],
                                (bezier_arc_coord*)&cx[3],
                                (bezier_arc_coord*)&cx[1],
                                (bezier_arc_coord*)&cx[2]);

 tr_setscale(&tr,radius,xradius,radius,yradius);
 tr_transformset(cx,4,&tr);

 tr_setshift(&tr,c->box.x+xradius,c->box.y+yradius);
 tr_transformset(cx,4,&tr);

 *p++=DRAWCURVE;
 *p++=cx[1].x;
 *p++=cx[1].y;
 *p++=cx[2].x;
 *p++=cx[2].y;
 *p++=cx[3].x;
 *p++=cx[3].y;

 *p++=DRAWDRAW;
 *p++=c->box.x+c->box.w-xradius;
 *p++=c->box.y;

 tr_setshift(&tr,2*c->box.x+c->box.w,0);
 tr_addscale(&tr,-1,1,1,1);

 tr_transformset(cx,4,&tr);

 *p++=DRAWCURVE;
 *p++=cx[2].x;
 *p++=cx[2].y;
 *p++=cx[1].x;
 *p++=cx[1].y;
 *p++=cx[0].x;
 *p++=cx[0].y;



 *p++=DRAWDRAW;
 *p++=c->box.x+c->box.w;
 *p++=c->box.y+c->box.h-yradius;

 tr_setshift(&tr,0,2*c->box.y+c->box.h);
 tr_addscale(&tr,1,1,1,-1);
 tr_transformset(cx,4,&tr);

 *p++=DRAWCURVE;
 *p++=cx[1].x;
 *p++=cx[1].y;
 *p++=cx[2].x;
 *p++=cx[2].y;
 *p++=cx[3].x;
 *p++=cx[3].y;


 *p++=DRAWDRAW;
 *p++=c->box.x+xradius;
 *p++=c->box.y+c->box.h;

 tr_setshift(&tr,2*c->box.x+c->box.w,0);
 tr_addscale(&tr,1,-1,1,1);
 tr_transformset(cx,4,&tr);

 *p++=DRAWCURVE;
 *p++=cx[2].x;
 *p++=cx[2].y;
 *p++=cx[1].x;
 *p++=cx[1].y;
 *p++=cx[0].x;
 *p++=cx[0].y;


 *p++=DRAWDRAW;
 *p++=c->box.x;
 *p++=c->box.y+yradius;

 *p++=DRAWCLOSE;
 *p++=DRAWEND;
}



static os_error * rboxtopath(curvestr * c,filestr * file)
{
 os_error * err;
 int        size;
 pathstr  * path;

 size=RBOXPATHSIZE*sizeof(int);

 err=fob_create(sizeof(pathstr)+size-sizeof(int),file,OPATH,(fobjptr)&path);
 if(!err)
 {
  rboxpath(path->data,c);
  c->type=CPATH;
  c->data.path.path=path->key;
 }

 return(err);
}




static void rpolypath(int * p,curvestr * c,int sides,int shape)
{
 int          i;
 coordstr     cx;
 coordstr     dx;
 coordstr     sx;
 coordstr     ix;
 transformstr tr;
 transformstr tx;
 double       angle;
 int          h;
 int          w;
 int          yshift;


 tr_setidentity(&tr);
 tr_setrotationa(&tr,(360*0x10000)/sides);


 angle=(2.0*PI)/(double)sides;

 if(sides & 0x1) /* odd  */
 {
  cx.x=0;
  cx.y=0x10000;
  h=0x10000+(int)(((double)(0x10000))*cos(angle/2.0));
  w=(int)(((double)(0x20000))*cos(angle/4.0));
  yshift=(0x20000-h)/2;
  yshift=scale(yshift,c->box.h,h);
 }
 else            /* even */
 {
  cx.x=(int)(((double)0x10000)*sin(angle/2.0));
  cx.y=(int)(((double)0x10000)*cos(angle/2.0));
  h=2*cx.y;
  if(sides & 0x2) w=2*0x10000;
  else            w=h;
  yshift=0;
 }


 if(shape==CSTAR)
 {
  ix.x=cx.x/2;
  ix.y=cx.y/2;
  tr_setidentity(&tx);
  tr_setrotationa(&tx,(360*0x10000)/(2*sides));
  tr_transformset(&ix,1,&tx);
 }

 tr_setshift(&tx,c->box.x+c->box.w/2,c->box.y+c->box.h/2-yshift);
 tr_addscale(&tx,w,c->box.w,h,c->box.h);


 for(i=0;i<sides;i++)
 {
  dx=cx;
  tr_transformset(&dx,1,&tx);

  if(i==0)
  {
   sx=dx;
   *p++=DRAWMOVE;
  }
  else
  {
   *p++=DRAWDRAW;
  }

  *p++=dx.x;
  *p++=dx.y;


  if(shape==CSTAR)
  {
   dx=ix;
   tr_transformset(&dx,1,&tx);
   *p++=DRAWDRAW;
   *p++=dx.x;
   *p++=dx.y;
   tr_transformset(&ix,1,&tr);
  }

  tr_transformset(&cx,1,&tr);
 }

 *p++=DRAWDRAW;
 *p++=sx.x;
 *p++=sx.y;

 *p++=DRAWCLOSE;
 *p++=DRAWEND;
}




static os_error * boxtorpolypath(curvestr * c,int sides,
                                              int shape,filestr * file)
{
 os_error * err;
 int        size;
 pathstr  * path;

 size=(3+3*(1+(shape==CSTAR))*sides+2)*sizeof(int);

 err=fob_create(sizeof(pathstr)+size-sizeof(int),file,OPATH,(fobjptr)&path);
 if(!err)
 {
  rpolypath(path->data,c,sides,shape);
  c->type=CPATH;
  c->data.path.path=path->key;
 }

 return(err);
}



/***************************************************************************/

static int  rc1x;
static int  rc1y;
static int  rc3x;
static int  rc3y;
static int  rc5x;
static int  rc5y;
static int  rc7x;
static int  rc7y;
static int  rx0;
static int  rx1;
static int  ry0;
static int  ry1;


/* constuct an rbox arc from (sx,sy) to (ex,ey) in given direction */
/* assumes that you are sat at given end point                     */
/* returns number of ints written                                  */


static void rboxendseg(int c,int * x,int * y,int clockwise)
{

 if(!clockwise)
 {
  c--;
  if(c<0) c=7;
 }

 switch(c)
 {
  case 0:
         *x=rx0;
         *y=rc1y;
         break;

  case 1:
         *x=rc1x;
         *y=ry0;
         break;

  case 2:
         *x=rc3x;
         *y=ry0;
         break;

  case 3:
         *x=rx1;
         *y=rc3y;
         break;

  case 4:
         *x=rx1;
         *y=rc5y;
         break;

  case 5:
         *x=rc5x;
         *y=ry1;
         break;

  case 6:
         *x=rc7x;
         *y=ry1;
         break;

  case 7:
         *x=rx0;
         *y=rc7y;
         break;
 }
}


static int rboxfindseg(int x,int y,int clockwise)
{
 if(y<rc1y)
 {
  if(x<rc1x) return(1);
  else
  if(x>rc3x) return(3);
  else       return(2);
 }
 else
 if(y>rc7y)
 {
  if(x<rc1x) return(7);
  else
  if(x>rc3x) return(5);
  else       return(6);
 }
 else
 {
  if(x<rc1x) return(0);
  else       return(4);
 }

 return(0);

 USE(clockwise);
}




static int rboxssubarc(int * p,curvestr * c,int cx,int cy,
                                            int sox,int soy,int eox,int eoy)
{
 transformstr     tr;
 bezier_arc_coord centre;
 bezier_arc_coord coord[4];
 int              xradius;
 int              yradius;
 double           start;
 double           end;
 int              ex;
 int              ey;
 int              sx;
 int              sy;




 sx=sox-cx;
 sy=soy-cy;
 ex=eox-cx;
 ey=eoy-cy;


 xradius=c->data.rbox.xradius;
 if(xradius>c->box.w/2) xradius=c->box.w/2;

 yradius=c->data.rbox.yradius;
 if(yradius>c->box.h/2) yradius=c->box.h/2;

 centre.x=centre.y=0;

 if(xradius>yradius)
 {
  sy=scale(sy,xradius,yradius);
  ey=scale(ey,xradius,yradius);

  end=atan2((double)ey,(double)ex);
  start=atan2((double)sy,(double)sx);

  bezier_arc_short(centre,start,end,xradius,&coord[0],&coord[3],
                                             &coord[1],&coord[2]);

  tr_setshift(&tr,cx,cy);
  tr_addscale(&tr,1,1,xradius,yradius);
 }
 else
 /* if(yradius>=xradius) */
 {
  sx=scale(sx,yradius,xradius);
  ex=scale(ex,yradius,xradius);

  end=atan2((double)ey,(double)ex);
  start=atan2((double)sy,(double)sx);

  bezier_arc_short(centre,start,end,yradius,&coord[0],&coord[3],
                                             &coord[1],&coord[2]);

  tr_setshift(&tr,cx,cy);
  tr_addscale(&tr,yradius,xradius,1,1);
 }

 tr_transformset((coordstr*)coord,4,&tr);


 *p++=DRAWCURVE;

 if(soy<eoy)
 {
  if(coord[1].y<soy) coord[1].y=soy;
  if(coord[2].y<soy) coord[2].y=soy;
  if(coord[1].y>eoy) coord[1].y=eoy;
  if(coord[2].y>eoy) coord[2].y=eoy;
 }
 else
 {
  if(coord[1].y>soy) coord[1].y=soy;
  if(coord[2].y>soy) coord[2].y=soy;
  if(coord[1].y<eoy) coord[1].y=eoy;
  if(coord[2].y<eoy) coord[2].y=eoy;
 }

 *p++=coord[1].x;
 *p++=coord[1].y;
 *p++=coord[2].x;
 *p++=coord[2].y;
 *p++=eox;
 *p++=eoy;

 return(7);
}



static int rboxsubarc(int * p,curvestr * c,int sx,int sy,int s,int ex,int ey)
{
 if(sx!=ex || sy!=ey)

 switch(s)
 {
  case 0:
  case 2:
  case 4:
  case 6:
         *p++=DRAWDRAW;
         *p++=ex;
         *p++=ey;
         return(3);
         break;

  case 1:
         return(rboxssubarc(p,c,rc1x,rc1y,sx,sy,ex,ey));
         break;

  case 3:
         return(rboxssubarc(p,c,rc3x,rc3y,sx,sy,ex,ey));
         break;

  case 5:
         return(rboxssubarc(p,c,rc5x,rc5y,sx,sy,ex,ey));
         break;

  case 7:
         return(rboxssubarc(p,c,rc7x,rc7y,sx,sy,ex,ey));
         break; 
 }

 return(0); /* never */
}



int rboxarc(int * p,curvestr * c,int sx,int sy,int ex,int ey,int clockwise)
{
 int s;
 int e;
 int n;
 int cx;
 int cy;
 int sum;
 int xradius;
 int yradius;


 xradius=c->data.rbox.xradius;
 if(xradius>c->box.w/2) xradius=c->box.w/2;

 yradius=c->data.rbox.yradius;
 if(yradius>c->box.h/2) yradius=c->box.h/2;


 rx0=c->box.x;
 rx1=rx0+c->box.w;
 ry0=c->box.y;
 ry1=ry0+c->box.h;

 rc1x=c->box.x+xradius;
 rc1y=c->box.y+yradius;
 rc3x=c->box.x+c->box.w-xradius;
 rc3y=rc1y;
 rc5x=rc3x;
 rc5y=c->box.y+c->box.h-yradius;
 rc7x=rc1x;
 rc7y=rc5y;

 sum=0;

 s=rboxfindseg(sx,sy,clockwise);
 e=rboxfindseg(ex,ey,clockwise);
 n=s;

 while(1)
 {
  if(n==e)
  {
   cx=ex;
   cy=ey;
  }
  else
  {
   rboxendseg(n,&cx,&cy,clockwise);
  }

  sum+=rboxsubarc(p+sum,c,sx,sy,n,cx,cy);

  sx=cx;
  sy=cy;

  if(n==e) break;

  if(clockwise)
  {
   n++;
   if(n>7) n=0;
  }
  else
  {
   n--;
   if(n<0) n=7;
  }
 }

 return(sum);  /* no of ints written */
}


/**************************************************************************/


void curvepath(curvestr * c,int * path)
{
 switch(c->type)
 {
  case     CBOX:
                boxpath(path,c);
                break;

  case    CBOXR:
                rboxpath(path,c);
                break;

  case CELLIPSE:
                ellipsepath(path,c,NULL);
                break;

     case CLINE:
    case CLINEX:
    case CLINER:
   case CLINEXR:
                linepath(path,c,c->type);
                break;
 }
}



os_error * curvetopath(curvestr * c,filestr * file)
{
 os_error * err;

 err=NULL;

 switch(c->type)
 {
  case     CBOX:
                err=boxtopath(c,file);
                break;

  case    CBOXR:
                err=rboxtopath(c,file);
                break;

  case CELLIPSE:
                err=ellipsetopath(c,file);
                break;

     case CLINE:
    case CLINEX:
   case CLINEXR:
    case CLINER:
                linetopath(c,c->type,file);
                break;

 }

 return(err);
}


/*****************************************************************************/

void rcreatepathstyle(pathstyle * s,int wimpcolour)
{
 int style;

 s->action=1;
 s->thickness=0;

 style=(wimpcolour>>8)&0xFF;

 if(!style) s->style=-1;
 else       s->style=(style<<24)|(style<<16)|(style<<8)|style;

 s->stylescale=256*MAX(cvdu.deltax,cvdu.deltay);
 s->endcaps=0;
 s->startcaps=0;
 s->endwidth=s->endheight=1;
 s->startwidth=s->startheight=1;
 s->join=2;
 s->windingrule=2;
}


static os_error * outcolour(viewredrawstr * v,int wimpcolour,
                            colourvalue * cv,int * fillp)
{
 os_error * err;
 int        transparent;
 int        fill;


 err=NULL;
 transparent=0;
 fill=*fillp;


 if(wimpcolour & OUTLINEEOR)
 {
  if(wimpcolour & OUTLINECVBACK)
  {
   rseteorcolpal(vfindcolour(v->view,cv),wimpcolour & 0xF);
  }
  else
  {
   rseteorcol((wimpcolour>>4) & 0xF,wimpcolour & 0xF);
  }
 }
 else
 {
  if(((wimpcolour & OUTLINECVFILL) && fill) ||
     ((wimpcolour & OUTLINECVLINE) && !fill))
  {
                                        /* assume does not move memory! */
//   err=vsetcolour(cv,&transparent,v->view);
   err=render_setcolour(cv,&transparent,v);
  }
  else
  if(((wimpcolour & OUTLINEWIMPFILL) && fill) ||
     ((wimpcolour & OUTLINEWIMPLINE) && !fill))
  {
   err=render_setwimpcolour(wimpcolour & 0xF,v);
  }
 }

 *fillp=!transparent;
 return(err);
}


/* width>=0 use Draw module else use OS */


os_error * outlinesections(viewredrawstr * v,int * p,int wimpcolour,
                                  colourvalue * cv,int width,transformstr * t)
{
 os_error * err;
 pathstyle  style;  
 int        fill;

 err=NULL;

// cprintf("outline sections");

 if(wimpcolour & (OUTLINECVLINE|OUTLINEWIMPLINE))
 {
  fill=0;
  outcolour(v,wimpcolour,cv,&fill);

  if(fill)
  {
   if(width>=0) 
   {
    rcreatepathstyle(&style,0);
    style.thickness=width;
    err=rpathline(p,&style,v,t);
   }
   else
   {
    err=rsimpleline(p,v,t);
   }
  }
 }
 return(err);
}


os_error * outlineline(viewredrawstr * v,curvestr * c,int wimpcolour,
                       colourvalue * cv,transformstr * t,int type)
{
 os_error * err;
 int        fill;
 int        style;
 int        x0;
 int        x1;
 int        y0;
 int        y1;

 err=NULL;

 if(wimpcolour & (OUTLINECVLINE|OUTLINEWIMPLINE))
 {
  fill=0;
  outcolour(v,wimpcolour,cv,&fill);

  if(fill)
  {
   x0=c->box.x;
   x1=c->box.x+c->box.w;
   y0=c->box.y;
   y1=c->box.y+c->box.h;
   if(type==CLINEX || type==CLINEXR) SWOP(x0,x1);
   if(type==CLINER || type==CLINEXR) SWOP(y0,y1);

   err=rtbbc_move(x0,y0,t,v);

   if(!err)
   {
    style=(wimpcolour>>8)&0xFF;
    if(style)
    {
     rsetdotstyle(style);
     err=rtbbc_drawstyle(x1,y1,t,v);
    }
    else
    {
     err=rtbbc_draw(x1,y1,t,v);
    }
   }
  }
 }

 return(err);
}



os_error * outlinebox(viewredrawstr * v,curvestr * c,int wimpcolour,
                                         colourvalue * cv,transformstr * t)
{
 os_error * err;
 int        ttype;
 coordstr   coord[3];
 int        fill;

 err=NULL;

// cprintf("outlinebox");

 coord[0].x=c->box.x;
 coord[0].y=c->box.y;

 coord[1].x=c->box.x+c->box.w;
 coord[1].y=c->box.y;

 coord[2].x=c->box.x+c->box.w;
 coord[2].y=c->box.y+c->box.h;

 tr_transformset(coord,3,t); /* NB this is always returns TR_GENERAL*/
 ttype=tr_type(t);
 if(ttype==TR_SHIFT) coord[1]=coord[2];

 if(wimpcolour & (OUTLINECVFILL|OUTLINEWIMPFILL))
 {                                         
  fill=1;
  outcolour(v,wimpcolour,cv,&fill);
  if(fill)
  {
   if(ttype==TR_SHIFT) err=rrectangle(coord,-1,v); /* fill in background */
   else                err=rparallel(coord,-1,v);
  }
 }

 if(wimpcolour & (OUTLINECVLINE|OUTLINEWIMPLINE))
 {
  fill=0;
  outcolour(v,wimpcolour,cv,&fill);

  if(fill)
  {
   if(ttype==TR_SHIFT) err=rrectangle(coord,0,v);
   else                err=rparallel(coord,0,v);
  }
 }

 return(err);
}




os_error * outlinecross(viewredrawstr * v,boxstr * box,int wimpcolour,
                                                         colourvalue * cv)
{
 os_error * err;
 coordstr   coord[8];
 int        fill;
 int        n;

 err=NULL;

 n=0;

 if(box->h)
 {
  coord[n].x=box->x;
  coord[n++].y=v->ylos;
  coord[n].x=box->x;
  coord[n++].y=v->yhis;
 }

 if(box->w && box->h)
 {      
  coord[n].x=box->x+box->w;
  coord[n++].y=v->ylos;
  coord[n].x=box->x+box->w;
  coord[n++].y=v->yhis;
 }

 if(box->w)
 {
  coord[n].x=v->xlos;
  coord[n++].y=box->y;
  coord[n].x=v->xhis;
  coord[n++].y=box->y;
 }

 if(box->h && box->w)
 {
  coord[n].x=v->xlos;
  coord[n++].y=box->y+box->h;
  coord[n].x=v->xhis;
  coord[n++].y=box->y+box->h;
 }


 if(wimpcolour & (OUTLINECVLINE|OUTLINEWIMPLINE))
 {
  fill=0;
  outcolour(v,wimpcolour,cv,&fill);

  if(fill)
  {
   err=rcross(coord,0,n,v);
  }
 }

 return(err);
}










static os_error *outlinepaths(viewredrawstr * v,curvestr * c,int wimpcolour,
                                         colourvalue * cv,transformstr * t)
{
 os_error * err;
 pathstr  * path;
 pathstyle  style;
 int        fill;
 int      * p;
 int        data[MAXPATHSIZE];


 err=NULL;

 rcreatepathstyle(&style,wimpcolour);

 if(c->type==CPATH)
 {
  path=(pathstr*)fob_find(c->data.path.path,v->file);
  p=path->data;
 }
 else
 {
  curvepath(c,data);
  p=data;
 }

 if(wimpcolour & (OUTLINECVFILL|OUTLINEWIMPFILL))
 {                                         
  fill=1;
  outcolour(v,wimpcolour,cv,&fill);
  if(fill)
  {
   style.action=2;
   err=rpathline(p,&style,v,t);
  }
 }

 if(wimpcolour & (OUTLINECVLINE|OUTLINEWIMPLINE))
 {
  fill=0;
  outcolour(v,wimpcolour,cv,&fill);

  if(fill)
  {
   style.action=1;
   if(!err) err=rpathline(p,&style,v,t);
  }
 }

 return(err);
}


/* use for outline boxes on screen */
/* for these functions wimpcolour is split into a number of fields    */
/* byte 0 lowest  4 bits wimp foreground                              */
/* byte 0 highest 4 bits wimp background                              */
/* byte 1 dot pattern byte                                            */
/* byte 2                    bit 0 EOR                                */
/*                           bit 1 USE cv for fill                    */
/*                           bit 2 USE cv for line                    */
/*                           bit 3 USE wimp for fill                  */
/*                           bit 4 USE wimp for line                  */
/*                           bit 5 USE cv   for background else wimp  */
/*                           bit 6 use draw module for output         */

os_error * outlinecurve(viewredrawstr * v,curvestr * c,int wimpcolour,
                                    colourvalue * cv,transformstr * t)
{

 if(wimpcolour)
 {
  switch(c->type)
  {
       case CBOX:
                 if(wimpcolour & OUTLINEUSEPATH) return(outlinepaths(v,c,wimpcolour,cv,t));
                 else                            return(outlinebox(v,c,wimpcolour,cv,t));
                 break;

#ifdef NEVER
     case CBOXR:
                return(outlinerbox(v,c,wimpcolour,cv,t));
                break;
 
  case CELLIPSE:
                return(outlineellipse(v,c,wimpcolour,cv,t));
                break;
#endif


      case CBOXR:
   case CELLIPSE:
      case CPATH:
                 return(outlinepaths(v,c,wimpcolour,cv,t));
                 break;

      case CLINE:
     case CLINEX:
     case CLINER:
    case CLINEXR:
                 if(wimpcolour & OUTLINEUSEPATH) return(outlinepaths(v,c,wimpcolour,cv,t));
                 else                            return(outlineline(v,c,wimpcolour,cv,t,c->type));
                 break;

  }
 }

 return(NULL);
}



/*****************************************************************************/




static os_error * renderpathsub(viewredrawstr * v,int * path,
                                      curvestylestr * cstyle,transformstr * t)
{
 os_error * err;
 int        transparent;
 int        action;

 err=NULL;

 action=cstyle->pstyle.action;

 if(t &&              /* balancing clause below */
    (cstyle->pstyle.thickness || cstyle->pstyle.style!=-1) && 
    (action & 0x1))
 {
  pathtransform(path,t);
  t=NULL;
 }

 if(action & 0x2)
 {
  err=render_setcolour(&cstyle->fillcolour,&transparent,v);
  if(!err)
  {
   if(!transparent)
   {              /* fill in background */
    cstyle->pstyle.action=2;
    rpathline(path,&cstyle->pstyle,v,t);
   }
  }
 }

 if(action & 0x1)
 {
  err=render_setcolour(&cstyle->linecolour,&transparent,v);
  if(!err)
  {
   if(!transparent)
   {
    cstyle->pstyle.action=1;
    rpathline(path,&cstyle->pstyle,v,t);
   }
  }
 }

 return(err);
}




os_error * rendercurve(viewredrawstr * v,curvestr * c,
                                 curvestylestr * cstyle,transformstr * t)
{
 os_error * err;
 pathstr  * path;
 int      * pathp;
 int        data[MAXPATHSIZE];

// cprintf("render curve");

 if(c->type==CPATH)
 {
  if(t &&                  /* balancing clause above */
     (cstyle->pstyle.thickness || cstyle->pstyle.style!=-1) &&
     (cstyle->pstyle.action & 0x1))
  {
   err=pathflex(&pathp,c->data.path.path,v->file);
   if(!err)
   {
    err=renderpathsub(v,pathp,cstyle,t);
    flex_free((flex_ptr)&pathp);
   }
  }
  else
  {
   path=(pathstr*)fob_find(c->data.path.path,v->file);
   err=renderpathsub(v,path->data,cstyle,t);
  }
 }
 else
 {
  curvepath(c,data);
  err=renderpathsub(v,data,cstyle,t);
 }

 return(err);
}


/****************************************************************************/



static os_error * truecurvesub(int * path,pathstyle * pstyle,
                                              transformstr * tr,boxstr * box)
{
 transformbitstr b;
 transformstr    t;
 transformstr    i;
 coordstr        cx[2];

 if(tr)
 {
  tr_getbits(tr,&b);
  tr_setbits(&t,&b,TR_XSCALE|TR_YSCALE);
  pathtransform(path,&t);
 }

 drawboxsize(cx,path,pstyle);

 if(tr)
 {
  tr_inverse(&t,&i);
  tr_transformset(cx,2,&i);
 }

 box->x=cx[0].x;
 box->y=cx[0].y;
 box->w=cx[1].x-box->x;
 box->h=cx[1].y-box->y;

 return(NULL);
}


os_error * truecurvebox(curvestr * c,pathstyle * pstyle,
                        filestr * file,transformstr * tr,boxstr * box)
{
 os_error * err;
 int        path[MAXPATHSIZE];
 int      * pathp;

 err=NULL;

 if(pstyle->thickness)
 {
  switch(c->type)
  {
      case CBOX:
     case CBOXR:
  case CELLIPSE:
     case CLINE:
    case CLINEX:
    case CLINER:
   case CLINEXR:
                curvepath(c,path);
                err=truecurvesub(path,pstyle,tr,box);
                break;

     case CPATH:
                err=pathflex(&pathp,c->data.path.path,file);
                if(!err)
                {
                 err=truecurvesub(pathp,pstyle,tr,box);
                 flex_free((flex_ptr)&pathp);
                }
                break;

  }
 }
 else
 {
  *box=c->box;
 }

 return(err);
}

/*************************************************************************/




void correctcurve(transformstr * xnew,transformstr * old,curvestr * curve)
{
 transformbitstr ob;
 transformbitstr nb;

 if(curve->type==CBOXR)
 {
  tr_getbits(xnew,&nb);
  tr_getbits(old,&ob);

  curve->data.rbox.xradius=scale(curve->data.rbox.xradius,ob.xscale,nb.xscale);
  curve->data.rbox.yradius=scale(curve->data.rbox.yradius,ob.yscale,nb.yscale);
 }
}



void setcurveradius(curvestr * f,int radius,transformstr * tr)
{
 transformbitstr b;

 if(f->type==CBOX || f->type==CBOXR)
 {
  if(radius)
  {
   tr_getbits(tr,&b);
   f->data.rbox.xradius=scale(radius,0x10000,b.xscale);
   f->data.rbox.yradius=scale(radius,0x10000,b.yscale);
   f->type=CBOXR;
  }
  else
  {
   f->type=CBOX;
  }
 }
}



/* assume src and dest are fixed in memory */

os_error * copycurve(curvestr * dest,filestr * dfile,
                                              curvestr * src,filestr * sfile)
{
 os_error * err;
 pathstr  * path;

 *dest=*src;
 if(src->type==CPATH)
 {
  err=fob_copy((fobjptr)&path,dfile,src->data.path.path,sfile);
  dest->data.path.path=path->key;
 }
 else err=NULL;

 return(err);
}



os_error * deletecurve(curvestr * c,filestr * file)
{
 int path;

 if(c->type==CPATH) path=c->data.path.path;
 else               path=0;

 memset(c,0,sizeof(curvestr));

 return(fob_delete(path,file));
}


os_error * newcurve(curvestr * c,int type,int ** xpath,int size,filestr * file)
{
 os_error * err;
 pathstr  * path;

 err=NULL;

 memset(c,0,sizeof(curvestr));
 c->type=type;
 if(c->type==CPATH)
 {
  err=fob_create(sizeof(pathstr)+size,file,OPATH,(fobjptr)&path);
  if(!err)
  {
   c->data.path.path=path->key;
   XMEMCPYX(path->data,*xpath,size);
  }
 }

 return(err);
}


os_error * makecurve(curvestr * c,int shape,int data,filestr * file)
{
 os_error * err;

 err=NULL;

 if(shape==CBOX || shape==CELLIPSE)
 {
  c->type=shape;
  c->data.rbox.xradius=c->data.rbox.yradius=0; /* zero data area of curve */
 }
 else
 if(shape==CPATH)
 {
  if(c->type==CELLIPSE) err=ellipsetopath(c,file);
  else
  if(c->type==CBOX)     err=boxtopath(c,file);
  else
  if(c->type==CBOXR)    err=rboxtopath(c,file);
 }
 else
 if(shape==CREGULAR || shape==CSTAR)
 {
  err=boxtorpolypath(c,data,shape,file);
 }

 return(err);
}




os_error * nullcurve(curvestr * c)
{
 memset(c,0,sizeof(curvestr));
 c->type=CNULL;
 return(NULL);
}


/***************************************************************************/

static os_error * transformpath(transformstr * tr,curvestr * cx,filestr * file)
{
 os_error * err;
 pathstr  * p;

 err=fob_mod(cx->data.path.path,file,(fobjptr)&p);
 if(!err)
 {
  pathtransform(p->data,tr);
 }

/* getmaxbox(&cx->box,&cx->box,tr); */

 validpath(&cx->box,&cx->data.path.flags,p->data);

 return(NULL);
}



static os_error * transformrbox(transformstr * tr,curvestr * c,filestr * file)
{
 os_error        * err;
 transformbitstr   b;
 int               type;

 err=NULL;

 type=tr_getbits(tr,&b);  /* better to use transform type */

 if(type & (TR_ROTATE | TR_SKEWXY | TR_SKEWYX))
 {
   /* convert to a path */
           err=rboxtopath(c,file);
  if(!err) err=transformpath(tr,c,file);
 }
 else
 {
  c->box.x=scale(c->box.x,b.xscale,0x10000)+b.xshift;
  c->box.y=scale(c->box.y,b.yscale,0x10000)+b.yshift;
  c->box.w=scale(c->box.w,b.xscale,0x10000);
  c->box.h=scale(c->box.h,b.yscale,0x10000);
  c->data.rbox.xradius=scale(c->data.rbox.xradius,b.xscale,0x10000);
  c->data.rbox.yradius=scale(c->data.rbox.yradius,b.yscale,0x10000);
 }

 return(err);
}



static os_error * transformbox(transformstr * tr,curvestr * c,filestr * file)
{
 os_error        * err;
 transformbitstr   b;
 int               type;

 err=NULL;

 type=tr_getbits(tr,&b);   /* better to use transform type */

 if(type & (TR_ROTATE | TR_SKEWXY | TR_SKEWYX))
 {
   /* convert to a path */
           err=boxtopath(c,file);
  if(!err) err=transformpath(tr,c,file);
 }
 else
 {
  c->box.x=scale(c->box.x,b.xscale,0x10000)+b.xshift;
  c->box.y=scale(c->box.y,b.yscale,0x10000)+b.yshift;
  c->box.w=scale(c->box.w,b.xscale,0x10000);
  c->box.h=scale(c->box.h,b.yscale,0x10000);
 }

 return(err);
}



static os_error * transformellipse(transformstr * tr,curvestr * c,
                                                     filestr * file)
{
 os_error        * err;
 transformbitstr   b;
 int               type;

 err=NULL;

 type=tr_getbits(tr,&b);   /* better to use transform type */

 if(type & (TR_ROTATE | TR_SKEWXY | TR_SKEWYX))
 {
   /* convert to a path */
           err=ellipsetopath(c,file);
  if(!err) err=transformpath(tr,c,file);
 }
 else
 {
  c->box.x=scale(c->box.x,b.xscale,0x10000)+b.xshift;
  c->box.y=scale(c->box.y,b.yscale,0x10000)+b.yshift;
  c->box.w=scale(c->box.w,b.xscale,0x10000);
  c->box.h=scale(c->box.h,b.yscale,0x10000);
 }

 return(err);
}



os_error * transformcurve(transformstr * tr,curvestr * c,filestr * file)
{
 os_error * err;

 err=NULL;

 switch(c->type)
 {
      case CBOX:
                err=transformbox(tr,c,file);
                break;

     case CBOXR:
                err=transformrbox(tr,c,file);
                break;

  case CELLIPSE:
                err=transformellipse(tr,c,file);
                break;

     case CPATH:
                err=transformpath(tr,c,file);
                break;
 }


 return(err);
}





/****************************************************************************/

/* get bounding box for a path */

void getpathbox(boxstr * box,int * p)
{
 int xlo;
 int xhi;
 int ylo;
 int yhi;
 int x;
 int y;
 int lx;
 int ly;
 bezier b;
 aboxstr abox;


 p++;                      /* skip move - we hope */
 x=xlo=xhi=*p++;
 y=ylo=yhi=*p++;

 while(*p!=DRAWEND)
 {
  lx=x;
  ly=y;

/*  printf("get path box=%d\n",*p); */


  if(*p==DRAWCLOSE)
  {
   p++;
  }
  else
  if(*p==DRAWDRAW || *p==DRAWMOVE)
  {
   p++;
   x=*p++;
   y=*p++;

   xlo=MIN(x,xlo);
   xhi=MAX(x,xhi);
   ylo=MIN(y,ylo);
   yhi=MAX(y,yhi);
  }
  else
  if(*p==DRAWCURVE)
  {
   p++;

   b.left.x=lx;
   b.left.y=ly;

   b.leftc.x=*p++;
   b.leftc.y=*p++;
   b.rightc.x=*p++;
   b.rightc.y=*p++;

   b.right.x=x=*p++;
   b.right.y=y=*p++;

   bezgetext(&b,&abox);
   xlo=MIN(xlo,abox.x0);
   xhi=MAX(xhi,abox.x1);
   ylo=MIN(ylo,abox.y0);
   yhi=MAX(yhi,abox.y1);
  }
 }

 box->x=xlo;
 box->w=xhi-xlo;
 box->y=ylo;
 box->h=yhi-ylo;
}




static os_error * cmppath(curvestr * c1,curvestr * c2,int * match,
                                                      filestr * file)
{
 os_error * err;
 pathstr  * p1;
 pathstr  * p2;
 int      * q1;
 int      * q2;
 int        x;
 int        y;

 err=fob_lock((fobjptr)&p1,c1->data.path.path,file);
 if(!err)
 {
  err=fob_lock((fobjptr)&p2,c2->data.path.path,file);
  if(!err)
  {
   q1=p1->data;
   q2=p2->data;

   while(1)
   {
    x=*q1++;
    y=*q2++;
    if(x!=y) break;

    if(x==DRAWEND || x==DRAWCLOSE)
    {
     *match=1;
     break;
    }
    else
    if(x==DRAWDRAW || x==DRAWMOVE)
    {
     if((*q1!=*q2) || (*(q1+1)!=*(q2+1))) break; 
     q1+=2;
     q2+=2;
    }
    else
    if(x==DRAWCURVE)
    {
     if(memcmp(q1,q2,sizeof(int)*3*2)) break;
     q1+=6;
     q2+=6;
    }
   }
   err=fob_unlock((fobjptr)&p2,file);
  }
  err=fob_unlock((fobjptr)&p1,file);
 }
 return(err);
}



/* are two curves identical ? */

os_error * cmpcurves(curvestr * c1,curvestr * c2,int * match,filestr * file)
{
 os_error * err;

 err=NULL;
 *match=0;

 if(c1->type==c2->type && cmpbox(&c1->box,&c2->box))
 {
  switch(c1->type)
  {
   case     CBOX:
   case CELLIPSE:
                 *match=1;
                 break;

   case    CBOXR:
                 *match=(c1->data.rbox.xradius==c2->data.rbox.xradius) &&
                        (c1->data.rbox.yradius==c2->data.rbox.yradius);
                 break;

   case    CPATH:
                 err=cmppath(c1,c2,match,file);
                 break;
  }
 }

 return(err);
}






int pathsize(int * p)
{
 int * q;

 q=p;

 while(1)
 {
  switch(*p)
  {
   case DRAWMOVE:
   case DRAWDRAW:
                 p+=3;
                 break;

  case DRAWCURVE:
                 p+=7;
                 break;

  case DRAWCLOSE:
                 p++;
                 break;

  case   DRAWEND:
                 p++;
                 return((p-q)*sizeof(int));
                 break;

  }
 }

 return(0);
}





os_error * pathflex(int ** pathpp,int path,filestr * file)
{
 os_error * err;
 pathstr  * pathp;
 int        size;
 int      * p;

 pathp=(pathstr*)fob_find(path,file);
 p=pathp->data;

 size=pathsize(p);

 err=flex_alloc((flex_ptr)pathpp,size);
 if(!err)
 {
  pathp=(pathstr*)fob_find(path,file);
  p=pathp->data;

  memmove(*pathpp,p,size);
 }

 return(err);
}



void pathtransform(int * q,transformstr * tr)
{
 coordstr c[3];
 int      x;

 while(1)
 {
  x=*q;

  if(x==DRAWEND)
  {
   break;
  }
  else
  if(x==DRAWCLOSE)
  {
   q++;
  }
  else
  if(x==DRAWDRAW || x==DRAWMOVE)
  {
   XMEMCPYX(c,q+1,2*sizeof(int));
   tr_transformset(c,1,tr);
   XMEMCPYX(q+1,c,2*sizeof(int));
   q+=3;
  }
  else
  if(x==DRAWCURVE)
  {
   XMEMCPYX(c,q+1,6*sizeof(int));
   tr_transformset(c,3,tr);
   XMEMCPYX(q+1,c,6*sizeof(int));
   q+=7;
  }
 }
}



int * clippath;


/* int getclipcurve(int n); */

int getclipcurve(int fp)
{
 return(clippath[stack[fp]]);
}




os_error * setupclipcurve(curvestr * c,transformstr * tr,filestr * file)
{
 os_error * err;

 err=NULL;

 if(!clippath)
 {
  if(c->type==CPATH)
  {
   err=pathflex(&clippath,c->data.path.path,file);
  }
  else
  {
   err=flex_alloc((flex_ptr)&clippath,MAXPATHSIZE*sizeof(int));
   if(!err) curvepath(c,clippath);
  }
  if(!err) pathtransform(clippath,tr);
 }

 return(err);
}



os_error * setclipcurve(curvestr * c,transformstr * tr,viewredrawstr * v)
{
// transformstr restr;
 transformstr drtr;
// transformstr ostr;

 setupclipcurve(c,tr,v->file);
 
 getdrtr(&drtr,NULL,v);

 transformstr result;

 tr_multiply(&result,&drtr,&v->tr);


 bbc_startpath();

// tr_setscale(&ostr,256,1,256,1);

// tr_multiply(&restr,&drtr,&ostr);

 Draw_ExpandPath(clippath,&result,DRAWMOD_OSUNITS);

 bbc_endpath();

 bbc_gclip(0);

 return(NULL);
}



os_error * clearclipcurve(void)
{
 return(flex_free((flex_ptr)&clippath));
}



