/*->h.str */


/*****************************************************************************/
/* structures common to main part of document */

typedef struct flagstr
{
 unsigned int selected:1;
 unsigned int invisible:1;
 unsigned int local:1;     /* master object made local */
 unsigned int changed:1;
 unsigned int locked:1;
 unsigned int spread:1;
 unsigned int masterdata:1;
 unsigned int full:1;
 unsigned int blanked:1;

 unsigned int version:5;


} flagstr;



typedef struct objectbitstr
{
 unsigned int local:1;
 unsigned int locked:1;
 unsigned int localdata:1;    /* e.g. added text to null master frame */
 unsigned int canmakelocal:1;
 unsigned int spread:1;
 unsigned int noresize:1;
 unsigned int vlock:1;
 unsigned int hlock:1;
 unsigned int norotate:1;
 unsigned int nodelete:1;
 unsigned int nocopy:1;
 unsigned int nosectionmove:1;
 unsigned int nopagemove:1;
 unsigned int nogroup:1;
 unsigned int noshape:1;
 unsigned int canmirror:1;    /* 16 bits */

} objectbitstr;



typedef struct chainstr
{
 int next;
 int prev;
 int child;
 int parent;
} chainstr;


typedef struct chainobject
{
 int      key;
 int      type;
 chainstr chain;
} chainobject;



typedef struct boxstr
{
 int x;
 int y;
 int w;
 int h;
} boxstr;



typedef struct aboxstr
{
 int x0;
 int y0;
 int x1;
 int y1;
} aboxstr;




#define DRAWEND   0     /* end of path                                   */
#define DRAWMOVE  2     /* move to (x,y), starts new subpath             */
#define DRAWDRAW  8     /* line to (x,y)                                 */
#define DRAWCURVE 6     /* bezier curve to (x3,y3) with 2 control points */
#define DRAWCLOSE 5     /* close current subpath with a line             */



/* types of curve */

#define CBOX     0x0
#define CBOXR    0x1
#define CELLIPSE 0x2
#define CPATH    0x3
#define CLINE    0x4
#define CREGULAR 0x5
#define CNULL    0x6
#define CLINEX   0x7
#define CLINER   0x8
#define CLINEXR  0x9
#define CSTAR    0xA



typedef struct curvestr
{
 int     type; /* box, ellipse, circle */
 boxstr  box;  /* bounding box         */

 union
 {

  struct
  {
   int xradius;
   int yradius;
  } rbox;

  struct
  {
   int path;
   int flags;
  } path;

 } data;

} curvestr;



typedef struct pathstr
{
 int            key;
 int            type;
 int            data[1];
} pathstr;



#define FRONONE    0
#define FROFRAME   1
#define FROUSER    2

#define FRFWIDEST  0
#define FRFLEFT    1
#define FRFRIGHT   2
#define FRFBOTH    3
#define FRFNEITHER 4


typedef struct textflowstr    /* controls how text flows around boxes */
{
 unsigned int o:24;
 unsigned int fmode:4;
 unsigned int omode:3;
 unsigned int changeo:1;
} textflowstr;



/****************************************************************************/

typedef struct zoomstr
{
 int mul;
 int div;
} zoomstr;


#define GRIDLOCK     0x1
#define GRIDBASELOCK 0x2

typedef struct gridstr
{
 unsigned int     show:1;
 unsigned int     baseline:1;
 unsigned int     basecolour:5;
 unsigned int     sparebits:25;
 int              lock;
 int              colour;
 int              spacing;    /* assume these two located together */
 int              spacediv;
 int              subdiv;
 int              spare[2];

} gridstr;


typedef struct borderstr
{
 int toppb;            /* paste board before page */
 int botpb;            /* paste board after  page */

 int leftpb;           /* paste board to left               */
 int mid1pb;           /* paste board to right of left page */
 int mid2pb;           /* paste board to left of right page */
 int rightpb;          /* paste board to right              */

 int firstborder;      /* first interpage border */
 int hmidborder;       /* mid interpage border */ 
 int lastborder;       /* last interpage border */

 int leftborder;       /* left interpage border */
 int vmidborder;       /* mid interpage border */ 
 int rightborder;      /* right interpage border */

} borderstr;



typedef struct viewdatastr
{
 int              dble;
 int              rulers;
 zoomstr          zoom;                /* scale              */
 zoomstr          defzoom;             /* var zoom value 80% */
 int              zoomopt;             /* 50%, 200% etc.     */
 int              units;
 int              tools;
 int              buttons;
 int              infopal;
 int              guides;              /* show guidelines */
 int              pictures;
 int              margins;
 int              invisibles;
 int              tabruler;
 int              colourpal;
 gridstr          grid;

 int              pasteboardxwidth;    /* pb width */
 int              pasteboardywidth;    /* pb width */
 int              pasteboard;

 zoomstr          prevzoom;            /* previous zoom */
 int              prevzoomopt;

 unsigned int     noclicks:1;
 unsigned int     noload:1;
 unsigned int     nodnd:1;
 unsigned int     nodeletewindow:1;
 unsigned int     help2:1;
 unsigned int     spell:1;
 unsigned int     noproxies:1;
 unsigned int     nomenu:1;            // 8
 unsigned int     toolboxdocked:3;
 unsigned int     infopaldocked:3;     // 14

 int              rulerx;
 int              rulery;
 int              guidelinefb;

 int              toolsxshift;
 int              toolsyshift;
 int              toolboxx;
 int              toolboxy;

 int              infoxshift;
 int              infoyshift;
 int              infowheight;

	
 int              spare[8];

} viewdatastr;


#define VIEWNAMELEN 128

typedef struct viewstr
{
 TCHAR            name[VIEWNAMELEN];
 struct filestr * file;
 wimp_w           handle;             /* main window handle for this view   */
// int              mainpane;

 struct viewstr * next;               /* pointer to next view for this file */
 struct viewstr * prev;               /* pointer to prev view for this file */

 int              branch;             /* master/doc branch                  */
 int              page;               /* first page */
 int              section;
 int              ps;                 /* y offset of first page             */

 int              dpage;              /* display page */
 int              dsection;

 windowstr        mainwindow;         /* current window info */
// windowstr        mainpanewindow;
 wimp_box         extent;             /* window extent       */

 viewdatastr      data;

// int              scalefactor;

 int              vruler;
 int              vshift;
 int              hruler;
 int              hshift;       // button bar height
 int              rulerpatch;
 
 wimp_w           vpane;
 wimp_w           hpane;
 wimp_w           info;         /* handle of info palette */

 int              hextent;           
 int              bextent;

 int              vpshift;      // vertical ruler width
 int              hpshift;      // horizontal ruler height

 borderstr        border;

 int              claimedzero;  /* have tools added a zero event */

 int              bshift;       /* button shift */

 wimp_w           hpane2;

 void           * vxcache;
 int              vxcachen;

 unsigned int     nosnap:1;
 unsigned int     printview:1;

 void           * process;
 int              processvalue;

 wimp_box         prevbox;

	int              mhandle;

 void           * olehandle;

 int              left;         // margins contain rulers, buttons, tools
 int              right;
 int              top;
 int              bottom;

	wimp_w           buttontips;

 int              spare;

} viewstr;



/*****************************************************************************/


#define DOC 0
#define MST 1

typedef struct filehdr
{
 int          key;
 int          type;
 chainstr     chain;
 flagstr      flags;

 int          colours;    /* pointer to colour defns              */
 int          section[2]; /* pointers to doc and master sections  */
 int          ns[2];      /* number of sections                   */
 int          np[2];      /* number of pages                      */
 viewdatastr  viewdata;

 int          ustyles;     /* user styles     */
 int          astyles;     /* absolute styles */
 int          tabs;        /* tab rulers      */
 int          cstyles;     /* style chains    */

 int          flowsection; /* highest priority unflowed section */

 int          spell;
 int          print;
 int          fmap;        /* font map */
 int          undo;
 int          preferences;
 int          indexcon;

 int          maxpagewidth[2]; /* max width of all pages */
 int          maxpageheight[2]; /* max height of all pages */

 int          border;

 int          bms;

 int          colsep;
 int          tag;
 int          fields;


 int          spare[21];

} filehdr;



/****************************************************************************/

#define FILENAMELEN (MAX_PATH)

typedef struct filestr
{
 char           * name;

 int              modded[2];            /* file modified ?        */

 struct filestr * next;                 /* next file in chain     */
 struct filestr * prev;                 /* prev file in chain     */

 viewstr        * view;                 /* pointer to first view  */
 int              viewn[2];             /* how many views         */

 int              xxxhdr;               /* pointer to file header */
                                        /* the key to everything  */

 struct domainstr * domain;               /* domain                 */

 unsigned int     permanent:1;          /* don't lose file when views closed */
 unsigned int     autosave:1;           /* flag if autosave scheduled        */
 unsigned int     saved:1;
 unsigned int     oldformat:1;
 unsigned int     noview:1;
 unsigned int     merge:1;
 unsigned int     savedonce:1;
 unsigned int     writeable:1;
 unsigned int     readable:1;
 unsigned int     notools:1;
 unsigned int     writeunlocked:1;
 unsigned int     readunlocked:1;
 unsigned int     alwayspermanent:1;    /* never lose file */
 unsigned int     mergedata:1;
 unsigned int     notuserfile:1;
 unsigned int     zeroundobuffer:1;


 struct toolboxstr * toolbox;           

 int              timex[2];

	int              compatability;

 int              spare[11];

 filehdr          hdr;


} filestr;


/*****************************************************************************/

#define VRONELAYER  0x10000

#define VRBACK      0
#define VRGRID      5
#define VRPRINT     6
#define VRSELECTION 9
#define VRTOOLS     10
#define VRTOOLS2    11


typedef struct sboxstr
{
 aboxstr      box;
 unsigned int blanked:1;

} sboxstr;



typedef struct viewredrawstr
{
 wimp_redrawstr redrawstr;
 viewstr      * view;
 filestr      * file;

/*                     Variables for redrawing pages                       */

 int xpo;              /* our coord for x page origin                      */
 int ypo;              /* our coord for y page origin                      */

 int xlo;              /* our coord for left of redraw                     */
 int xhi;              /* our coord for right of redraw                    */
 int ylo;              /* our coord for top of redraw                      */ 
 int yhi;              /* our coord for bot of redraw                      */
 int xps;              /* our coord for left of page                       */
 int xpe;              /* our coord for right of page                      */
 int yps;              /* our coord for top of page                        */
 int ype;              /* our coord for bot of page                        */
 int ox;               /* real OS coords of top left hand corner of window */
 int oy;
 int ylos;             /* ylo-yps */
 int yhis;             /* yhi-yps */
 int xlos;             /* xlo-xps */
 int xhis;             /* xhi-xps */

 int ylox; 
 int yhix;             /* limits on redraw in os units  */
 int xlox;
 int xhix;
 int yloz;
 int yhiz;             /* limits on redraw in our units */
 int xloz;
 int xhiz;

 int  page;            /* page that we're redrawing             */
 int  section;         /* section for rpage                     */

 int  ignore;          /* ignore redraws until you see */
 int  object;          /* this object then start       */
 int  xobject;         /* don't redraw this object     */
 int  layer;           /* layer to redraw              */
 int  onelayer;        /* just one layer                           */
 int  isprinter;       /* going to printer             */

 wimp_box pagebox;

 int  pageignore;      /* ignore pages before this one */
 int  sectignore;      /* section for pageignore       */


 int       sobject;    /* doing a redraw of an area of this object */
 int       sboxn;      /* no of areas                              */
 sboxstr * sbox;       /* areas to redraw                          */
 int       szone;      /* just zone marking                        */
 int       sunit;      /* frame/header/footer etc.                 */
 int       ssub;       /* column number                            */
 int       sshortcuts; /* transparent object overlies sobject...   */
 int       sflush;     /* OK to flush any cached areas             */


 void    * prjob;

 transformstr tr;

 int       spare[7];


} viewredrawstr;



typedef struct posnstr
{
 int          mousex;
 int          mousey;
 int          buttons;

 unsigned int shift:1;
 unsigned int ctrl:1;
 unsigned int alt:1;

 viewstr    * view;
 filestr    * file;

 int          x;
 int          y;
 int          xpo;
 int          ypo;
 int          xps;
 int          xpe;
 int          yps;
 int          ype;
 int          ghost;
 int          valid;

 int          section;
 int          page;
 int          object;
 int          otype;
 int          handle;
 int          htype;
 int          master;
 objectbitstr bits;

 int          scrap1;
 int          scrap2;

 int          branch;
 int          chain;

 int          spare[6];

} posnstr;






/*****************************************************************************/

#define TINTPAL   0xC0
#define TINTTRANS 0xA0


typedef struct tintdatastr
{
 unsigned int pal0:9;
 unsigned int tag:8;     // top bit + at least one other
 unsigned int pal1:15;
} tintdatastr;


typedef struct tintstr
{
 unsigned int tint:17;
 unsigned int key:14;
 unsigned int overprint:1;
} tintstr;


typedef struct colourvalue
{
 union
 {
  tintstr     tint;              /* takes precedence    */
  tintdatastr tintdata;          /* local colour values */
 } data;
} colourvalue;


/****************************************************************************/

#define FILLSTR    0

typedef struct fillstr
{

 int fill;


} fillstr;


/*****************************************************************************/

#define PNFNUMERIC    2
#define PNFROMANLOWER 1
#define PNFROMANUPPER 0

/* chapterbits() allocates 4 bits to return the format */


typedef struct pagenumberstr
{
 int          omitheader;
 int          omitfooter;

 int          format;        /* page number format               */
 int          start;         /* 1 == start numbering from here   */

 unsigned int left:1;        /* start on left                    */
 unsigned int blanks:31;     /* count blank pages when numbering */

 int          number;        /* starting number                  */

} pagenumberstr;




typedef struct sectiondatastr
{
 int           width;           /* true values for page */
 int           height;
 char          size[12];
 colourvalue   colour[2];
 fillstr       fill[2];
 int           mainkey[2];

 unsigned int  shape:1;
 unsigned int  usersize:1;
 unsigned int  spare:30;

 int           dble;

 pagenumberstr pagenumber;

 int           number;   /* section number                 */
} sectiondatastr;



typedef struct sectionmapstr
{
 int msource;
 int dsource;
 int mvsource;
 int dvsource;
} sectionmapstr;



typedef struct sectionstr
{                     
 int            key;
 int            type;
 chainstr       chain;

 int            np;               /* number of pages                */
 int            dual;             /* related section number         */
 unsigned int   branch:1;         /* is this in master/doc          */
 unsigned int   changed:1;

 sectiondatastr data;

 int            oldmaster;
 int            flows;            /* chain of unflowed vsources */

 int            spare[8];

 int            mapentries;
 sectionmapstr  map[1];

} sectionstr;


/*****************************************************************************/

#define MAXPAGE   999
#define MAXSECS   99


typedef struct pagedatastr
{
 int      number;

} pagedatastr;



#define PGLINECHAIN (sizeof(pagestr)-sizeof(chainstr))


typedef struct pagestr
{
 int            key;
 int            type;

 chainstr       chain;
 flagstr        flags;
 boxstr         box;

 int            section;
 int            branch; /* a flag to say this page is in master branch */

 pagedatastr    data;
 int            spare[4];

 chainstr       gchain; /* could have many chains here for layers */
} pagestr;




typedef struct
{
 int page;
 int section;

} pagesectionstr;



