/* pxwmf.cpp - WMF and EMF handling */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "transform.h"
#include "sprite.h"
#include "sprx.h"
#include "xprocess.h"


#include "err.h"
#include "filetype.h"
#include "fsx.h"
#include "bf.h"
#include "alloc.h"
#include "colour.h"
#include "etc.h"
#include "poll.h"
#include "xmath.h"
#include "imlib.h"
#include "bmphdr.h"
#include "bmpim.h"
#include "deb.h"
#include "wmflib.h"


#include "constants.h"
#include "str.h"
#include "reslink.h"

#include "obj.h"
#include "object.h"
#include "dscript.h"

#include "xprint.h"
#include "prstr.h"

#include "pxstr.h"
#include "pict.h"
#include "file.h"
#include "view.h"
#include "impex.h"
#include "pixtool.h"
#include "splib.h"
#include "splink.h"
#include "fit.h"
#include "xtrace.h"
#include "pxtags.h"
#include "pxref.h"
#include "pxcache.h"
#include "draw.h"
#include "xrender.h"


#include "pxwmf.h"






/* render picture                */
/* given box in picture space    */
/* given transform to real space */

static os_error * wmfrender(aboxstr * abox,transformstr * tr,
                  processstr * pr,picturestr * picturep,viewredrawstr * v)
{
 os_error        * err;
 int               oldmode;
 pxproxyimstr    * proxy;
 void            * data=NULL;
	transformstr      result;


// cprintf("bmprender");

 err=NULL;

 if(!pr->spotplate || pr->knockoutgraphics)
 {
  oldmode=pr->mode;
  if(pr->spotplate && pr->knockoutgraphics) pr->mode=PROCESSWHITE;


  if(pr->useproxies)
  {
   proxy=(pxproxyimstr*)(picturep->data+((picturep->imagelen+0x3)&(~0x3)));
   data=proxy->data;
  }
  else
  {
   if(picturep->imagelen==0)
   {
    err=pxcache2(&picturep,v->file,NULL,NULL);
   }

   if(!err)
   {
    data=picturep->data;
   }
  }

  if(!err)
  {
//   bmppaint(data,tr->e>>8,tr->f>>8,scale(picturep->abox.x1-picturep->abox.x0,tr->a,256*0x10000),scale(picturep->abox.y1-picturep->abox.y0,tr->d,256*0x10000));
//   cprintf("wmfrender %x",picturep->type);

	  if(v) tr_multiply(&result,tr,&v->tr);
   else  result=*tr;

   if(v && v->prjob && (((prjobstr*)(v->prjob))->bits & PSDIRECT))
			{

			}
			else
			{
    if(picturep->ftype==WMETAF) err=wmfrender(data,&result,(wmfboxstr*)&picturep->abox,pr);
    else                        err=emfrender(data,&result,(wmfboxstr*)&picturep->abox,pr);
			}
  }

  pr->mode=oldmode;
 }

 return(err);
}




static os_error * wmfpxinfo(int object,int ftype,filestr * file,int * info)
{
 os_error         * err;
 picturestr       * picturep;
 char               string[128];
 int                len;
// int              * buff;

 err=NULL;

 picturep=(picturestr*)fob_find(object,file);

 if(picturep->imagelen==0)
 {
  err=pxcache2(&picturep,file,NULL,NULL);
 }

 if(!err)
 {
  *info++=PT_LENGTH;
  *info++=3*sizeof(int);
  *info++=picturep->imagelen;

  strcpy(string,impexname(ftype));
  len=strlen(string)+1;
  len=(len+3)&(~0x3);
  *info++=PT_INFO;
  *info++=2*sizeof(int)+len;
  strcpy((char*)info,string);
  info+=(len/sizeof(int));

  *info++=PT_DIM;
  *info++=4*sizeof(int);
  *info++=picturep->abox.x1-picturep->abox.x0;
  *info++=picturep->abox.y1-picturep->abox.y0;

  *info++=PT_CLASS;
  *info++=8*sizeof(int);
  *info++=PT_VEC;

  *info++=PT_RGB;
  *info++=3;
  *info++=8;
  *info++=0;
  *info++=0;
 
  *info++=PT_END;
  *info++=2*sizeof(int);
 }

 return(err);

 USE(ftype);
}











static os_error * wmfimportpost(int ftype,int object,filestr * file)
{
 USE(ftype);
 USE(object);
 USE(file);

 return(NULL);
}


static os_error * wmfinitpicture(int object,int ftype,filestr * file)
{
 os_error * err;

 picturestr * picturep;

 picturep=(picturestr*)fob_find(object,file);
 picturep->ftype=ftype;

 cprintf("wmf init type %x",ftype);

 if(ftype==WMETAF) err=wmfinit(picturep->data,(wmfboxstr*)&picturep->abox);
 else              err=emfinit(picturep->data,(wmfboxstr*)&picturep->abox);



	return(NULL);
}




static os_error * wmfimport(char * name,int ftype,int xv,
                                                int * object,filestr * file)
{
 os_error   * err;

 err=dxloadchunk(name,0,object,file);
 if(!err)	err=wmfinitpicture(*object,ftype,file);

 return(err);

 USE(xv);
}




static os_error * wmfembed(int ftype,int object,filestr * file)
{
 os_error * err;

 err=dxembedchunk(0,object,file);
 if(!err)
 {
  pxtagsize(object,file,PXPREF1TAG,0);
  pxtagsize(object,file,PXPREF2TAG,0);
 }

 return(err);

	USE(ftype);
}





static os_error * wmfdsexport(int ftype,int object,
                                           filestr * file,buffer * bf)
{
 picturestr * picturep;
 char       * buff;
 int          len;

 picturep=(picturestr*)fob_find(object,file);
 buff=picturep->data;

 len=picturep->imagelen;

 bf_printfeol(bf,"%d",len);

 dxwritedata(NULL,-1,bf);

 dxwritedata(buff,len,bf);

 return(dxwritedata(NULL,-2,bf));

 USE(ftype);
}




static os_error * wmfdsimport(int ftype,int fp,int argc,objectloc * loc)
{              
 os_error          * err;
 filestr           * file;
 dsobjectstr       * ds;
 int                 object;

 file=loc->file;

 err=dxloadblock(0,fp,argc,loc);

 if(argc==DSIMMLAST)
 {
  ds=(dsobjectstr*)(dsstack+fp);
  object=ds->data.integer;

		err=wmfinitpicture(object,ftype,file);
  if(err)
  {
   dxscrapchunk(&object,file);
   ds=(dsobjectstr*)(dsstack+fp);
   ds->data.integer=object;
  }
 }

 return(err);
}





static os_error * wmfexport(char * name,int object,int ftype,int xv,filestr * file)
{
 os_error   * err;
 picturestr * picturep;

 err=NULL;

 picturep=(picturestr*)fob_find(object,file);

 if(picturep->imagelen==0)
 {
  err=pxcache2(&picturep,file,NULL,NULL);
 }

 if(!err) err=bf_saveblock(name,ftype,picturep->data,picturep->imagelen);

 return(err);

 USE(xv);
}



static os_error * wmfinfo(int object,int ftype,filestr * file,int * len)
{
 os_error   * err;
 picturestr * picturep;

 err=NULL;

 picturep=(picturestr*)fob_find(object,file);

 if(picturep->imagelen==0)
 {
  err=pxcache2(&picturep,file,NULL,NULL);
 }

 if(!err) *len=picturep->imagelen;

 return(err);

 USE(ftype);
}









static int wmfshift(picturestr * picturep)
{
 return(0);
 USE(picturep);
}



static os_error * wmfproxy(int ftype,int object,filestr * file,pxproxystr * proxy)
{
 return(spproxysub(SPPXMASK,object,file,proxy));
 USE(ftype);
}






static impexstr wmfstr=
{
 WMETAF,
 "Windows Metafile",

 IMPXLOAD|IMPXPICT,   /* flags       */
 wmfimport,           /* import      */
 wmfimportpost,       /* import post */
 wmfexport,           /* export      */
 wmfrender,           /* render      */
 wmfdsexport,         /* dsexport    */
 wmfdsimport,         /* dsimport    */

 pictintexport,       /* internal export */
 wmfinfo,             /* exportinfo      */

 NULL,                /* cache */

 NULL,                /* rendermem */

 wmfshift,            /* data shift */

 wmfpxinfo,           /* info */
 dxrefload,           /* reference load */
 wmfproxy,            /* make proxy */
 dxremoveproxy,       /* remove proxy */

 wmfembed,            /* embed referenced image */
 dxrefreload,         /* reload                 */
 NULL,                /* list fonts             */
 NULL,                /* x3                     */

};






static impexstr emfstr=
{
 EMETAF,
 "Windows Enhanced Metafile",

 IMPXLOAD|IMPXPICT,   /* flags       */
 wmfimport,           /* import      */
 wmfimportpost,       /* import post */
 wmfexport,           /* export      */
 wmfrender,           /* render      */
 wmfdsexport,         /* dsexport    */
 wmfdsimport,         /* dsimport    */

 pictintexport,       /* internal export */
 wmfinfo,             /* exportinfo      */

 NULL,                /* cache */

 NULL,                /* rendermem */

 wmfshift,            /* data shift */

 wmfpxinfo,           /* info */
 dxrefload,           /* reference load */
 wmfproxy,            /* make proxy */
 dxremoveproxy,       /* remove proxy */

 wmfembed,            /* embed referenced image */
 dxrefreload,         /* reload                 */
 NULL,                /* list fonts             */
 NULL,                /* x3                     */

};






os_error * wmfinit(void)
{
 os_error * err;

          err=impexregister(&wmfstr);
 if(!err) err=impexregister(&emfstr);

 return(err);
}

