/* font.h
 * Copyright (C) Acorn Computers Ltd., 1989
 * Copyright (C) Codemist Ltd., 1989
 */
#ifndef __font_h
#define __font_h
#ifndef __os_h
#include "os.h"
#endif


#ifdef _WIN64

typedef __int64 font;

#else

typedef int font; 

#endif



typedef struct font_metric
{
 int ascent;
 int descent;
 int height;
 int intleading;
 int extleading;
 int maxwidth;
 int averagewidth;

} font_metric;


extern os_error * font_getmetric(font_metric * metric);
extern os_error * font_getmetric2(font_metric * metric,oshdc hdcptr);




#define FD_CHARSET 0
#define FD_REGULAR 400
#define FD_BOLD 700

typedef struct font_description
{
 int       height;
 int       width;
 char      name[64];
 int       charset;
 int       weight;
 int       italic;

} font_description;


extern os_error * font_find_description(font_description * fd,font * handle);
extern os_error * font_getmetric_description(font_description * fd,font_metric * metric);
extern os_error * font_getwidths_description(font_description * fd,int * width);

extern os_error * font_embelish(font oldfont,font * newfont,int italic);


#define FONTENUM_TT 0x1
#define FONTENUM_OT 0x2
#define FONTENUM_T1 0x3


typedef struct font_enumdata
{
 int weight;
 int italic;
 int emsize;
 int emheight;
 int emwidth;
 int charset;
 int pitch;
 int family;
 int type;
	int bold;
	int regular;

} font_enumdata;



typedef os_error * (*fontenumfn)(char * facename,char * fullfontname,font_enumdata * data);

os_error * font_enumerate(font_description * fd,fontenumfn fn);






#define FA_LEFT     0x1
#define FA_RIGHT    0x2
#define FA_CENTRE   0x4
#define FA_TOP      0x8
#define FA_BOTTOM   0x10
#define FA_BASELINE 0x20
#define FA_UPDATE   0x40
#define FA_NOUPDATE 0x80

extern os_error * font_align(int bits);


os_error * font_find(char* name,int xsize, int ysize,int xres, int yres,font * f);
os_error * font_find1(char * name,int xsize,int ysize,int angle,font * fontadd);
os_error * font_find2(char* name,int xsize, int ysize,int angle,font * f);
os_error * font_find3(char * name,int charset,int xsize,int ysize,int angle,font * fontadd);
os_error * font_find4(char * name,int charset,int weight,int italic,int family,int pitch,int xsize,int ysize,int angle,font * fontadd);
os_error * font_getstandardfont(int type,font * handle);


os_error * font_lose(font f);




typedef struct font_string
{
 char * s;
 int    x; 
 int    y; 
 int    split; 
 
 int    term;

 font   font;
 int    xpts;
 int    ypts;
} font_string;



typedef struct font_string16
{
 wchar_t * s;
 int       x; 
 int       y; 
 int       split; 
 
 int       term;

 font      font;
 int       xpts;
 int       ypts;

} font_string16;



os_error * font_strwidth(font_string *fs);
os_error * font_strwidth2(font_string *fs,oshdc hdcptr);

#define font_JUSTIFY  0x01 
#define font_RUBOUT   0x02 
#define font_ABS      0x0 
#define font_AUTOKERN (1<<9)
#define font_RIGHTLEFT (1<<10)
 
#define font_OSCOORDS 0x10 
#define font_TRANS    0x40 

#define font_BACKBLEND (1<<11)

#define font_IGNORE (1<<12)



os_error * font_paint(char*, int options, int x, int y);
os_error * font_paint16(wchar_t *, int options, int x, int y);

os_error * font_paintlen(char*,int len, int options, int x, int y);
os_error * font_painttabbed(char * s,int options,int x,int y,int ntabs,int * tabs,int tx);


os_error * font_setfont(font f);
os_error * font_setfont2(font f,oshdc hdcptr);



os_error * font_setcolour(font f, int background, int foreground, int offset);
os_error * font_setforeground(int foreground);


typedef struct font_info {
 int minx, miny, maxx, maxy;
}font_info;


os_error *font_stringbbox(char * s, font_info *fi);
os_error *font_stringbbox2(char * s,font_info * fi,oshdc hdcptr);

os_error *font_stringbbox16(wchar_t * s,font_info * fi);
os_error *font_stringbbox162(char * s,font_info * fi,oshdc hdcptr);

os_error * font_extent(font_string * fs,int * exts);
os_error * font_extent2(font_string * fs,int * exts,oshdc hdcptr);

typedef struct 
{
 int a;
 int b;
 int c;
 int d;
 int e;
 int f;
} font_transform;


os_error * font_painttrans(char*, int options, int x, int y,
                                                font_transform * trans);


os_error * font_setjustification(int space,int brks);
os_error * font_setjustification2(int space,int brks,oshdc hdcptr);

os_error * font_setextraspace(int space);
os_error * font_setextraspace2(int space,oshdc hdcptr);



typedef struct fontrect
{
 int x0; // left
 int y1; // top
 int x1;
 int y0;

} fontrect;

os_error * font_paintext(char * s,int len,int options,int x,int y,fontrect * rectp,int * exts);
os_error * font_paintexttr(char * s,int len,int options,int x,int y,fontrect * rectp,int * exts,font_transform * tr);
os_error * font_paintextscale(char * s,int len,int options,int x,int y,fontrect * rectp,int xscale,int yscale,int * exts);
os_error * font_paintextscale16(wchar_t * s,int len,int options,int x,int y,fontrect * rectp,int xscale,int yscale,int * exts);
os_error * font_paintext16(wchar_t * s,int len,int options,int x,int y,fontrect * rectp,int * exts);
os_error * font_paintext16tr(wchar_t * s,int len,int options,int x,int y,fontrect * rectp,int * exts,font_transform * tr);

os_error * font_glyphs(oshdc hdcptr,int ** glyphs,int * nglyph);

#define FP_KERNING 0x1
#define FP_EXTENT  0x2
#define FP_NEAREST 0x4

os_error * font_placement(font_string * fs,int * exts,int xflags);
os_error * font_placement16(font_string16 * fs,int * exts,int xflags);
os_error * font_placement2(font_string * fs,int * exts,int xflags,oshdc hdcptr);
os_error * font_placement162(font_string * fs,int * exts,int xflags,oshdc hdcptr);

os_error * font_charbbox2(font f,int ch,int options,font_info * i,oshdc hdcptr);
os_error * font_charbbox(font f,int ch,int options,font_info * i);
os_error * font_charwidth2(int fonth,int c,int * w,int * h,oshdc hdcptr);
os_error * font_charwidth(int fonth,int c,int * w,int * h);

os_error * font_chardata2(int c,font_info * i,int * w,int * h,oshdc hdcptr);
os_error * font_chardata162(int c,font_info * i,int * w,int * h,oshdc hdcptr);
os_error * font_chardata(int c,font_info * i,int * w,int * h);
os_error * font_chardata16(int c,font_info * i,int * w,int * h);


typedef struct font_kerningdatastr
{
 unsigned short first;
 unsigned short second;
 int      shift;
} font_kerningdatastr;

os_error * font_kerningdata2(font_kerningdatastr * data,int * npairs,oshdc hdcptr);
os_error * font_kerningdata(font_kerningdatastr * data,int * npairs);

os_error * font_kerningdata162(font_kerningdatastr * data,int * npairs,oshdc hdcptr);
os_error * font_kerningdata16(font_kerningdatastr * data,int * npairs);

os_error * font_exists(char * facename,int * exists);



os_error * font_glyphindices16(oshdc hdcptr,wchar_t * s,int n,unsigned short int * indexp);
         
extern int font_code16wintab[256];

extern int font_code16win(int win8,int charset);


typedef struct fontsubtablestr
{
	TCHAR * facename;
	TCHAR * subname;

} fontsubtablestr;

os_error * font_subtable(fontsubtablestr	** table,int * n);
os_error * font_subtablefree(fontsubtablestr	** tablep);


#endif
