#include "stdafx.h"



#include "os.h"
#include "sprite.h"
#include "wimp.h"
#include "bbc.h"
#include "transform.h"
#include "xprocess.h"
#include "flex.h"

#include "alloc.h"
#include "etc.h"
#include "deb.h"

#include "bitrev.h"

#include "colour.h"
#include "sprx.h"
#include "xmath.h"


#include "oserr.h"
#include "osgraphics.h"
#include "ossprite.h"
#include "osglue.h"



char ln2bpptab[33]=
{
 0,     // 0
 0,     // 1
 1,     // 2
 0,     // 3
 2,     // 4
 0,     // 5
 0,     // 6
 0,     // 7
 3,     // 8
 0,     // 9
 0,     // 10
 0,     // 11
 0,     // 12
 0,     // 13
 0,     // 14
 0,     // 15
 4,     // 16
 0,     // 17
 0,     // 18
 0,     // 19
 0,     // 20
 0,     // 21
 0,     // 22
 0,     // 23
 0,     // 24
 0,     // 25
 0,     // 26
 0,     // 27
 0,     // 28
 0,     // 29
 0,     // 30
 0,     // 31
 5,     // 32
};



static int xln2(int x)
{
 int n;
 n=0;
 while(x>1) {x=x>>1;n++;}
 return(n);
}


int ln2(int n)
{
 if(n<=32) return(ln2bpptab[n]);
 return(xln2(n));
}



typedef struct bmphdrstr
{
 BITMAPINFOHEADER bmi;
 RGBQUAD          pal[256];


} bmphdrstr;





static void osgetpal(int spbpp,newsprite_header * sph,RGBQUAD * quad,int * blackindex,int * pixtrans)
{
 os_error * err;
 int        palentries;
 int      * paltab;
 int      * pal2;
 int        i;
 int        c1;

 err=NULL;

 if(spbpp>8) return;

 paltab=(int*)quad;
 palentries=(sph->image-sizeof(sprite_header))/8;
 if(palentries<0) palentries=0;
 pal2=(int*)(((char*)sph)+sizeof(sprite_header));

//	cprintf("spbpp %d palentries %d",spbpp,palentries);

//		cprintf("gen pal %d pix %x",palentries,pixtrans);

 if(pixtrans)
 {
  palentries=(1<<spbpp);

  for(i=0;i<palentries;i++)
  {
   paltab[i]=QUADFROMPAL(pixtrans[i]);
  }
 }
 else
 if(palentries==0)
 {
//		cprintf("gen pal");
  /* if there is no palette, we'd like it to look like wimp cols */
  genpalbpp(paltab,spbpp,&palentries);

//		cprintf("gen pal %d",palentries);

  for(i=0;i<palentries;i++)
  {
   paltab[i]=QUADFROMPAL(paltab[i]);
  }
 }
 else
 if(spbpp==8 && (palentries==64 || palentries==16))
 {
  if(palentries==64)
  {
   pal2=pal2+48*2;
  }

  for(i=255;i>=0;i--)
  {
   c1=pal2[(i & 0xF)*2];

   if(i & 0x10) c1|=0x8000;
   else         c1&=~0x8000;

   if(i & 0x20) c1|=0x400000;
   else         c1&=~0x400000;

   if(i & 0x40) c1|=0x800000;
   else         c1&=~0x800000;

   if(i & 0x80) c1|=0x80000000;
   else         c1&=~0x80000000;

   c1&=0xF0F0F000;
   c1|=(c1>>4)&0xF0F0F00;

   paltab[i]=QUADFROMPAL(c1);
  }

  palentries=256;
 }
 else
 {
  for(i=0;i<palentries;i++)
  {
   paltab[i]=QUADFROMPAL(pal2[i*2]);

//			cprintf("i=%d pal=%x",i,paltab[i]);
  }
 }

 if(blackindex)
 {
  *blackindex=0;
  for(i=0;i<palentries;i++)
  {
   if(paltab[i]==0)
   {
    *blackindex=i;
    break;
   }
  }
 }
}



static void osgreypal(int spbpp,newsprite_header * sph,RGBQUAD * quad,int * whiteindex,int * pixtrans)
{
 os_error * err;
 int        palentries;
	int        i;
 int      * paltab;


 err=NULL;

 if(spbpp>8) return;

 osgetpal(spbpp,sph,quad,NULL,pixtrans);

 palentries=(1<<spbpp);
 paltab=(int*)quad;

 greyscalequad(paltab,palentries);

 if(whiteindex)
 {
  *whiteindex=0;
  for(i=0;i<palentries;i++)
  {
   if(paltab[i]==0xFFFFFF)
   {
    *whiteindex=i;
    break;
   }
  }
 }
}





static void osgetmpal(int bpp,RGBQUAD * quad)
{
 int * paltab;

 paltab=(int*)quad;

 paltab[0]=QUADFROMPAL(0xFFFFFF00);
 paltab[(1<<bpp)-1]=QUADFROMPAL(0);
}



static void osgetmgpal(int bpp,RGBQUAD * quad)
{
 int * paltab;

 paltab=(int*)quad;

 paltab[0]=QUADFROMPAL(0);
 paltab[(1<<bpp)-1]=QUADFROMPAL(0xFFFFFF00);
}





static void setspritebits(int * bits,int spbpp,int dbpp,int xpix,int ypix,int * dbits,int sshift)
{
 char       * pr;
 char       * pw;
 char       * plt;
 int          bytewidth;
 int          dbytewidth;
 char       * plt2;
 char       * plt3;
 short      * pws;
 short      * pwlt;
 short      * pwlt2;
 int        * prw;
 int        * pww;
 int        * pltw;
 int        * pltw2;
 int          c;


 bytewidth=((xpix*spbpp+31)&~31)/8;

 if(spbpp==1)
 {
  plt=(char*)dbits;
  pw=plt+ypix*bytewidth;
  pr=(char*)bits;
  while(pw>plt)
  {
   plt2=pw;
   pw-=bytewidth;
   while(pw<plt2)
   {
    *pw++=bitrev[*pr++];
   }
   pw-=bytewidth;
  }
 }
 else
 if(spbpp==2)
 {
  dbytewidth=((xpix*dbpp+31)&~31)/8;
  pr=(char*)bits;
  plt=pr+bytewidth;
  pwlt=(short*)dbits;
  pws=(short*)(((char*)pwlt)+ypix*dbytewidth);

  while(pws>pwlt)
  {
   pwlt2=pws;
   pws=(short*)(((char*)pws)-dbytewidth);
   while(pws<pwlt2)
   {
    *pws++=dbitexp[*pr++];
   }
   pws=(short*)(((char*)pws)-dbytewidth);
   pr=plt;
   plt+=bytewidth;
  }
 }
 else
 if(spbpp==4)
 {
  pr=(char*)bits;
  plt=(char*)dbits;
  pw=plt+ypix*bytewidth;
  while(pw>plt)
  {
   plt2=pw;
   pw-=bytewidth;
   while(pw<plt2)
   {
    *pw++=nybrev[*pr++];
   }
   pw-=bytewidth;
  }
 }
 else
 if(spbpp==8)
 {
  pr=(char*)bits;
  plt=(char*)dbits;
  pw=plt+ypix*bytewidth;
  while(pw>plt)
  {
   plt2=pw;
   pw-=bytewidth;
   while(pw<plt2)
   {
    *pw++=*pr++;
   }
   pw-=bytewidth;
  }
 }
 else
 if(spbpp==16)
 {
  prw=bits;
  pltw=dbits;
  pww=(int*)(((char*)pltw )+ypix*bytewidth);
  while(pww>pltw)
  {
   pltw2=pww;
   pww=(int*)(((char*)pww)-bytewidth);
   while(pww<pltw2)
   {
    c=*prw++;
    *pww++=((c & 0x1F)<<10)|((c & 0x3E0))|((c & 0x7C00)>>10)|
            ((c & 0x1F0000)<<10)|((c & 0x3E00000))|((c & 0x7C000000)>>10);
   }
   pww=(int*)(((char*)pww)-bytewidth);
  }
 }
 else
 if(spbpp==32)
 {
  dbytewidth=((xpix*dbpp+31)&~31)/8;
  prw=bits;
  plt=(char*)dbits;
  pw=plt+ypix*dbytewidth;

  while(pw>plt)
  {
   plt2=pw;
   pw-=dbytewidth;
   plt3=pw+3*xpix;

   while(pw<plt3)
   {
    c=*prw++;
    *pw++=(char)((c & 0xFF0000)>>16);
    *pw++=(char)((c & 0xFF00)>>8);
    *pw++=(char)((c & 0xFF));
   }

   pw=plt2-dbytewidth;
  }
 }

	USE(sshift);
}




static void setspritembits(int * bits,int spbpp,int * mbits,int dbpp,int xpix,int ypix,int * dbits,int backindex,int sshift)
{
 char       * pr;
 char       * pw;
 char       * plt;
 char       * pm;
 char       * pmlt;
 int          bytewidth;
 int          c;
 int          backcol;
 int        * prw;
 int        * pww;
 int        * pmw;
 char       * plt2;
 short      * pws;
 short      * pwlt;
 short      * pwlt2;
 int          dbytewidth;
 int        * pltw;
 int        * pltw2;


 bytewidth=((xpix*spbpp+31)&~31)/8;


 if(spbpp==1)
 {
  backcol=backindex|(backindex<<1);
  backcol=backcol|(backcol<<2);
  backcol=backcol|(backcol<<4);

  plt=(char*)dbits;
  pw=plt+ypix*bytewidth;
  pr=(char*)bits;
  pm=(char*)mbits;

  while(pw>plt)
  {
   plt2=pw;
   pw-=bytewidth;
   while(pw<plt2)
   {
    c=*pr++;
    c^=backcol;
    c&=*pm++;
    c^=backcol;
    *pw++=bitrev[c];
   }
   pw-=bytewidth;
  }
 }
 else
 if(spbpp==2)
 {
  backcol=backindex|(backindex<<4);
  dbytewidth=((xpix*dbpp+31)&~31)/8;
  pr=(char*)bits;
  pm=(char*)mbits;
  plt=pr+bytewidth;
  pmlt=pm+bytewidth;
  pwlt=(short*)dbits;
  pws=(short*)(((char*)pwlt)+ypix*dbytewidth);

  while(pws>pwlt)
  {
   pwlt2=pws;
   pws=(short*)(((char*)pws)-dbytewidth);
   while(pws<pwlt2)
   {
    c=*pr++;
    c^=backcol;
    c&=*pm++;
    c^=backcol;
    *pws++=dbitexp[c];
   }
   pws=(short*)(((char*)pws)-dbytewidth);

   pr=plt;
   pm=pmlt;
   plt+=bytewidth;
   pmlt+=bytewidth;
  }
 }
 else
 if(spbpp==4)
 {
  backcol=backindex|(backindex<<4); // backindex is the index in the sprite which maps to black, i.e. 0 colour number
  pr=(char*)bits;
  plt=(char*)dbits;
  pm=(char*)mbits;
  pw=plt+ypix*bytewidth;

  while(pw>plt)
  {
   plt2=pw;
   pw-=bytewidth;
   while(pw<plt2)
   {
    c=*pr++;
    c^=backcol;
    c&=*pm++;
    c^=backcol;
    *pw++=nybrev[c];
   }
   pw-=bytewidth;
  }
 }
 else
 if(spbpp==8)
 {
  backcol=backindex|(backindex<<8);
  backcol=backcol|(backcol<<16);

  prw=(int*)bits;
  pltw=(int*)dbits;
  pmw=(int*)mbits;
  pww=(int*)(((char*)pltw)+ypix*bytewidth);

  while(pww>pltw)
  {
   pltw2=pww;
   pww=(int*)(((char*)pww)-bytewidth);
   while(pww<pltw2)
   {
    c=*prw++;
    c^=backcol;
    c&=*pmw++;
    c^=backcol;
    *pww++=c;
   }
   pww=(int*)(((char*)pww)-bytewidth);
  }
 }
 if(spbpp==16)
 {
  // never
 
 }
 else
 if(spbpp==32)
 {
  // never
 
 }

	USE(sshift);
}




static void setspritembits1(int * sbits,int sbpp,int * mbits,int dbpp,int xpix,int ypix,int * dbits,int backindex)
{
 char       * pr;
 char       * prrow;
 char       * pw;
 char       * pwrow;
 char       * pwlt;
 char       * pwrowlt;

 int        * prroww;
 int        * pmroww;
 int        * pwroww;
 int        * pwltw;
 int        * pwrowltw;

 int          shiftr;
 int          shiftw;
 int          shiftm;
 int          shiftltw;
 int          mask;
 int          cr;
 int          cr2;
 int          cw;
 int          cm;
 
 int          xdisc;
 int          backcol;
 int          dbytewidth;
 int          mbytewidth;
 int          sbytewidth;
 int        * prw;
 int        * pww;
 int        * pmw;
 int          incx;
 int          inc1x;
 int          incy;
 int          sshift=0;


 incx=sbpp;
 inc1x=1;

 dbytewidth=((xpix*dbpp+31)&~31)/8;
 sbytewidth=((xpix*sbpp+31)&~31)/8;
 mbytewidth=((xpix*1+31)&~31)/8;


 incy=-dbytewidth;
 pwltw=(int*)dbits;
 pwroww=(int*)(((char*)pwltw)+(ypix-1)*dbytewidth);
 prroww=(int*)sbits;
 pmroww=(int*)mbits;



 if(sbpp==1)
 {
// can never get here
 
 }
 if(sbpp==2)
 {
  backcol=backindex;

  while(1)
  {
   if(pwroww<pwltw) break;

   pww=pwroww;
   pwrowltw=pwroww+xpix/8;
   shiftltw=(xpix & 7)*4;
   if(shiftltw & 4) shiftltw-=4;
   else             shiftltw+=4;


   shiftw=4;
   cw=0;

   prw=prroww;
   pmw=pmroww;
   cr=*prw++;
   cm=*pmw++;
   shiftr=sshift;
   shiftm=0;


   if(((cm>>shiftm)&0x1)==0) cr2=backcol;
   else                      cr2=(cr>>shiftr)&0x3;

   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    cw|=(cr2<<shiftw);

    if(shiftw & 4) shiftw-=4;
    else           shiftw+=12;

    if(shiftw>=32)
    {
     shiftw=4;
     *pww++=cw;
     cw=0;
    }
    
    shiftr+=incx;
    if(shiftr>=32)
    {
     cr=*prw++;
     shiftr=0;
    }

    shiftm+=inc1x;
    if(shiftm>=32)
    {
     cm=*pmw++;
     shiftm=0;
    }

    if(((cm>>shiftm)&0x1)==0) cr2=backcol;
    else                      cr2=(cr>>shiftr)&0x3;
   }

   if(shiftw!=4) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   prroww=(int*)(((char*)prroww)+sbytewidth);
   pmroww=(int*)(((char*)pmroww)+mbytewidth);
  }
 }
 if(sbpp==4)
 {
  mask=15;
  backcol=backindex;


  while(1)
  {
   if(pwroww<pwltw) break;
  
   pww=pwroww;
   pwrowltw=pwroww+xpix/8;
   shiftltw=(xpix & 7)*4;
   if(shiftltw & 4) shiftltw-=4;
   else             shiftltw+=4;
   xdisc=0;
   shiftw=4;
   cw=0;

   prw=prroww;
   pmw=pmroww;
   shiftr=sshift;
   shiftm=0;
   cr=*prw++;
   cm=*pmw++;

   if(((cm>>shiftm)&0x1)==0) cr2=backcol;
   else                      cr2=(cr>>shiftr)&0xF;

   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    cw|=(cr2<<shiftw);
    
    if(shiftw & 4) shiftw-=4;
    else           shiftw+=12;

    if(shiftw>=32)
    {
     shiftw=4;
     *pww++=cw;
     cw=0;
    }
    
    shiftr+=incx;
    if(shiftr>=32)
    {
     cr=*prw++;
     shiftr=0;
    }

    shiftm+=inc1x;
    if(shiftm>=32)
    {
     cm=*pmw++;
     shiftm=0;
    }

    if(((cm>>shiftm)&0x1)==0) cr2=backcol;
    else                      cr2=(cr>>shiftr)&0xF;
   }

   if(shiftw!=4) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   prroww=(int*)(((char*)prroww)+sbytewidth);
   pmroww=(int*)(((char*)pmroww)+mbytewidth);
  }
 }
 else
 if(sbpp==8)
 {
  prrow=(char*)prroww;
  pwrow=(char*)pwroww;
  pwlt=(char*)pwltw;

  backcol=backindex;

  while(1)
  {
   if(pwrow<pwlt) break;
   
   
   pw=pwrow;
   pwrowlt=pwrow+xpix;
   xdisc=0;

   pr=prrow+(sshift/8);
   pmw=pmroww+(sshift/8)/32;
   cm=*pmw++;
   shiftm=(sshift/8);
   cr=*pr++;

   if(((cm>>shiftm)&0x1)==0) cr=backcol;

   while(pw<pwrowlt)
   {
    *pw++=(char)cr;
    
    cr=*pr++;

    shiftm+=inc1x;
    if(shiftm>=32)
    {
     cm=*pmw++;
     shiftm=0;
    }

    if(((cm>>shiftm)&0x1)==0) cr=backcol;
   }

   pwrow+=incy;

   prrow+=sbytewidth;
   pmroww=(int*)(((char*)pmroww)+mbytewidth);
  }
 }
 else
 if(sbpp==16)
 {
  while(1)
  {
   if(pwroww<pwltw) break;
  
   
   pww=pwroww;
   pwrowltw=pwroww+xpix/2;
   shiftltw=(xpix & 1)*16;

   xdisc=0;
   shiftw=0;
   cw=0;

   prw=prroww;
   pmw=pmroww;
   shiftr=0;
   shiftm=0;
   cr=*prw++;
   cm=*pmw++;

   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(((cm>>shiftm)&0x1))
    {
     if(shiftw==0)
     {
      if(shiftr==0)  cw=((cr & 0x1F)<<10)|((cr & 0x3E0))|((cr & 0x7C00)>>10);
      else           cw=((cr & 0x1F0000)>>6)|((cr & 0x3E00000)>>16)|((cr & 0x7C000000)>>26);
     }
     else
     {
      if(shiftr==0) cw|=((cr & 0x1F)<<26)|((cr & 0x3E0)<<16)|((cr & 0x7C00)<<6);
      else          cw|=((cr & 0x1F0000)<<10)|((cr & 0x3E00000))|((cr & 0x7C000000)>>10);
     }
    }
    else
    {
     if(shiftw==0) cw=0;
    }


    shiftw+=16;

    if(shiftw>=32)
    {
     shiftw=0;
     *pww++=cw;
    }
    
    shiftr+=incx;
    if(shiftr>=32)
    {
     cr=*prw++;
     shiftr=0;
    }

    shiftm+=inc1x;
    if(shiftm>=32)
    {
     cm=*pmw++;
     shiftm=0;
    }
   }

   if(shiftw!=0) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   prroww=(int*)(((char*)prroww)+sbytewidth);
   pmroww=(int*)(((char*)pmroww)+mbytewidth);
  }
 }
 else
 if(sbpp==32)
 {
  while(1)
  {
   if(pwroww<pwltw) break;
   
   pw=(char*)pwroww;
   pwrowlt=pw+3*xpix;
   cw=0;

   shiftm=0;
   prw=prroww;
   pmw=pmroww;
   cr=*prw++;
   cm=*pmw++;

   while(pw<pwrowlt)
   {
    if((cm>>shiftm)&0x1)
    {
     *pw++=(char)((cr & 0xFF0000)>>16);
     *pw++=(char)((cr & 0xFF00)>>8);
     *pw++=(char)((cr & 0xFF));
    }
    else
    {
     *pw++=0;
     *pw++=0;
     *pw++=0;
    }

    cr=*prw++;

    shiftm+=inc1x;
    if(shiftm>=32)
    {
     cm=*pmw++;
     shiftm=0;
    }
   }


   pwroww=(int*)(((char*)pwroww)+incy);

   prroww=(int*)(((char*)prroww)+sbytewidth);
   pmroww=(int*)(((char*)pmroww)+mbytewidth);
  }
 }
}





os_error * sprite_render2(newsprite_header * sph,HDC hdc,int x,int y,int grey)
{
 os_error   * err;
 int          xpix;
 int          ypix;
 int        * bits;
 int        * mbits;
 bmphdrstr    bmi;
 int          bitwidth;
 int          spbpp;
 int          mbpp;
 int        * dbits;
 HBITMAP      hbm;
// BITMAP       bitmap;
 HDC          hdcmem;
 int          dbpp;
 int          backindex;
 int          sshift;

 err=NULL;


// cprintf("sprite_render %d",grey);

 getspriteinfo((newsprite_header*)sph,NULL,NULL,&spbpp);

 bits=(int*)(((char*)sph)+sph->image);
 mbits=(int*)(((char*)sph)+sph->mask);
 bitwidth=32*(sph->width+1)-sph->lbit-(31-sph->rbit);
 sshift=sph->lbit;

 xpix=bitwidth/spbpp;
 ypix=sph->height+1;

 if(spbpp==2)  dbpp=4;
 else
 if(spbpp==32) dbpp=24;
 else          dbpp=spbpp;


// SetDIBits(hdc,hbm,0,ypix,bits,(BITMAPINFO*)&bmi,DIB_RGB_COLORS);

 if(mbits==bits || (((newsprite_header*)sph)->selector.type.alpha))   // the no mask case
 {
  memset(&bmi,0,sizeof(bmi));

//		if(grey) cprintf("grey");

		if(grey)	osgreypal(spbpp,(newsprite_header *)sph,&bmi.pal[0],NULL,NULL);
  else     osgetpal(spbpp,(newsprite_header *)sph,&bmi.pal[0],NULL,NULL);

  bmi.bmi.biWidth=xpix;
  bmi.bmi.biHeight=ypix;
  bmi.bmi.biXPelsPerMeter=0;
  bmi.bmi.biYPelsPerMeter=0;
  bmi.bmi.biBitCount=(unsigned short)dbpp;
  bmi.bmi.biPlanes=1;
  bmi.bmi.biCompression=BI_RGB;
  bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
  bmi.bmi.biSizeImage=0;
  bmi.bmi.biClrImportant=0;
  if(spbpp==2) bmi.bmi.biClrUsed=(1<<spbpp);
  else         bmi.bmi.biClrUsed=0;

  err=XCreateDIBSection(hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
  if(!err)
  {
//  cprintf("setspbits s %d d %d",spbpp,dbpp);

   setspritebits(bits,spbpp,dbpp,xpix,ypix,dbits,sshift);

   SetStretchBltMode(hdc,COLORONCOLOR);

//   GetObject(hbm,sizeof(BITMAP),&bitmap);

   hdcmem=CreateCompatibleDC(hdc);
   SelectObject(hdcmem,hbm);
   BitBlt(hdc,x,y,xpix,ypix,hdcmem,0,0,SRCCOPY);
   DeleteDC(hdcmem);
 
   DeleteObject(hbm);
  }
 }
 else            // we have a mask
 {
  if(((newsprite_header*)sph)->selector.type.type)
  {
   mbpp=dbpp=1;
  }
  else
  {
   mbpp=spbpp;
   if(spbpp==2) dbpp=4;
   else         dbpp=spbpp;
  }

  if(((newsprite_header*)sph)->selector.type.alpha) mbpp=8;


  if(grey)  osgetmgpal(mbpp,&bmi.pal[0]);
  else      osgetmpal(mbpp,&bmi.pal[0]);

  bmi.bmi.biWidth=xpix;
  bmi.bmi.biHeight=ypix;
  bmi.bmi.biXPelsPerMeter=0;
  bmi.bmi.biYPelsPerMeter=0;
  bmi.bmi.biBitCount=(unsigned short)dbpp;
  bmi.bmi.biPlanes=1;
  bmi.bmi.biCompression=BI_RGB;
  bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
  bmi.bmi.biSizeImage=0;
  bmi.bmi.biClrImportant=0;
  if(mbpp==2) bmi.bmi.biClrUsed=(1<<mbpp);
  else        bmi.bmi.biClrUsed=0;

  err=XCreateDIBSection(hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
  if(!err)
  {
   setspritebits(mbits,mbpp,dbpp,xpix,ypix,dbits,sshift);

//   GetObject(hbm,sizeof(BITMAP),&bitmap);

   hdcmem=CreateCompatibleDC(hdc);
   SelectObject(hdcmem,hbm);
   BitBlt(hdc,x,y,xpix,ypix,hdcmem,0,0,grey?SRCPAINT:SRCAND);  // and mask
   DeleteDC(hdcmem);
   DeleteObject(hbm);

   if(spbpp==2)  dbpp=4;
   else
   if(spbpp==32) dbpp=24;
   else          dbpp=spbpp;

 		if(grey)	osgreypal(dbpp,(newsprite_header *)sph,&bmi.pal[0],&backindex,NULL);
   else     osgetpal(spbpp,(newsprite_header *)sph,&bmi.pal[0],&backindex,NULL);

   bmi.bmi.biWidth=xpix;
   bmi.bmi.biHeight=ypix;
   bmi.bmi.biXPelsPerMeter=0;
   bmi.bmi.biYPelsPerMeter=0;
   bmi.bmi.biBitCount=(unsigned short)dbpp;
   bmi.bmi.biPlanes=1;
   bmi.bmi.biCompression=BI_RGB;
   bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
   bmi.bmi.biSizeImage=0;
   bmi.bmi.biClrImportant=0;
   if(spbpp==2) bmi.bmi.biClrUsed=(1<<spbpp);
   else         bmi.bmi.biClrUsed=0;

   err=XCreateDIBSection(hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
   if(!err)
   {
    if(spbpp==mbpp) setspritembits(bits,spbpp,mbits,dbpp,xpix,ypix,dbits,backindex,sshift);
    else            setspritembits1(bits,spbpp,mbits,dbpp,xpix,ypix,dbits,backindex);

//    GetObject(hbm,sizeof(BITMAP),&bitmap);

    hdcmem=CreateCompatibleDC(hdc);
    SelectObject(hdcmem,hbm);
    BitBlt(hdc,x,y,xpix,ypix,hdcmem,0,0,grey?SRCAND:SRCPAINT); // or data
    DeleteDC(hdcmem);
    DeleteObject(hbm);
   }
  }
 }

// SetDIBitsToDevice(hdc,x,y,xpix,ypix,0,0,0,ypix,bits,(BITMAPINFO*)&bmi,DIB_RGB_COLORS);


 return(err);
}



os_error * sprite_render(newsprite_header * sph,HDC hdc,int x,int y)
{
	return(sprite_render2(sph,hdc,x,y,0));
}



/****************************************************************************/

// routines for scaling sprites


#define SPWORKCHUNK 0x400

static char * spworkspace;


static void setspritebitsscaled(int * sbits,int sbpp,int sbytewidth,int sshift,int dbpp,int sxpix,int sypix,int dxpix,int dypix,int * dbits,int sx,int ex,int sy,int ey,processstr * prc,newsprite_header * sph)
{
	os_error   * err;
 char       * pr;
 char       * prrow;
 char       * pw;
 char       * pwrow;
 char       * pwlt;
 char       * pwrowlt;
 int        * prroww;
 int        * pwroww;
 int        * pwltw;
 int        * pwrowltw;
 int          shiftr;
 int          shiftw;
 int          shiftltw;
 int          mask;
 int          cr=0;
 int          cw;
 int          xdisc;
 int          ydisc;
 int          dbytewidth;
 int        * prw;
 int        * pww;
 int          incx;
 int          inc1x;
 int          incy;
 int          xdiscp;
 int          nx;
 int          ny;
 int          crf;


//  cprintf("setspritebitsscaled sx %d ex %d sy %d ey %d dx %d dy %d sshift %d",sx,ex,sy,ey,dxpix,dypix,sshift);

	if(sph->selector.type.type==SPT_CMYK)
	{
		err=schunk((flex_ptr)&spworkspace,(ex-sx)*4,SPWORKCHUNK);
	}
 
 if(dxpix<0)
 {
  incx=-sbpp;
  inc1x=-1;
  dxpix=-dxpix;
 }
 else
 {
  incx=sbpp;
  inc1x=1;
 }

 dbytewidth=(((ex-sx)*dbpp+31)&~31)/8;

 if(dypix<0)
 {
  incy=-dbytewidth;
  dypix=-dypix;
  pwltw=(int*)dbits;
  pwroww=(int*)(((char*)pwltw)+((ey-sy)-1)*dbytewidth);
 }
 else
 {
  incy=dbytewidth;
  pwroww=(int*)dbits;
  pwltw=(int*)(((char*)pwroww)+(ey-sy)*dbytewidth);
 }



 xdiscp=sx*sxpix;
 nx=(xdiscp/dxpix);
 xdiscp-=nx*dxpix;

// cprintf("incy=%d dbytewidth %d",incy,dbytewidth);

// an upside down image has incy>=0

 if(incy>=0)
 {
  ydisc=(dypix-ey)*sypix;
 }
 else
 {
  ydisc=sy*sypix;
 }

 ny=(ydisc/dypix);
 ydisc-=ny*dypix;


 sshift/=sbpp;


 prroww=(int*)(((char*)sbits)+ny*sbytewidth);


 if(sbpp==1)
 {
  mask=1;

  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }
   
   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/32;
   shiftltw=((ex-sx) & 31)*1;
   shiftltw=(7-(shiftltw & 7))+(shiftltw & 24);
   xdisc=xdiscp;
   shiftw=7;
   cw=0;

   crf=0;

   if(incx<0)
   {
    prw=prroww+(sshift+sxpix-1-nx)/32;
    shiftr=((sshift+sxpix-1-nx) & 0x1F);
   }
   else
   {
    prw=prroww+((nx+sshift)/32);
    shiftr=((nx+sshift) & 0x1F);
   }

   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     crf=1;
    }
    
    cw|=((cr>>shiftr)&mask)<<shiftw;
    if(shiftw & 7) shiftw--;
    else           shiftw+=15;

    if(shiftw>=32)
    {
     shiftw=7;
     *pww++=cw;
     cw=0;
    }
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     shiftr+=incx;
     if(shiftr>=32)
     {
      crf=0;
      prw++;
      shiftr=0;
     }
     else
     if(shiftr<0)
     {
      crf=0;
      prw--;
      shiftr=31;
     }
    }
   }

   if(shiftw!=7) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
   } 
  }
 }
 else
 if(sbpp==2)
 {
  mask=3;

  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }
   
   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/8;
   shiftltw=((ex-sx) & 7)*4;
   if(shiftltw & 4) shiftltw-=4;
   else             shiftltw+=4;
   xdisc=xdiscp;

   shiftw=4;
   cw=0;

   crf=0;

   if(incx<0)
   {
    prw=prroww+(sshift+sxpix-1-nx)/16;
    shiftr=((sshift+sxpix-1-nx) & 0xF)*2;
   }
   else
   {
    prw=prroww+((nx+sshift)/16);
    shiftr=((nx+sshift) & 0xF)*2;
   }

   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     crf=1;
    }
    
    cw|=((cr>>shiftr)&mask)<<shiftw;
    if(shiftw & 4) shiftw-=4;
    else           shiftw+=12;

    if(shiftw>=32)
    {
     shiftw=4;
     *pww++=cw;
     cw=0;
    }
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     shiftr+=incx;
     if(shiftr>=32)
     {
      crf=0;
      prw++;
      shiftr=0;
     }
     else
     if(shiftr<0)
     {
      crf=0;
      prw--;
      shiftr=30;
     }
    }
   }

   if(shiftw!=4) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
   } 
  }
 }
 else
 if(sbpp==4)
 {
  mask=15;

  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/8;
   shiftltw=((ex-sx) & 7)*4;
   if(shiftltw & 4) shiftltw-=4;
   else             shiftltw+=4;
   xdisc=xdiscp;

   crf=0;

   if(incx<0)
   {
    prw=prroww+(sshift+sxpix-1-nx)/8;
    shiftr=((sshift+sxpix-1-nx) & 0x7)*4;
   }
   else
   {
    prw=prroww+((nx+sshift)/8);
    shiftr=((nx+sshift) & 0x7)*4;
   }

   shiftw=4;
   cw=0;

   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     crf=1;
    }
    
    cw|=((cr>>shiftr)&mask)<<shiftw;
    if(shiftw & 4) shiftw-=4;
    else           shiftw+=12;

    if(shiftw>=32)
    {
     shiftw=4;
     *pww++=cw;
     cw=0;
    }
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     shiftr+=incx;
     if(shiftr>=32)
     {
      crf=0;
      prw++;
      shiftr=0;
     }
     else
     if(shiftr<0)
     {
      crf=0;
      prw--;
      shiftr=28;
     }
    }
   }

   if(shiftw!=4) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
   } 
  }
 }
 else
 if(sbpp==8)
 {
  prrow=(char*)prroww;
  pwrow=(char*)pwroww;
  pwlt=(char*)pwltw;


  while(1)
  {
   if(incy>0)
   {
    if(pwrow>=pwlt) break;
   }
   else
   {
    if(pwrow<pwlt) break;
   }

   
   if(incx<0) pr=prrow+sshift+sxpix-1-nx;
   else       pr=prrow+sshift+nx;
   pw=pwrow;
   pwrowlt=pwrow+(ex-sx);
   xdisc=xdiscp;

   while(pw<pwrowlt)
   {
    *pw++=*pr;
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     pr+=inc1x;
    }
   }

   pwrow+=incy;

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prrow+=sbytewidth;
   } 
  }
 }
 else
 if(sbpp==16)
 {
  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

   crf=0;
   
   if(incx<0)
   {
    prw=prroww+(sxpix-1-nx)/2;
    shiftr=((sxpix-1-nx) & 0x1)*16;
   }
   else
   {
    prw=prroww+(nx/2);
    shiftr=(nx & 0x1)*16;
   }

   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/2;
   shiftltw=((ex-sx) & 1)*16;

   xdisc=xdiscp;
   shiftw=0;
   cw=0;


   if(prc && prc->mode==PROCESSBLACK) memset(pww,0,(ex-sx)*2);
   else
   if(prc && prc->mode==PROCESSWHITE) memset(pww,0xFF,(ex-sx)*2);
   else
   {
	   while(pww<pwrowltw || shiftw!=shiftltw)
    {
     if(crf==0)
     {
      cr=*prw;
      crf=1;
     }
    
     if(shiftw==0)
     {
      if(shiftr==0)  cw=((cr & 0x1F)<<10)|((cr & 0x3E0))|((cr & 0x7C00)>>10);
      else           cw=((cr & 0x1F0000)>>6)|((cr & 0x3E00000)>>16)|((cr & 0x7C000000)>>26);
     }
     else
     {
      if(shiftr==0) cw|=((cr & 0x1F)<<26)|((cr & 0x3E0)<<16)|((cr & 0x7C00)<<6);
      else          cw|=((cr & 0x1F0000)<<10)|((cr & 0x3E00000))|((cr & 0x7C000000)>>10);
     }

     shiftw+=16;

     if(shiftw>=32)
     {
      shiftw=0;
      *pww++=cw;
      cw=0;
     }
    
     xdisc+=sxpix;
     while(xdisc>=dxpix)
     {
      xdisc-=dxpix;
      shiftr+=incx;
      if(shiftr>=32)
      {
       crf=0;
       prw++;
       shiftr=0;
      }
      else
      if(shiftr<0)
      {
       crf=0;
       prw--;
       shiftr=16;
      }
     }
    }

    if(shiftw!=0) *pww++=cw;

    if(prc && prc->mode==PROCESSFN)
    {
     prc->process(prc->handle,pwroww,pwroww,(ex-sx),PROCESST_BGR16,PROCESST_BGR16,NULL,NULL);
    }
			}

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
   } 
  }
 }
 else
 if(sbpp==32)
 {
  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

   crf=0;

   if(incx>0)
   {
    prw=prroww+nx;
   }
   else
   {
    prw=prroww+sxpix-1-nx;
   }



   xdisc=xdiscp;

   if(prc && prc->mode==PROCESSBLACK) memset(pwroww,0,(ex-sx)*3);
   else
   if(prc && prc->mode==PROCESSWHITE) memset(pwroww,0xFF,(ex-sx)*3);
   else
   {
				if(sph->selector.type.type==SPT_CMYK)
				{
     pww=(int*)spworkspace;
     pwrowltw=(int*)((char*)pww+4*(ex-sx));

     while(pww<pwrowltw)
     {
      if(crf==0)
      {
       cr=*prw;
       crf=1;
      }
    
      *pww++=cr;
 
      xdisc+=sxpix;
      while(xdisc>=dxpix)
      {
       xdisc-=dxpix;
       if(incx>0) prw++;
       else       prw--;
       crf=0;
      }
 				}

     if(prc && prc->mode==PROCESSFN)
     {
      prc->process(prc->handle,spworkspace,spworkspace,(ex-sx),PROCESST_RGB32,PROCESST_CMYK,NULL,NULL);
     }
					else
					if(prc)
					{
						prc->cmyk2rgb(prc->handle2,(int*)spworkspace,(int*)spworkspace,(ex-sx),NULL,NULL);
					}

     pw=(char*)pwroww;
					prw=(int*)spworkspace;
     pwrowlt=pw+3*(ex-sx);

     while(pw<pwrowlt)
     {
						cr=*prw++;
      *pw++=(char)((cr & 0xFF0000)>>16);
      *pw++=(char)((cr & 0xFF00)>>8);
      *pw++=(char)((cr & 0xFF));
					}
				}
				else
				{
     pw=(char*)pwroww;
     pwrowlt=pw+3*(ex-sx);

     while(pw<pwrowlt)
     {
      if(crf==0)
      {
       cr=*prw;
       crf=1;
      }
    
      *pw++=(char)((cr & 0xFF0000)>>16);
      *pw++=(char)((cr & 0xFF00)>>8);
      *pw++=(char)((cr & 0xFF));
 
      xdisc+=sxpix;
      while(xdisc>=dxpix)
      {
       xdisc-=dxpix;
       if(incx>0) prw++;
       else       prw--;
       crf=0;
      }
 				}

     if(prc && prc->mode==PROCESSFN)
     {
      prc->process(prc->handle,pwroww,pwroww,(ex-sx),PROCESST_BGR24,PROCESST_BGR24,NULL,NULL);
     }
				}
   }



   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
   } 
  }
 }
}





static void setspritembitsscaled(int * sbits,int sbpp,int sbytewidth,int sshift,int * mbits,int dbpp,int sxpix,int sypix,int dxpix,int dypix,int * dbits,int backindex,int sx,int ex,int sy,int ey,processstr * prc)
{
 char       * pr;
 char       * pm;
 char       * prrow;
 char       * pmrow;
 char       * pw;
 char       * pwrow;
 char       * pwlt;
 char       * pwrowlt;

 int        * prroww;
 int        * pmroww;
 int        * pwroww;
 int        * pwltw;
 int        * pwrowltw;

 int          shiftr;
 int          shiftw;
 int          shiftltw;
 int          mask;
 int          cr=0;
 int          cw;

 int          xdisc;
 int          ydisc;
 int          backcol;
 int          dbytewidth;
 int        * prw;
 int        * pww;
 int        * pmw;
 int          incx;
 int          incy;
 int          xdiscp;
 int          nx;
 int          ny;
 int          crf;


 if(dxpix<0)
 {
  incx=-sbpp;
  dxpix=-dxpix;
 }
 else
 {
  incx=sbpp;
 }


 dbytewidth=(((ex-sx)*dbpp+31)&~31)/8;


 if(dypix<0)
 {
  incy=-dbytewidth;
  dypix=-dypix;
  pwltw=(int*)dbits;
  pwroww=(int*)(((char*)pwltw)+((ey-sy)-1)*dbytewidth);
 }
 else
 {
  incy=dbytewidth;
  pwroww=(int*)dbits;
  pwltw=(int*)(((char*)pwroww)+(ey-sy)*dbytewidth);
 }

 xdiscp=sx*sxpix;
 nx=(xdiscp/dxpix);
 xdiscp-=nx*dxpix;

 if(incy>=0)
 {
  ydisc=(dypix-ey)*sypix;
 }
 else
 {
  ydisc=sy*sypix;
 }

 ny=(ydisc/dypix);
 ydisc-=ny*dypix;


 prroww=(int*)(((char*)sbits)+ny*sbytewidth);
 pmroww=(int*)(((char*)mbits)+ny*sbytewidth);

 sshift/=sbpp;


 if(sbpp==1)
 {
  
  mask=1;
  if(backindex) backcol=0xFFFFFFFF;
  else          backcol=0;

  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/32;
   shiftltw=((ex-sx) & 31)*1;
   shiftltw=(7-(shiftltw & 7))+(shiftltw & 24);
   xdisc=xdiscp;
   shiftw=7;
   cw=0;

   crf=0;

   if(incx>0)
   {
    pmw=pmroww+((sshift+nx)/32);
    prw=prroww+((sshift+nx)/32);
    shiftr=((sshift+nx) & 0x1F);
   }
   else
   {
    pmw=pmroww+(sshift+sxpix-1-nx)/32;
    prw=prroww+(sshift+sxpix-1-nx)/32;
    shiftr=(sshift+sxpix-1-nx)&0x1F;
   }


   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     cr^=backcol;
     cr&=*pmw;
     cr^=backcol;
     crf=1;
    }
    
    cw|=((cr>>shiftr)&mask)<<shiftw;
    if(shiftw & 7) shiftw--;
    else           shiftw+=15;

    if(shiftw>=32)
    {
     shiftw=7;
     *pww++=cw;
     cw=0;
    }
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     shiftr+=incx;
     if(shiftr>=32)
     {
      crf=0;
      prw++;
      pmw++;
      shiftr=0;
     }
     else
     if(shiftr<0)
     {
      crf=0;
      prw--;
      pmw--;
      shiftr=31;
     }
    }
   }

   if(shiftw!=7) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
    pmroww=(int*)(((char*)pmroww)+sbytewidth);
   } 
  }
 }
 else
 if(sbpp==2)
 {
  
  mask=3;
  backcol=(backindex<<2)|backindex;
  backcol|=(backcol<<4);
  backcol|=(backcol<<8);
  backcol|=(backcol<<16);

  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

  
   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/8;
   shiftltw=((ex-sx) & 7)*4;
   if(shiftltw & 4) shiftltw-=4;
   else             shiftltw+=4;
   xdisc=xdiscp;

   shiftw=4;
   cw=0;

   crf=0;

   if(incx>0)
   {
    prw=prroww+((sshift+nx)/16);
    pmw=pmroww+((sshift+nx)/16);
    shiftr=((sshift+nx) & 0xF)*2;
   }
   else
   {
    prw=prroww+(sshift+sxpix-1-nx)/16;
    pmw=pmroww+(sshift+sxpix-1-nx)/16;
    shiftr=((sshift+sxpix-1-nx)&0xF)*2;
   }


   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     cr^=backcol;
     cr&=*pmw;
     cr^=backcol;
     crf=1;
    }
    
    cw|=((cr>>shiftr)&mask)<<shiftw;
    if(shiftw & 4) shiftw-=4;
    else           shiftw+=12;

    if(shiftw>=32)
    {
     shiftw=4;
     *pww++=cw;
     cw=0;
    }
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     shiftr+=incx;
     if(shiftr>=32)
     {
      crf=0;
      prw++;
      pmw++;
      shiftr=0;
     }
     else
     if(shiftr<0)
     {
      crf=0;
      prw--;
      pmw--;
      shiftr=30;
     }
    }
   }

   if(shiftw!=4) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
    pmroww=(int*)(((char*)pmroww)+sbytewidth);
   } 
  }
 }
 else
 if(sbpp==4)
 {
  mask=15;
  backcol=(backindex<<4)|backindex;
  backcol|=(backcol<<8);
  backcol|=(backcol<<16);


  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

   
   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/8;
   shiftltw=((ex-sx) & 7)*4;
   if(shiftltw & 4) shiftltw-=4;
   else             shiftltw+=4;
   xdisc=xdiscp;
   shiftw=4;
   cw=0;

   crf=0;

   if(incx>0)
   {
    prw=prroww+((sshift+nx)/8);
    pmw=pmroww+((sshift+nx)/8);
    shiftr=((sshift+nx) & 0x7)*4;
   }
   else
   {
    prw=prroww+(sshift+sxpix-1-nx)/8;
    pmw=pmroww+(sshift+sxpix-1-nx)/8;
    shiftr=((sshift+sxpix-1-nx)&0x7)*4;
   }


   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     cr^=backcol;
     cr&=*pmw;
     cr^=backcol;
     crf=1;
    }
    
    cw|=((cr>>shiftr)&mask)<<shiftw;
    if(shiftw & 4) shiftw-=4;
    else           shiftw+=12;

    if(shiftw>=32)
    {
     shiftw=4;
     *pww++=cw;
     cw=0;
    }
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     shiftr+=incx;
     if(shiftr>=32)
     {
      crf=0;
      prw++;
      pmw++;
      shiftr=0;
     }
     else
     if(shiftr<0)
     {
      crf=0;
      prw--;
      pmw--;
      shiftr=28;
     }
    }
   }

   if(shiftw!=4) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
    pmroww=(int*)(((char*)pmroww)+sbytewidth);
   } 
  }
 }
 else
 if(sbpp==8)
 {

  prrow=(char*)prroww;
  pmrow=(char*)pmroww;
  pwrow=(char*)pwroww;
  pwlt=(char*)pwltw;

  backcol=backindex;

  while(1)
  {
   if(incy>0)
   {
    if(pwrow>=pwlt) break;
   }
   else
   {
    if(pwrow<pwlt) break;
   }

  
   pw=pwrow;
   pwrowlt=pwrow+(ex-sx);
   xdisc=xdiscp;

   crf=0;
   
   if(incx>0)
   {
    pr=prrow+(sshift+nx);
    pm=pmrow+(sshift+nx);
   }
   else
   {
    pr=prrow+(sshift+sxpix-1-nx);
    pm=pmrow+(sshift+sxpix-1-nx);
   }

   while(pw<pwrowlt)
   {
    if(crf==0)
    {
     cr=*pr;
     cr^=backcol;
     cr&=*pm;
     cr^=backcol;
     crf=1;
    }

    *pw++=(char)cr;
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     if(incx>0)
     {
      crf=0;
      pr++;
      pm++;
     }
     else
     {
      crf=0;
      pr--;
      pm--;
     }
    }
   }

   pwrow+=incy;

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prrow+=sbytewidth;
    pmrow+=sbytewidth;
   } 
  }
 }
 else
 if(sbpp==16)
 {
  // can never happen

 	USE(prc);
 }
 else
 if(sbpp==32)
 {
  // can never happen

 	USE(prc);
 }
}





static void setspritembitsscaled1(int * sbits,int sbpp,int sbytewidth,int sshift,int * mbits,int dbpp,int sxpix,int sypix,int dxpix,int dypix,int * dbits,int backindex,int sx,int ex,int sy,int ey,processstr * prc,newsprite_header * sph)
{
	os_error   * err;
 char       * pr;
 char       * prrow;
 char       * pw;
 char       * pwrow;
 char       * pwlt;
 char       * pwrowlt;

 int        * prroww;
 int        * pmroww;
 int        * pwroww;
 int        * pwltw;
 int        * pwrowltw;

 int          shiftr;
 int          shiftw;
 int          shiftm;
 int          shiftltw;
 int          mask;
 int          cr=0;
 int          cr2;
 int          cw;
 int          cm=0;
 
 int          xdisc;
 int          ydisc;
 int          backcol;
 int          dbytewidth;
 int          mbytewidth;
 int        * prw;
 int        * pww;
 int        * pmw;
 int          incx;
 int          inc1x;
 int          incy;
 int          xdiscp;
 int          nx;
 int          ny;
 int          crf;
 int          cmf;


 if(dxpix<0)
 {
  incx=-sbpp;
  inc1x=-1;
  dxpix=-dxpix;
 }
 else
 {
  incx=sbpp;
  inc1x=1;
 }


 dbytewidth=(((ex-sx)*dbpp+31)&~31)/8;
 mbytewidth=((sxpix*1+31)&~31)/8;


 if(dypix<0)
 {
  incy=-dbytewidth;
  dypix=-dypix;
  pwltw=(int*)dbits;
  pwroww=(int*)(((char*)pwltw)+((ey-sy)-1)*dbytewidth);
 }
 else
 {
  incy=dbytewidth;
  pwroww=(int*)dbits;
  pwltw=(int*)(((char*)pwroww)+(ey-sy)*dbytewidth);
 }


 xdiscp=sx*sxpix;
 nx=(xdiscp/dxpix);
 xdiscp-=nx*dxpix;

 if(incy>=0)
 {
  ydisc=(dypix-ey)*sypix;
 }
 else
 {
  ydisc=sy*sypix;
 }

 ny=(ydisc/dypix);
 ydisc-=ny*dypix;


 prroww=(int*)(((char*)sbits)+ny*sbytewidth);
 pmroww=(int*)(((char*)mbits)+ny*mbytewidth);



 if(sbpp==1)
 {
// can never get here
 
 }
 if(sbpp==2)
 {
  
  backcol=backindex;

  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

  
   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/8;
   shiftltw=((ex-sx) & 7)*4;
   if(shiftltw & 4) shiftltw-=4;
   else             shiftltw+=4;
   xdisc=xdiscp;

   shiftw=4;
   cw=0;

   cmf=crf=0;

   if(incx<0)
   {
    prw=prroww+(sxpix-1-nx)/16;
    pmw=pmroww+(sxpix-1-nx)/32;
    shiftr=((sxpix-1-nx) & 0xF)*2;
    shiftm=(sxpix-1-nx)&0x1F;
   }
   else
   {
    prw=prroww+(nx/16);
    pmw=pmroww+(nx/32);
    shiftr=sshift+(nx & 0xF)*2;
    shiftm=nx & 0x1F;
   }

   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(cmf==0)
    {
     cm=*pmw;
     cmf=1;
    }

    if(crf==0)
    {
     crf=1;
     cr=*prw;
    }

    if(((cm>>shiftm)&0x1)==0) cr2=backcol;
    else                      cr2=(cr>>shiftr)&0x3;
    
    cw|=(cr2<<shiftw);

    if(shiftw & 4) shiftw-=4;
    else           shiftw+=12;

    if(shiftw>=32)
    {
     shiftw=4;
     *pww++=cw;
     cw=0;
    }
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     shiftr+=incx;
     if(shiftr>=32)
     {
      crf=0;
      prw++;
      shiftr=0;
     }
     else
     if(shiftr<0)
     {
      crf=0;
      prw--;
      shiftr=30;
     }

     shiftm+=inc1x;
     if(shiftm>=32)
     {
      cmf=0;
      pmw++;
      shiftm=0;
     }
     else
     if(shiftm<0)
     {
      cmf=0;
      pmw--;
      shiftm=31;
     }
    }
   }

   if(shiftw!=4) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
    pmroww=(int*)(((char*)pmroww)+mbytewidth);
   } 
  }
 }
 if(sbpp==4)
 {
  
  mask=15;
  backcol=backindex;


  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

  
   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/8;
   shiftltw=((ex-sx) & 7)*4;
   if(shiftltw & 4) shiftltw-=4;
   else             shiftltw+=4;
   xdisc=xdiscp;
   shiftw=4;
   cw=0;

   cmf=crf=0;

   if(incx>0)
   {
    prw=prroww+(nx/8);
    pmw=pmroww+(nx/32);
    shiftr=sshift+(nx & 0x7)*4;
    shiftm=(nx & 0x1F);
   }
   else
   {
    prw=prroww+(sxpix-1-nx)/8;
    pmw=pmroww+(sxpix-1-nx)/32;
    shiftr=((sxpix-1-nx)&0x7)*4;
    shiftm=(sxpix-1-nx)&0x1F;
   }


   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(cmf==0)
    {
     cm=*pmw;
     cmf=1;
    }

    if(crf==0)
    {
     crf=1;
     cr=*prw;
    }

    if(((cm>>shiftm)&0x1)==0) cr2=backcol;
    else                      cr2=(cr>>shiftr)&0xF;
    
    cw|=(cr2<<shiftw);
    
    if(shiftw & 4) shiftw-=4;
    else           shiftw+=12;

    if(shiftw>=32)
    {
     shiftw=4;
     *pww++=cw;
     cw=0;
    }
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     shiftr+=incx;
     if(shiftr>=32)
     {
      crf=0;
      prw++;
      shiftr=0;
     }
     else
     if(shiftr<0)
     {
      crf=0;
      prw--;
      shiftr=28;
     }

     shiftm+=inc1x;
     if(shiftm>=32)
     {
      cmf=0;
      pmw++;
      shiftm=0;
     }
     else
     if(shiftm<0)
     {
      cmf=0;
      pmw--;
      shiftm=31;
     }
    }
   }

   if(shiftw!=4) *pww++=cw;

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
    pmroww=(int*)(((char*)pmroww)+mbytewidth);
   } 
  }
 }
 else
 if(sbpp==8)
 {
  
  prrow=(char*)prroww;
  pwrow=(char*)pwroww;
  pwlt=(char*)pwltw;

  backcol=backindex;

  while(1)
  {
   if(incy>0)
   {
    if(pwrow>=pwlt) break;
   }
   else
   {
    if(pwrow<pwlt) break;
   }
   
   
   pw=pwrow;
   pwrowlt=pwrow+(ex-sx);
   xdisc=xdiscp;

   if(incx>0)
   {
    pr=prrow+(sshift/8)+nx;
    pmw=pmroww+(sshift/8+nx)/32;
    shiftm=(sshift/8+nx)&0x1F;
   }
   else
   {
    pr=prrow+(sxpix-1-nx);
    pmw=pmroww+(sxpix-1-nx)/32;
    shiftm=(sxpix-1-nx)&0x1F;
   }

   crf=cmf=0;

   while(pw<pwrowlt)
   {
    if(cmf==0)
    {
     cm=*pmw;
     cmf=1;
    }

    if(crf==0)
    {
     crf=1;
     if(((cm>>shiftm)&0x1)==0) cr=backcol;
     else                      cr=*pr;
    }

    *pw++=(char)cr;
    
    xdisc+=sxpix;
    while(xdisc>=dxpix)
    {
     xdisc-=dxpix;
     crf=0;
     if(incx>0) pr++;
     else       pr--;

     shiftm+=inc1x;
     if(shiftm>=32)
     {
      cmf=0;
      pmw++;
      shiftm=0;
     }
     else
     if(shiftm<0)
     {
      cmf=0;
      pmw--;
      shiftm=31;
     }
    }
   }

   pwrow+=incy;

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prrow+=sbytewidth;
    pmroww=(int*)(((char*)pmroww)+mbytewidth);
   } 
  }
 }
 else
 if(sbpp==16)
 {

  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

  
   pww=pwroww;
   pwrowltw=pwroww+(ex-sx)/2;
   shiftltw=((ex-sx) & 1)*16;

   xdisc=xdiscp;
   shiftw=0;
   cw=0;

   crf=cmf=0;

   if(incx>0)
   {
    prw=prroww+(nx/2);
    pmw=pmroww+(nx/32);
    shiftr=(nx & 0x1)*16;
    shiftm=(nx & 0x1F);
   }
   else
   {
    prw=prroww+(sxpix-1-nx)/2;
    pmw=pmroww+(sxpix-1-nx)/32;
    shiftr=((sxpix-1-nx)&1)*16;
    shiftm=((sxpix-1-nx)&0x1F);
   }


			if(prc && (prc->mode==PROCESSFN || prc->mode==PROCESSBLACK || prc->mode==PROCESSWHITE))
			{
				int   x;
				int   xlt;
				int * pww32;
				int * pwwcopy;
				int * pwwskip;
				int   lastw;
				int   nval;
				int   copy;
				int   skip;

				x=0;
				xlt=(ex-sx);

		  err=schunk((flex_ptr)&spworkspace,(xlt+1)*4*3,SPWORKCHUNK);

				pww32=((int*)spworkspace)+1;
				pwwcopy=(int*)(spworkspace+(xlt+1)*4);
				pwwskip=(int*)(spworkspace+(xlt+1)*8);
				*pwwskip=0;
				*pwwcopy=0;
				lastw=-1;

    while(x++<xlt)
    {
     if(crf==0)
     {
      cr=*prw;
      crf=1;
     }

     if(cmf==0)
     {
      cm=*pmw;
      cmf=1;
     }

     if(((cm>>shiftm)&0x1))
     {
      if(shiftr==0)  cw=(expand16[(cr & 0x1F)]<<16)|(expand16[(cr & 0x3E0)>>5]<<8)|(expand16[(cr & 0x7C00)>>10]);
      else           cw=(expand16[(cr & 0x1F0000)>>16]<<16)|(expand16[(cr & 0x3E00000)>>21]<<8)|(expand16[(cr & 0x7C000000)>>26]);

						if(cw==lastw)
      {
       *pwwcopy=(*pwwcopy)+1;
						}
						else
    		{
							lastw=cw;
							*pww32++=cw;
							*++pwwcopy=1;
							*++pwwskip=0;
						}
     }
     else
     {
						lastw=-1;
						*pwwskip=(*pwwskip)+1;
     }

   
     xdisc+=sxpix;
     while(xdisc>=dxpix)
     {
      xdisc-=dxpix;

      shiftr+=incx;
      if(shiftr>=32)
      {
       crf=0;
       prw++;
       shiftr=0;
      }
      else
      if(shiftr<0)
      {
       crf=0;
       prw--;
       shiftr=16;
      }

      shiftm+=inc1x;
      if(shiftm>=32)
      {
       cmf=0;
       pmw++;
       shiftm=0;
      }
      else
      if(shiftm<0)
      {
       cmf=0;
       pmw--;
       shiftm=31;
      }
     }
    }

				nval=pww32-((int*)spworkspace);

//				cprintf("nval %d pww32 %x spwork %x",nval,pww32,spworkspace);

				if(prc->mode==PROCESSBLACK)
				{
					memset(spworkspace,0,nval*sizeof(int));
				}
				else
				if(prc->mode==PROCESSWHITE)
				{
					memset(spworkspace,0xFF,nval*sizeof(int));
				}
				else			// PROCESSFN
				{
     prc->process(prc->handle,spworkspace,spworkspace,nval,PROCESST_BGR32,PROCESST_BGR32,NULL,NULL);
				}

				pww32=(int*)spworkspace;
				pwwcopy=(int*)(spworkspace+(xlt+1)*4);
				pwwskip=(int*)(spworkspace+(xlt+1)*8);
				cw=0;

    while(pww<pwrowltw || shiftw!=shiftltw)
    {
 				cr=*pww32++;
					copy=*pwwcopy++;
					skip=*pwwskip++;
 				while(copy--)
 				{
 					if(shiftw==0) cw=((cr & 0xF8)>>3)|((cr & 0xF800)>>6)|((cr & 0xF80000)>>9);
 					else										cw|=((cr & 0xF8)<<13)|((cr & 0xF800)<<10)|((cr & 0xF80000)<<7);

      shiftw+=16;

      if(shiftw>=32)
      {
       shiftw=0;
       *pww++=cw;
      }
 				}

 				while(skip--)
 				{
 					if(shiftw==0) cw=0;

      shiftw+=16;

      if(shiftw>=32)
      {
       shiftw=0;
       *pww++=cw;
      }
 				}

				}

				if(shiftw!=0) *pww++=cw;
			}
			else
			{
    while(pww<pwrowltw || shiftw!=shiftltw)
    {
     if(crf==0)
     {
      cr=*prw;
      crf=1;
     }

     if(cmf==0)
     {
      cm=*pmw;
      cmf=1;
     }

     if(((cm>>shiftm)&0x1))
     {
      if(shiftw==0)
      {
       if(shiftr==0)  cw=((cr & 0x1F)<<10)|((cr & 0x3E0))|((cr & 0x7C00)>>10);
       else           cw=((cr & 0x1F0000)>>6)|((cr & 0x3E00000)>>16)|((cr & 0x7C000000)>>26);
      }
      else
      {
       if(shiftr==0) cw|=((cr & 0x1F)<<26)|((cr & 0x3E0)<<16)|((cr & 0x7C00)<<6);
       else          cw|=((cr & 0x1F0000)<<10)|((cr & 0x3E00000))|((cr & 0x7C000000)>>10);
      }
     }
     else
     {
      if(shiftw==0) cw=0;
     }


     shiftw+=16;

     if(shiftw>=32)
     {
      shiftw=0;
      *pww++=cw;
     }
    
     xdisc+=sxpix;
     while(xdisc>=dxpix)
     {
      xdisc-=dxpix;

      shiftr+=incx;
      if(shiftr>=32)
      {
       crf=0;
       prw++;
       shiftr=0;
      }
      else
      if(shiftr<0)
      {
       crf=0;
       prw--;
       shiftr=16;
      }

      shiftm+=inc1x;
      if(shiftm>=32)
      {
       cmf=0;
       pmw++;
       shiftm=0;
      }
      else
      if(shiftm<0)
      {
       cmf=0;
       pmw--;
       shiftm=31;
      }
     }
    }

    if(shiftw!=0) *pww++=cw;
			}



   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
    pmroww=(int*)(((char*)pmroww)+mbytewidth);
   } 
  }


 }
 else
 if(sbpp==32)
 {
  while(1)
  {
   if(incy>0)
   {
    if(pwroww>=pwltw) break;
   }
   else
   {
    if(pwroww<pwltw) break;
   }

   
   pw=(char*)pwroww;
   pwrowlt=pw+(ex-sx)*3;
   cw=0;
   xdisc=xdiscp;

   crf=cmf=0;

   if(incx>0)
   {
    shiftm=(nx & 0x1F);
    prw=prroww+nx;
    pmw=pmroww+(nx/32);
   }
   else
   {
    shiftm=(sxpix-1-nx)&0x1F;
    prw=prroww+(sxpix-1-nx);
    pmw=pmroww+(sxpix-1-nx)/32;
   }

			if(prc && (prc->mode==PROCESSFN || prc->mode==PROCESSBLACK || prc->mode==PROCESSWHITE || sph->selector.type.type==SPT_CMYK))
			{
				int   x;
				int   xlt;
				int * pww32;
				int * pwwcopy;
				int * pwwskip;
				int   lastw;
				int   nval;
				int   copy;
				int   skip;

				x=0;
				xlt=(ex-sx);

		  err=schunk((flex_ptr)&spworkspace,(xlt+1)*4*3,SPWORKCHUNK);

				pww32=((int*)spworkspace)+1;
				pwwcopy=(int*)(spworkspace+(xlt+1)*4);
				pwwskip=(int*)(spworkspace+(xlt+1)*8);
				*pwwskip=0;
				*pwwcopy=0;
				lastw=-1;

    while(x++<xlt)
    {
     if(crf==0)
     {
      cr=*prw;
      crf=1;
     }

     if(cmf==0)
     {
      cm=*pmw;
      cmf=1;
     }

     if(((cm>>shiftm)&0x1))
     {
						if(cr==lastw)
      {
       *pwwcopy=(*pwwcopy)+1;
						}
						else
    		{
							lastw=cr;
							*pww32++=cr;
							*++pwwcopy=1;
							*++pwwskip=0;
						}
     }
     else
     {
						lastw=-1;
						*pwwskip=(*pwwskip)+1;
     }

   
     xdisc+=sxpix;
     while(xdisc>=dxpix)
     {
      xdisc-=dxpix;

      if(incx>0)
      {
       crf=0;
       prw++;
      }
      else
      {
       crf=0;
       prw--;
      }

      shiftm+=inc1x;
      if(shiftm>=32)
      {
       cmf=0;
       pmw++;
       shiftm=0;
      }
      else
      if(shiftm<0)
      {
       cmf=0;
       pmw--;
       shiftm=31;
      }
     }
    }

				nval=pww32-((int*)spworkspace);

//				cprintf("nval %d pww32 %x spwork %x",nval,pww32,spworkspace);

				if(prc->mode==PROCESSBLACK)
				{
					memset(spworkspace,0,nval*sizeof(int));
				}
				else
				if(prc->mode==PROCESSWHITE)
				{
					memset(spworkspace,0xFF,nval*sizeof(int));
				}
				else
				if(prc->mode==PROCESSFN)
				{
					if(sph->selector.type.type==SPT_CMYK)
 				{
      prc->process(prc->handle,spworkspace,spworkspace,nval,PROCESST_RGB32,PROCESST_CMYK,NULL,NULL);
 				}
 				else
 				{
      prc->process(prc->handle,spworkspace,spworkspace,nval,PROCESST_RGB32,PROCESST_RGB32,NULL,NULL);
 				}
				}
				else
				if(sph->selector.type.type==SPT_CMYK)
				{
     prc->cmyk2rgb(prc->handle2,(int*)spworkspace,(int*)spworkspace,nval,NULL,NULL);
				}

				pww32=(int*)spworkspace;
				pwwcopy=(int*)(spworkspace+(xlt+1)*4);
				pwwskip=(int*)(spworkspace+(xlt+1)*8);

    while(pw<pwrowlt)
    {
 				cr=*pww32++;
					copy=*pwwcopy++;
					skip=*pwwskip++;
 				while(copy--)
 				{
      *pw++=(char)((cr & 0xFF0000)>>16);
      *pw++=(char)((cr & 0xFF00)>>8);
      *pw++=(char)((cr & 0xFF));
 				}

 				while(skip--)
 				{
      *pw++=0;
      *pw++=0;
      *pw++=0;
 				}
				}

			}
			else
			{
    while(pw<pwrowlt)
    {
     if(crf==0)
     {
      cr=*prw;
      crf=1;
     }

     if(cmf==0)
     {
      cm=*pmw;
      cmf=1;
     }

     if((cm>>shiftm)&0x1)
     {
      *pw++=(char)((cr & 0xFF0000)>>16);
      *pw++=(char)((cr & 0xFF00)>>8);
      *pw++=(char)((cr & 0xFF));
     }
     else
     {
      *pw++=0;
      *pw++=0;
      *pw++=0;
     }

     xdisc+=sxpix;
     while(xdisc>=dxpix)
     {
      xdisc-=dxpix;
      crf=0;
      if(incx>0) prw++;
      else       prw--;

      shiftm+=inc1x;
      if(shiftm>=32)
      {
       cmf=0;
       pmw++;
       shiftm=0;
      }
      else
      if(shiftm<0)
      {
       cmf=0;
       pmw--;
       shiftm=31;
      }
     }
    }
			}

   pwroww=(int*)(((char*)pwroww)+incy);

   ydisc+=sypix;
   while(ydisc>=dypix)
   {
    ydisc-=dypix;
    prroww=(int*)(((char*)prroww)+sbytewidth);
    pmroww=(int*)(((char*)pmroww)+mbytewidth);
   } 
  }
 }
}



























os_error * sprite_renderscaled(newsprite_header * sph,HDC hdc,int x,int y,int dxpix,int dypix,int * pixtrans,processstr * prc)
{
 os_error   * err;
 int          xpix;
 int          ypix;
 int        * bits;
 int        * mbits;
 bmphdrstr    bmi;
 int          bitwidth;
 int          sbytewidth;
 int          mbytewidth;
 int          spbpp;
 int          mbpp;
 int        * dbits;
 HBITMAP      hbm;
// BITMAP       bitmap;
 HDC          hdcmem;
 int          dbpp;
 int          backindex;
 int          sshift;
 int          absdxpix;
 int          absdypix;
 RECT         clipbox;
 int          code;
 int          sx;
 int          ex;
 int          sy;
 int          ey;




// cprintf("renderbox");

 err=NULL;

 code=GetClipBox(hdc,&clipbox);
 if(code!=NULLREGION)
 {
// cprintf("render scaled");

  absdxpix=ABS(dxpix);
  absdypix=ABS(dypix);

  if(dypix>0) y-=dypix;
  if(dxpix<0) x+=dxpix;
  
  sy=clipbox.top-y;
  if(sy<0) sy=0;

  ey=clipbox.bottom-y;
  if(ey>absdypix) ey=absdypix;

  sx=clipbox.left-x;
  if(sx<0) sx=0;

  ex=clipbox.right-x;
  if(ex>absdxpix) ex=absdxpix;

  if(sx<ex && sy<ey)
  {

   getspriteinfo((newsprite_header*)sph,NULL,NULL,&spbpp);

//cprintf("render scaled bpp %d",spbpp);

   bits=(int*)(((char*)sph)+sph->image);
   mbits=(int*)(((char*)sph)+sph->mask);
   bitwidth=32*(sph->width+1)-sph->lbit-(31-sph->rbit);
   sbytewidth=4*(sph->width+1);
   sshift=sph->lbit;

   xpix=bitwidth/spbpp;
   ypix=sph->height+1;


   if(spbpp==2)  dbpp=4;
   else
   if(spbpp==32) dbpp=24;
   else          dbpp=spbpp;


// SetDIBits(hdc,hbm,0,ypix,bits,(BITMAPINFO*)&bmi,DIB_RGB_COLORS);

   if(mbits==bits || (((newsprite_header*)sph)->selector.type.alpha))   // the no mask case
   {
//				cprintf("no mask");

    memset(&bmi,0,sizeof(bmi));

    osgetpal(spbpp,(newsprite_header *)sph,&bmi.pal[0],NULL,pixtrans);

    bmi.bmi.biWidth=ex-sx;
    bmi.bmi.biHeight=ey-sy;
    bmi.bmi.biXPelsPerMeter=0;
    bmi.bmi.biYPelsPerMeter=0;
    bmi.bmi.biBitCount=(unsigned short)dbpp;
    bmi.bmi.biPlanes=1;
    bmi.bmi.biCompression=BI_RGB;
    bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
    bmi.bmi.biSizeImage=0;
    bmi.bmi.biClrImportant=0;
    if(spbpp==2) bmi.bmi.biClrUsed=(1<<spbpp);
    else         bmi.bmi.biClrUsed=0;

    err=XCreateDIBSection(hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
    if(!err)
    {
     setspritebitsscaled(bits,spbpp,sbytewidth,sshift,dbpp,xpix,ypix,dxpix,-dypix,dbits,sx,ex,sy,ey,prc,(newsprite_header*)sph);

     SetStretchBltMode(hdc,COLORONCOLOR);

//   GetObject(hbm,sizeof(BITMAP),&bitmap);

     hdcmem=CreateCompatibleDC(hdc);
     SelectObject(hdcmem,hbm);
     BitBlt(hdc,x+sx,y+sy,ex-sx,ey-sy,hdcmem,0,0,SRCCOPY);
     DeleteDC(hdcmem);
 
     DeleteObject(hbm);
    }
   }
   else            // we have a mask
   {
    if(((newsprite_header*)sph)->selector.type.type)
    {
     mbpp=dbpp=1;
     mbytewidth=((xpix+31)&(~31))/8;
    }
    else
    {
     mbpp=spbpp;
     if(spbpp==2) dbpp=4;
     else         dbpp=spbpp;
     mbytewidth=sbytewidth;
    }

    if(((newsprite_header*)sph)->selector.type.alpha)
    {
     mbpp=8;
					mbytewidth=((xpix+3)&(~3));
				}

    osgetmpal(mbpp,&bmi.pal[0]);

    bmi.bmi.biWidth=ex-sx;
    bmi.bmi.biHeight=ey-sy;
    bmi.bmi.biXPelsPerMeter=0;
    bmi.bmi.biYPelsPerMeter=0;
    bmi.bmi.biBitCount=(unsigned short)dbpp;
    bmi.bmi.biPlanes=1;
    bmi.bmi.biCompression=BI_RGB;
    bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
    bmi.bmi.biSizeImage=0;
    bmi.bmi.biClrImportant=0;
    if(mbpp==2) bmi.bmi.biClrUsed=(1<<mbpp);
    else        bmi.bmi.biClrUsed=0;

    err=XCreateDIBSection(hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
    if(!err)
    {
// we can assume bytewidth and start shift are the same for the mask because
// new format masks don't have left hand wastage
   
     setspritebitsscaled(mbits,mbpp,mbytewidth,sshift,dbpp,xpix,ypix,dxpix,-dypix,dbits,sx,ex,sy,ey,prc,(newsprite_header*)sph);

//   GetObject(hbm,sizeof(BITMAP),&bitmap);

     hdcmem=CreateCompatibleDC(hdc);
     SelectObject(hdcmem,hbm);
     BitBlt(hdc,x+sx,y+sy,ex-sx,ey-sy,hdcmem,0,0,SRCAND);  // and mask
     DeleteDC(hdcmem);
     DeleteObject(hbm);

// Now start on main image

     if(spbpp==2)  dbpp=4;
     else
     if(spbpp==32) dbpp=24;
     else          dbpp=spbpp;
   
     osgetpal(spbpp,(newsprite_header *)sph,&bmi.pal[0],&backindex,pixtrans);

     bmi.bmi.biWidth=ex-sx;
     bmi.bmi.biHeight=ey-sy;
     bmi.bmi.biXPelsPerMeter=0;
     bmi.bmi.biYPelsPerMeter=0;
     bmi.bmi.biBitCount=(unsigned short)dbpp;
     bmi.bmi.biPlanes=1;
     bmi.bmi.biCompression=BI_RGB;
     bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
     bmi.bmi.biSizeImage=0;
     bmi.bmi.biClrImportant=0;
     if(spbpp==2) bmi.bmi.biClrUsed=(1<<spbpp);
     else         bmi.bmi.biClrUsed=0;

     err=XCreateDIBSection(hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
     if(!err)
     {
      if(spbpp==mbpp) setspritembitsscaled(bits,spbpp,sbytewidth,sshift,mbits,dbpp,xpix,ypix,dxpix,-dypix,dbits,backindex,sx,ex,sy,ey,prc);
      else            setspritembitsscaled1(bits,spbpp,sbytewidth,sshift,mbits,dbpp,xpix,ypix,dxpix,-dypix,dbits,backindex,sx,ex,sy,ey,prc,(newsprite_header*)sph);

//    GetObject(hbm,sizeof(BITMAP),&bitmap);

      hdcmem=CreateCompatibleDC(hdc);
      SelectObject(hdcmem,hbm);
      BitBlt(hdc,x+sx,y+sy,ex-sx,ey-sy,hdcmem,0,0,SRCPAINT); // or data
      DeleteDC(hdcmem);
      DeleteObject(hbm);
     }
    }
   }
  }
 }

// SetDIBitsToDevice(hdc,x,y,xpix,ypix,0,0,0,ypix,bits,(BITMAPINFO*)&bmi,DIB_RGB_COLORS);


 return(err);
}



/*******************************************************************************/

// Rotated sprite plotting

/*
    scan bitmap in y axis
      find intersections with sprite area in bitmap space
         map to sprite space
      fill bitmap - copy from sprite - fill bitmap

*/




static void setmaskbitsscaledspan(int dx,int dy,int dspanx,int dbpp,int * dbits,int sx,int ex,processstr * prc)
{
 char * pw;
 int    n;
 int    m;
 int    c;
 int    dbytewidth;
 int    dxlt;
 int    dspanxp;

 dxlt=dx+dspanx;
 if(dx<sx)
 {
  dx=sx;
 }

 if(dxlt>(ex-1)) dxlt=ex-1;
 dspanxp=dxlt-dx+1;


 dbytewidth=(((ex-sx)*dbpp+31)&~31)/8;
 dbits=(int*)(((char*)dbits)+dbytewidth*dy);

 if(dbpp==1)
 {
  n=(dx-sx)&0x7;  // clear top n bits
  if(n)
  {
   m=MIN(8-n,dspanxp); // set m bits
   c=0xFF;
   c=c>>(8-m);
   c=c<<(8-n-m);
   pw=((char*)dbits)+(dx-sx)/8;
   *pw=(char)c;
   dspanxp-=m;
   dx+=m;
  }

  if(dspanxp)
  {
   pw=((char*)dbits)+(dx-sx)/8;
   n=dspanxp/8;
   memset(pw,0xFF,n);
   pw+=n;
   n=dspanxp&0x7; // set top n bits
   if(n)
   {
    *pw=(char)((0xFF)<<(8-n));
   }
  }
 }

	USE(prc);
}



// passed 2 points, start and end

static void setspritebitsscaledspan(POINT * pts,int * sbits,int sbpp,int sbytewidth,int sshift,int dx,int dy,int dspanx,int dbpp,int * dbits,
                                    int sx,int ex,processstr * prc,newsprite_header * sph)
{
	os_error   * err;
 char       * pr;
 char       * pw;
 char       * pwrowlt;
	int        * pwroww;
 int        * pwrowltw;

 int          shiftr;
 int          shiftw;
 int          shiftltw;
 int          cr;
 int          cw;

 int          xdisc;
 int          ydisc;
 int          dbytewidth;
 int        * prw;
 int        * pww;

 int          incx;
 int          incy;
 int          sspanx;
 int          sspany;

 int          inc1x;

 int          dxlt;
 int          dspanxp;
 int          nx;
 int          ny;
 int          crf;


//	cprintf("setspritebitsscaledspan sx %d ex %d",sx,ex);


	if(sph->selector.type.type==SPT_CMYK)
	{
		err=schunk((flex_ptr)&spworkspace,(ex-sx)*4,SPWORKCHUNK);
	}


 dbytewidth=(((ex-sx)*dbpp+31)&~31)/8;

 sspanx=pts[1].x-pts[0].x;
 if(sspanx>=0)
 {
  inc1x=1;
  incx=sbpp;
 }
 else
 {
  inc1x=-1;
  incx=-sbpp;
  sspanx=-sspanx;
 }

 sspany=pts[1].y-pts[0].y;
 if(sspany>=0) incy=sbytewidth;
 else
 {
  incy=-sbytewidth;
  sspany=-sspany;
 }



 dxlt=dx+dspanx;
 if(dx<sx)
 {
  xdisc=(sx-dx)*sspanx;
  nx=(xdisc/dspanx);
  xdisc-=nx*dspanx;

  ydisc=(sx-dx)*sspany;
  ny=(ydisc/dspanx);
  ydisc-=ny*dspanx;

  dx=sx;
 }
 else
 {
  nx=ny=xdisc=ydisc=0;
 }

 if(dxlt>(ex-1)) dxlt=ex-1;
 dspanxp=dxlt-dx;


 if(sbpp==1)
 {
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x+sshift+nx*incx)/32;
  shiftr=((pts[0].x+sshift+nx*incx) & 0x1F);
  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/32;
  shiftw=((dx-sx) & 0x1F);
  shiftw=(7-(shiftw & 7))+(shiftw & 24);
  pwrowltw=pww+(((dx-sx) & 0x1F)+(dxlt-dx+1))/32;
  shiftltw=(((dx-sx) & 0x1F)+(dxlt-dx+1)) & 0x1F;
  shiftltw=(7-(shiftltw & 7))+(shiftltw & 24);

  cr=crf=0;
  cw=0;

  while(pww<pwrowltw || shiftw!=shiftltw)
  {
   if(crf==0)
   {
    cr=*prw;
    crf=1;
   }

   
   cw|=((cr>>shiftr)&0x1)<<shiftw;

   if(shiftw & 7) shiftw--;
   else           shiftw+=15;

   if(shiftw>=32)
   {
    shiftw=7;
    *pww++=cw;
    cw=0;
   }
   
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    shiftr+=incx;
    if(shiftr>=32)
    {
     shiftr=0;
     prw++;
     crf=0;
    }
    else
    if(shiftr<0)
    {
     shiftr=31;
     prw--;
     crf=0;
    }
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    prw=(int*)((char*)prw+incy);
    crf=0;
   } 
  }

  if(shiftw!=7) *pww=cw;
 }
 else
 if(sbpp==2)
 {
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x*2+sshift+nx*incx)/32;
  shiftr=((pts[0].x*2+sshift+incx*nx) & 0x1F);
  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/8;
  shiftw=((dx-sx) & 0x7)*4;
  if(shiftw & 4) shiftw-=4;
  else           shiftw+=4;
  pwrowltw=pww+(((dx-sx) & 0x7)+(dxlt-dx+1))/8;
  shiftltw=((((dx-sx) & 0x7)+(dxlt-dx+1)) & 0x7)*4;
  if(shiftltw & 4) shiftltw-=4;
  else             shiftltw+=4;

  cr=crf=0;
  cw=0;

  
  while(pww<pwrowltw || shiftw!=shiftltw)
  {
   if(crf==0)
   {
    cr=*prw;
    crf=1;
   }
   
   cw|=((cr>>shiftr)&0x3)<<shiftw;

   if(shiftw & 4) shiftw-=4;
   else           shiftw+=12;

   if(shiftw>=32)
   {
    shiftw=4;
    *pww++=cw;
    cw=0;
   }
   
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    shiftr+=incx;
    if(shiftr>=32)
    {
     shiftr=0;
     prw++;
     crf=0;
    }
    else
    if(shiftr<0)
    {
     shiftr=30;
     prw--;
     crf=0;
    }
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    prw=(int*)((char*)prw+incy);
    crf=0;
   } 
  }

  if(shiftw!=4) *pww=cw;
 }
 else
 if(sbpp==4)
 {
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x*4+sshift+nx*incx)/32;
  shiftr=((pts[0].x*4+sshift+nx*incx) & 0x1F);
  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/8;
  shiftw=((dx-sx) & 0x7)*4;
  if(shiftw & 4) shiftw-=4;
  else           shiftw+=4;
  pwrowltw=pww+(((dx-sx) & 0x7)+(dxlt-dx+1))/8;
  shiftltw=((((dx-sx) & 0x7)+(dxlt-dx+1)) & 0x7)*4;
  if(shiftltw & 4) shiftltw-=4;
  else             shiftltw+=4;

  cr=crf=0;
  cw=0;

  
  while(pww<pwrowltw || shiftw!=shiftltw)
  {
   if(crf==0)
   {
    cr=*prw;
    crf=1;
   }
   
   cw|=((cr>>shiftr)&0xF)<<shiftw;

   if(shiftw & 4) shiftw-=4;
   else           shiftw+=12;

   if(shiftw>=32)
   {
    shiftw=4;
    *pww++=cw;
    cw=0;
   }
   
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    shiftr+=incx;
    if(shiftr>=32)
    {
     shiftr=0;
     prw++;
     crf=0;
    }
    else
    if(shiftr<0)
    {
     shiftr=28;
     prw--;
     crf=0;
    }
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    prw=(int*)((char*)prw+incy);
    crf=0;
   } 
  }

  if(shiftw!=4) *pww=cw;
 }
 else
 if(sbpp==8)
 {
  pr=(char*)sbits+sbytewidth*pts[0].y+pts[0].x+(sshift/8);
  pr+=incy*ny+inc1x*nx;
  pw=(char*)dbits+dbytewidth*dy+(dx-sx);
  pwrowlt=pw+(dxlt-dx);

  while(pw<=pwrowlt)
  {
   *pw++=*pr;
    
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    pr+=inc1x;
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    pr+=incy;
   } 
  }
 }
 else
 if(sbpp==16)
 {
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y))+(pts[0].x*16+sshift+nx*incx)/32;
  prw=(int*)((char*)prw+incy*ny);

  shiftr=((pts[0].x*16+sshift+nx*incx) & 0x1F);

  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/2;
		pwroww=pww;
  shiftw=((dx-sx) & 0x1)*16;
  pwrowltw=pww+(((dx-sx) & 0x1)+(dxlt-dx+1))/2;
  shiftltw=((((dx-sx) & 0x1)+(dxlt-dx+1)) & 0x1)*16;
		pw=((char*)pwroww)+shiftw/8;

  crf=cr=0;
  cw=0;


  if(prc && prc->mode==PROCESSBLACK) memset(pw,0,(dxlt-dx+1)*2);
  else
  if(prc && prc->mode==PROCESSWHITE) memset(pw,0xFF,(dxlt-dx+1)*2);
  else
  {
   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     crf=1;
    }
   
    if(shiftw==0)
    {
     if(shiftr==0)  cw=((cr & 0x1F)<<10)|((cr & 0x3E0))|((cr & 0x7C00)>>10);
     else           cw=((cr & 0x1F0000)>>6)|((cr & 0x3E00000)>>16)|((cr & 0x7C000000)>>26);
    }
    else
    {
     if(shiftr==0) cw|=((cr & 0x1F)<<26)|((cr & 0x3E0)<<16)|((cr & 0x7C00)<<6);
     else          cw|=((cr & 0x1F0000)<<10)|((cr & 0x3E00000))|((cr & 0x7C000000)>>10);
    }

    shiftw+=16;

    if(shiftw>=32)
    {
     shiftw=0;
     *pww++=cw;
    }
   
    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;
     shiftr+=incx;
     if(shiftr>=32)
     {
      shiftr=0;
      prw++;
      crf=0;
     }
     else
     if(shiftr<0)
     {
      shiftr=16;
      prw--;
      crf=0;
     }
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     prw=(int*)((char*)prw+incy);
     crf=0;
    }
			}

	  if(shiftw) *pww=cw;

   if(prc && prc->mode==PROCESSFN)
   {
    prc->process(prc->handle,pw,pw,(dxlt-dx+1),PROCESST_BGR16,PROCESST_BGR16,NULL,NULL);
   }
  }
 }
 else
 if(sbpp==32)
 {

//  cprintf("spanx %d spany %d dspanx %d",sspanx,sspany,dspanx);

  prw=(int*)(((char*)sbits)+sbytewidth*pts[0].y+pts[0].x*sizeof(int));
  prw+=inc1x*nx;
  prw=(int*)((char*)prw+incy*ny);

  pw=((char*)dbits)+dbytewidth*dy+(dx-sx)*3;
		pwroww=(int*)pw;
  pwrowlt=pw+(dxlt-dx)*3;

  cr=cw=0;


  if(prc && prc->mode==PROCESSBLACK) memset(pw,0,(dxlt-dx+1)*3);
  else
  if(prc && prc->mode==PROCESSWHITE) memset(pw,0xFF,(dxlt-dx+1)*3);
  else
  {
			if(sph->selector.type.type==SPT_CMYK)
			{
				pww=(int*)spworkspace;
				pwrowltw=pww+(dxlt-dx);

				while(pww<=pwrowltw)
				{
     if(cw==0)
     {
      cr=*prw;
      cw=1;
     }
   
     *pww++=cr;

     xdisc+=sspanx;
     while(xdisc>=dspanx)
     {
      xdisc-=dspanx;
      prw+=inc1x;
      cw=0;
     }

     ydisc+=sspany;
     while(ydisc>=dspanx)
     {
      ydisc-=dspanx;
      prw=(int*)(((char*)prw)+incy);
      cw=0;
     } 
				}

    if(prc && prc->mode==PROCESSFN)
    {
     prc->process(prc->handle,spworkspace,spworkspace,(dxlt-dx+1),PROCESST_RGB32,PROCESST_CMYK,NULL,NULL);
    }
				else
				if(prc)
				{
					prc->cmyk2rgb(prc->handle2,(int*)spworkspace,(int*)spworkspace,(dxlt-dx+1),NULL,NULL);
				}

    pw=((char*)dbits)+dbytewidth*dy+(dx-sx)*3;
				prw=(int*)spworkspace;

    while(pw<=pwrowlt)
    {
					cr=*prw++;
     *pw++=(char)((cr & 0xFF0000)>>16);
     *pw++=(char)((cr & 0xFF00)>>8);
     *pw++=(char)((cr & 0xFF));
				}

			}
			else
			{
    while(pw<=pwrowlt)
    {
     if(cw==0)
     {
      cr=*prw;
      cw=1;
     }
   
     *pw++=(char)((cr & 0xFF0000)>>16);
     *pw++=(char)((cr & 0xFF00)>>8);
     *pw++=(char)((cr & 0xFF));

     xdisc+=sspanx;
     while(xdisc>=dspanx)
     {
      xdisc-=dspanx;
      prw+=inc1x;
      cw=0;
     }

     ydisc+=sspany;
     while(ydisc>=dspanx)
     {
      ydisc-=dspanx;
      prw=(int*)(((char*)prw)+incy);
      cw=0;
     } 
    }

    if(prc && prc->mode==PROCESSFN)
    {
     prc->process(prc->handle,pwroww,pwroww,(dxlt-dx+1),PROCESST_BGR24,PROCESST_BGR24,NULL,NULL);
    }
			}
		}
 }
}





// passed 2 points, start and end

static void setspritembitsscaledspan(POINT * pts,int * sbits,int sbpp,int sbytewidth,int sshift,int * mbits,int dx,int dy,int dspanx,int dbpp,int * dbits,int backindex,int sx,int ex,processstr * prc)
{
 char       * pr;
 char       * pm;
 char       * pw;
 char       * pwrowlt;

 int        * pwrowltw;

 int          shiftr;
 int          shiftw;
 int          shiftltw;
 int          cr;
 int          cw;

 int          xdisc;
 int          ydisc;
 int          dbytewidth;
 int        * prw;
 int        * pmw;
 int        * pww;

 int          incx;
 int          incy;
 int          sspanx;
 int          sspany;

 int          inc1x;
 int          backcol;

 int          nx;
 int          ny;
 int          dxlt;
 int          dspanxp;
 int          crf;


 
 dbytewidth=(((ex-sx)*dbpp+31)&~31)/8;

 sspanx=pts[1].x-pts[0].x;
 if(sspanx>=0)
 {
  inc1x=1;
  incx=sbpp;
 }
 else
 {
  inc1x=-1;
  incx=-sbpp;
  sspanx=-sspanx;
 }

 sspany=pts[1].y-pts[0].y;
 if(sspany>=0) incy=sbytewidth;
 else
 {
  incy=-sbytewidth;
  sspany=-sspany;
 }



 dxlt=dx+dspanx;
 if(dx<sx)
 {
  xdisc=(sx-dx)*sspanx;
  nx=(xdisc/dspanx);
  xdisc-=nx*dspanx;

  ydisc=(sx-dx)*sspany;
  ny=(ydisc/dspanx);
  ydisc-=ny*dspanx;

  dx=sx;
 }
 else
 {
  nx=ny=xdisc=ydisc=0;
 }

 if(dxlt>(ex-1)) dxlt=ex-1;
 dspanxp=dxlt-dx;




 if(sbpp==1)
 {
  if(backindex) backcol=0xFFFFFFFF;
  else          backcol=0;

  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x+sshift+nx*incx)/32;
  shiftr=((pts[0].x+sshift+nx*incx) & 0x1F);
  pmw=((int*)((char*)mbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x+sshift+nx*incx)/32;
  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/32;
  shiftw=((dx-sx) & 0x1F);
  shiftw=(7-(shiftw & 7))+(shiftw & 24);
  pwrowltw=pww+(((dx-sx) & 0x1F)+(dxlt-dx+1))/32;
  shiftltw=(((dx-sx) & 0x1F)+(dxlt-dx+1)) & 0x1F;
  shiftltw=(7-(shiftltw & 7))+(shiftltw & 24);

  cr=crf=0;
  cw=0;

  while(pww<pwrowltw || shiftw!=shiftltw)
  {
   if(crf==0)
   {
    cr=*prw;
    cr^=backcol;
    cr&=*pmw;
    cr^=backcol;
    crf=1;
   }
   
   cw|=((cr>>shiftr)&0x1)<<shiftw;

   if(shiftw & 7) shiftw--;
   else           shiftw+=15;

   if(shiftw>=32)
   {
    shiftw=7;
    *pww++=cw;
    cw=0;
   }
   
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    shiftr+=incx;
    if(shiftr>=32)
    {
     shiftr=0;
     prw++;
     pmw++;
     crf=0;
    }
    else
    if(shiftr<0)
    {
     shiftr=31;
     prw--;
     pmw--;
     crf=0;
    }
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    prw=(int*)((char*)prw+incy);
    pmw=(int*)((char*)pmw+incy);
    crf=0;
   } 
  }

  if(shiftw!=7) *pww=cw;
 }
 else
 if(sbpp==2)
 {
  backcol=(backindex<<2)|backindex;
  backcol|=(backcol<<4);
  backcol|=(backcol<<8);
  backcol|=(backcol<<16);
  
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x*2+sshift+nx*incx)/32;
  shiftr=((pts[0].x*2+sshift+nx*incx) & 0x1F);
  pmw=((int*)((char*)mbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x*2+sshift+nx*incx)/32;
  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/8;
  shiftw=((dx-sx) & 0x7)*4;
  if(shiftw & 4) shiftw-=4;
  else           shiftw+=4;
  pwrowltw=pww+(((dx-sx) & 0x7)+(dxlt-dx+1))/8;
  shiftltw=((((dx-sx) & 0x7)+(dxlt-dx+1)) & 0x7)*4;
  if(shiftltw & 4) shiftltw-=4;
  else             shiftltw+=4;

  cr=crf=0;
  cw=0;

  
  while(pww<pwrowltw || shiftw!=shiftltw)
  {
   if(crf==0)
   {
    cr=*prw;
    cr^=backcol;
    cr&=*pmw;
    cr^=backcol;
    crf=1;
   }
   
   cw|=((cr>>shiftr)&0x3)<<shiftw;

   if(shiftw & 4) shiftw-=4;
   else           shiftw+=12;

   if(shiftw>=32)
   {
    shiftw=4;
    *pww++=cw;
    cw=0;
   }
   
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    shiftr+=incx;
    if(shiftr>=32)
    {
     shiftr=0;
     prw++;
     pmw++;
     crf=0;
    }
    else
    if(shiftr<0)
    {
     shiftr=30;
     prw--;
     pmw--;
     crf=0;
    }
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    prw=(int*)((char*)prw+incy);
    pmw=(int*)((char*)pmw+incy);
    crf=0;
   } 
  }

  if(shiftw!=4) *pww=cw;
 }
 else
 if(sbpp==4)
 {
  backcol=(backindex<<4)|backindex;
  backcol|=(backcol<<8);
  backcol|=(backcol<<16);

  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x*4+sshift+nx*incx)/32;
  shiftr=((pts[0].x*4+sshift+nx*incx) & 0x1F);
  pmw=((int*)((char*)mbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x*4+sshift+nx*incx)/32;

  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/8;
  shiftw=((dx-sx) & 0x7)*4;
  if(shiftw & 4) shiftw-=4;
  else           shiftw+=4;
  pwrowltw=pww+(((dx-sx) & 0x7)+(dxlt-dx+1))/8;
  shiftltw=((((dx-sx) & 0x7)+(dxlt-dx+1)) & 0x7)*4;
  if(shiftltw & 4) shiftltw-=4;
  else             shiftltw+=4;

  cr=crf=0;
  cw=0;
  
  while(pww<pwrowltw || shiftw!=shiftltw)
  {
   if(crf==0)
   {
    cr=*prw;
    cr^=backcol;
    cr&=*pmw;
    cr^=backcol;
    crf=1;
   }
   
   cw|=((cr>>shiftr)&0xF)<<shiftw;

   if(shiftw & 4) shiftw-=4;
   else           shiftw+=12;

   if(shiftw>=32)
   {
    shiftw=4;
    *pww++=cw;
    cw=0;
   }
   
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    shiftr+=incx;
    if(shiftr>=32)
    {
     shiftr=0;
     crf=0;
     prw++;
     pmw++;
    }
    else
    if(shiftr<0)
    {
     shiftr=28;
     crf=0;
     prw--;
     pmw--;
    }
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    prw=(int*)((char*)prw+incy);
    pmw=(int*)((char*)pmw+incy);
    crf=0;
   } 
  }

  if(shiftw!=4) *pww=cw;
 }
 else
 if(sbpp==8)
 {
  backcol=backindex;

  pr=(char*)sbits+sbytewidth*pts[0].y+pts[0].x+(sshift/8)+nx*inc1x+ny*incy;
  pm=(char*)mbits+sbytewidth*pts[0].y+pts[0].x+(sshift/8)+nx*inc1x+ny*incy;
  pw=(char*)dbits+dbytewidth*dy+(dx-sx);
  pwrowlt=pw+(dxlt-dx);

  cr=crf=0;

  while(pw<=pwrowlt)
  {
   if(crf==0)
   {
    cr=*pr;
    cr^=backcol;
    cr&=*pm;
    cr^=backcol;
    crf=1;
   }
   
   *pw++=(char)cr;

   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    pr+=inc1x;
    pm+=inc1x;
    crf=0;
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    pr+=incy;
    pm+=incy;
    crf=0;
   } 
  }
 }
 else
 if(sbpp==16)
 {

// can never happen
 
 }
 else
 if(sbpp==32)
 {

// can never happen
 
 }

	USE(prc);
}





// passed 2 points, start and end

static void setspritembitsscaledspan1(POINT * pts,int * sbits,int sbpp,int sbytewidth,int sxpix,int sshift,int * mbits,int dx,int dy,int dspanx,int dbpp,int * dbits,int backindex,int sx,int ex,processstr * prc,newsprite_header * sph)
{
	os_error   * err;
 char       * pr;
 char       * pw;
 char       * pwrowlt;

 int        * pwrowltw;

 int          shiftr;
 int          shiftw;
 int          shiftm;
 int          shiftltw;
 int          cr;
 int          cw;
 int          cm;
 int          cr2;

 int          xdisc;
 int          ydisc;
 int          dbytewidth;
 int          mbytewidth;
 int        * prw;
 int        * pmw;
 int        * pww;

 int          incx;
 int          incy;
 int          sspanx;
 int          sspany;

 int          inc1x;
 int          incmy;
 int          backcol;

 int          nx;
 int          ny;
 int          dxlt;
 int          dspanxp;
 int          cmf;
 int          crf;

 
 dbytewidth=(((ex-sx)*dbpp+31)&~31)/8;
 mbytewidth=((sxpix*1+31)&~31)/8;

 sspanx=pts[1].x-pts[0].x;
 if(sspanx>=0)
 {
  inc1x=1;
  incx=sbpp;
 }
 else
 {
  inc1x=-1;
  incx=-sbpp;
  sspanx=-sspanx;
 }

 sspany=pts[1].y-pts[0].y;
 if(sspany>=0)
 {
  incy=sbytewidth;
  incmy=mbytewidth;
 }
 else
 {
  incy=-sbytewidth;
  incmy=-mbytewidth;
  sspany=-sspany;
 }


 dxlt=dx+dspanx;
 if(dx<sx)
 {
  xdisc=(sx-dx)*sspanx;
  nx=(xdisc/dspanx);
  xdisc-=nx*dspanx;

  ydisc=(sx-dx)*sspany;
  ny=(ydisc/dspanx);
  ydisc-=ny*dspanx;

  dx=sx;
 }
 else
 {
  nx=ny=xdisc=ydisc=0;
 }

 if(dxlt>(ex-1)) dxlt=ex-1;
 dspanxp=dxlt-dx;


// cprintf("inc1x %d incmy %d",inc1x,incmy);

 if(sbpp==1)
 {
// Can never get here
 
 }
 else
 if(sbpp==2)
 {
  backcol=backindex;
  
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x*2+sshift+nx*incx)/32;
  shiftr=((pts[0].x*2+sshift+nx*incx) & 0x1F);

  pmw=((int*)((char*)mbits+mbytewidth*pts[0].y+ny*incmy))+(pts[0].x+sshift+nx*inc1x)/32;
  shiftm=(pts[0].x+sshift+nx*inc1x)&0x1F;

  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/8;
  shiftw=((dx-sx) & 0x7)*4;
  if(shiftw & 4) shiftw-=4;
  else           shiftw+=4;
  pwrowltw=pww+(((dx-sx) & 0x7)+(dxlt-dx+1))/8;
  shiftltw=((((dx-sx) & 0x7)+(dxlt-dx+1)) & 0x7)*4;
  if(shiftltw & 4) shiftltw-=4;
  else             shiftltw+=4;

  cr=crf=0;
  cm=cmf=0;
  cw=0;

  
  while(pww<pwrowltw || shiftw!=shiftltw)
  {
   if(cmf==0)
   {
    cm=*pmw;
    cmf=1;
   }

   if(crf==0)
   {
    cr=*prw;
    crf=1;
   }

   if(((cm>>shiftm)&0x1)==0) cr2=backcol;
   else                      cr2=(cr>>shiftr)&0x3;

   cw|=(cr2<<shiftw);

   if(shiftw & 4) shiftw-=4;
   else           shiftw+=12;

   if(shiftw>=32)
   {
    shiftw=4;
    *pww++=cw;
    cw=0;
   }
   
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    shiftr+=incx;
    if(shiftr>=32)
    {
     shiftr=0;
     prw++;
     crf=0;
    }
    else
    if(shiftr<0)
    {
     shiftr=30;
     prw--;
     crf=0;
    }

    shiftm+=inc1x;
    if(shiftm>=32)
    {
     pmw++;
     shiftm=0;
     cmf=0;
    }
    else
    if(shiftm<0)
    {
     pmw--;
     shiftm=31;
     cmf=0;
    }
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    prw=(int*)((char*)prw+incy);
    pmw=(int*)((char*)pmw+incmy);
    cmf=crf=0;
   }
  }

  if(shiftw!=4) *pww=cw;
 }
 else
 if(sbpp==4)
 {
  backcol=backindex;

  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x*4+sshift+nx*incx)/32;
  shiftr=((pts[0].x*4+sshift+nx*incx) & 0x1F);

  pmw=((int*)((char*)mbits+mbytewidth*pts[0].y+ny*incmy))+(pts[0].x+sshift+nx*inc1x)/32;
  shiftm=(pts[0].x+sshift+nx*inc1x)&0x1F;

  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/8;
  shiftw=((dx-sx) & 0x7)*4;
  if(shiftw & 4) shiftw-=4;
  else           shiftw+=4;

  pwrowltw=pww+(((dx-sx) & 0x7)+(dxlt-dx+1))/8;
  shiftltw=((((dx-sx) & 0x7)+(dxlt-dx+1)) & 0x7)*4;
  if(shiftltw & 4) shiftltw-=4;
  else             shiftltw+=4;

  cr=crf=0;
  cm=cmf=0;
  cw=0;

  
  while(pww<pwrowltw || shiftw!=shiftltw)
  {
   if(cmf==0)
   {
    cm=*pmw;
    cmf=1;
   }

   if(crf==0)
   {
    cr=*prw;
    crf=1;
   }

   if(((cm>>shiftm)&0x1)==0) cr2=backcol;
   else                      cr2=(cr>>shiftr)&0xF;
   
   cw|=(cr2<<shiftw);

   if(shiftw & 4) shiftw-=4;
   else           shiftw+=12;

   if(shiftw>=32)
   {
    shiftw=4;
    *pww++=cw;
    cw=0;
   }
   
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    shiftr+=incx;
    if(shiftr>=32)
    {
     shiftr=0;
     prw++;
     crf=0;
    }
    else
    if(shiftr<0)
    {
     shiftr=28;
     prw--;
     crf=0;
    }

    shiftm+=inc1x;
    if(shiftm>=32)
    {
     pmw++;
     cmf=0;
     shiftm=0;
    }
    else
    if(shiftm<0)
    {
     pmw--;
     cmf=0;
     shiftm=31;
    }
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    prw=(int*)((char*)prw+incy);
    pmw=(int*)((char*)pmw+incmy);
    crf=0;
    cmf=0;
   }
  }

  if(shiftw!=4) *pww=cw;
 }
 else
 if(sbpp==8)
 {
  backcol=backindex;

  pr=(char*)sbits+sbytewidth*pts[0].y+pts[0].x+(sshift/8)+nx*inc1x+ny*incy;
  pmw=((int*)(((char*)mbits)+mbytewidth*pts[0].y+ny*incmy))+(pts[0].x+inc1x*nx)/32;
  shiftm=(pts[0].x+inc1x*nx) & 0x1F;

  pw=(char*)dbits+dbytewidth*dy+(dx-sx);
  pwrowlt=pw+(dxlt-dx);

  cr=crf=0;
  cm=cmf=0;

  while(pw<=pwrowlt)
  {
   if(cmf==0)
   {
    cm=*pmw;
    cmf=1;
   }

   if(crf==0)
   {
    if(((cm>>shiftm)&0x1)==0) cr=backcol;
    else                      cr=*pr;
    crf=1;
   }
   
   *pw++=(char)cr;

   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    pr+=inc1x;
    crf=0;

    shiftm+=inc1x;
    if(shiftm>=32)
    {
     pmw++;
     cmf=0;
     shiftm=0;
    }
    else
    if(shiftm<0)
    {
     pmw--;
     cmf=0;
     shiftm=31;
    }

   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    pr+=incy;
    pmw=(int*)(((char*)pmw)+incmy);
    cmf=0;
    crf=0;
   } 
  }
 }
 else
 if(sbpp==16)
 {
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+ny*incy))+(pts[0].x+nx*inc1x)/2;
  shiftr=((pts[0].x+nx*inc1x) & 0x1)*16;

  pmw=((int*)(((char*)mbits)+mbytewidth*pts[0].y+ny*incmy))+(pts[0].x+nx*inc1x)/32;
  shiftm=(pts[0].x+nx*inc1x) & 0x1F;

  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/2;
  shiftw=((dx-sx) & 0x1)*16;
  pwrowltw=pww+(((dx-sx) & 0x1)+(dxlt-dx+1))/2;
  shiftltw=((((dx-sx) & 0x1)+(dxlt-dx+1)) & 0x1)*16;

  cr=crf=0;
  cm=cmf=0;
  cw=0;


		if(prc && (prc->mode==PROCESSFN || prc->mode==PROCESSWHITE || prc->mode==PROCESSBLACK))
		{
			int   x;
			int   xlt;
			int * pww32;
			int * pwwcopy;
			int * pwwskip;
			int   lastw;
			int   nval;
			int   copy;
			int   skip;

			x=0;
			xlt=(dxlt-dx+1);

	  err=schunk((flex_ptr)&spworkspace,(xlt+1)*4*3,SPWORKCHUNK);

			pww32=((int*)spworkspace)+1;
			pwwcopy=(int*)(spworkspace+(xlt+1)*4);
			pwwskip=(int*)(spworkspace+(xlt+1)*8);
			*pwwskip=0;
			*pwwcopy=0;
			lastw=-1;


   while(x++<xlt)
   {
    if(crf==0)
    {
     cr=*prw;
     crf=1;
    }

    if(cmf==0)
    {
     cm=*pmw;
     cmf=1;
    }

    if(((cm>>shiftm)&0x1))
    {
     if(shiftr==0)  cw=(expand16[(cr & 0x1F)]<<16)|(expand16[(cr & 0x3E0)>>5]<<8)|(expand16[(cr & 0x7C00)>>10]);
     else           cw=(expand16[(cr & 0x1F0000)>>16]<<16)|(expand16[(cr & 0x3E00000)>>21]<<8)|(expand16[(cr & 0x7C000000)>>26]);

					if(cw==lastw)
     {
      *pwwcopy=(*pwwcopy)+1;
					}
					else
   		{
						lastw=cw;
						*pww32++=cw;
						*++pwwcopy=1;
						*++pwwskip=0;
					}
    }
    else
    {
					lastw=-1;
					*pwwskip=(*pwwskip)+1;
    }
   
    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;
     shiftr+=incx;
     if(shiftr>=32)
     {
      shiftr=0;
      prw++;
      crf=0;
     }
     else
     if(shiftr<0)
     {
      shiftr=16;
      prw--;
      crf=0;
     }

     shiftm+=inc1x;
     if(shiftm>=32)
     {
      pmw++;
      cmf=0;
      shiftm=0;
     }
     else
     if(shiftm<0)
     {
      pmw--;
      cmf=0;
      shiftm=31;
     }
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     prw=(int*)(((char*)prw)+incy);
     pmw=(int*)(((char*)pmw)+incmy);
     crf=0;
     cmf=0;
    } 
   }

			nval=pww32-((int*)spworkspace);

//			cprintf("nval %d pww32 %x spwork %x",nval,pww32,spworkspace);

			if(prc->mode==PROCESSBLACK)
			{
				memset(spworkspace,0,nval*sizeof(int));
			}
			else
			if(prc->mode==PROCESSWHITE)
			{
				memset(spworkspace,0xFF,nval*sizeof(int));
			}
			else
			{
    prc->process(prc->handle,spworkspace,spworkspace,nval,PROCESST_BGR32,PROCESST_BGR32,NULL,NULL);
			}

			pww32=(int*)spworkspace;
			pwwcopy=(int*)(spworkspace+(xlt+1)*4);
			pwwskip=(int*)(spworkspace+(xlt+1)*8);
			cw=0;

   while(pww<pwrowltw || shiftw!=shiftltw)
   {
				cr=*pww32++;
				copy=*pwwcopy++;
				skip=*pwwskip++;
				while(copy--)
				{
					if(shiftw==0) cw=((cr & 0xF8)>>3)|((cr & 0xF800)>>6)|((cr & 0xF80000)>>9);
					else										cw|=((cr & 0xF8)<<13)|((cr & 0xF800)<<10)|((cr & 0xF80000)<<7);

     shiftw+=16;

     if(shiftw>=32)
     {
      shiftw=0;
      *pww++=cw;
     }
				}

				while(skip--)
				{
					if(shiftw==0) cw=0;

     shiftw+=16;

     if(shiftw>=32)
     {
      shiftw=0;
      *pww++=cw;
     }
				}
			}

			if(shiftw!=0) *pww++=cw;

		}
  else
		{
   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     crf=1;
    }

    if(cmf==0)
    {
     cm=*pmw;
     cmf=1;
    }

    if(((cm>>shiftm)&0x1))
    {
     if(shiftw==0)
     {
      if(shiftr==0)  cw=((cr & 0x1F)<<10)|((cr & 0x3E0))|((cr & 0x7C00)>>10);
      else           cw=((cr & 0x1F0000)>>6)|((cr & 0x3E00000)>>16)|((cr & 0x7C000000)>>26);
     }
     else
     {
      if(shiftr==0) cw|=((cr & 0x1F)<<26)|((cr & 0x3E0)<<16)|((cr & 0x7C00)<<6);
      else          cw|=((cr & 0x1F0000)<<10)|((cr & 0x3E00000))|((cr & 0x7C000000)>>10);
     }
    }
    else
    {
     if(shiftw==0) cw=0;
    }

    shiftw+=16;

    if(shiftw>=32)
    {
     shiftw=0;
     *pww++=cw;
    }
   
    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;
     shiftr+=incx;
     if(shiftr>=32)
     {
      shiftr=0;
      prw++;
      crf=0;
     }
     else
     if(shiftr<0)
     {
      shiftr=16;
      prw--;
      crf=0;
     }

     shiftm+=inc1x;
     if(shiftm>=32)
     {
      pmw++;
      cmf=0;
      shiftm=0;
     }
     else
     if(shiftm<0)
     {
      pmw--;
      cmf=0;
      shiftm=31;
     }
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     prw=(int*)(((char*)prw)+incy);
     pmw=(int*)(((char*)pmw)+incmy);
     crf=0;
     cmf=0;
    } 
   }

   if(shiftw) *pww=cw;
		}
 }
 else
 if(sbpp==32)
 {
  prw=(int*)(((char*)sbits)+sbytewidth*pts[0].y+pts[0].x*sizeof(int));
  prw+=inc1x*nx;
  prw=(int*)((char*)prw+incy*ny);

  pmw=((int*)(((char*)mbits)+mbytewidth*pts[0].y+ny*incmy))+(pts[0].x+nx*inc1x)/32;
  shiftm=(pts[0].x+nx*inc1x) & 0x1F;

  pw=((char*)dbits)+dbytewidth*dy+(dx-sx)*3;
  pwrowlt=pw+(dxlt-dx)*3;


  cr=cw=0;
  cm=cmf=0;


	 if(prc && (prc->mode==PROCESSFN || prc->mode==PROCESSBLACK || prc->mode==PROCESSWHITE || sph->selector.type.type==SPT_CMYK))
		{
			int   x;
			int   xlt;
			int * pww32;
			int * pwwcopy;
			int * pwwskip;
			int   lastw;
			int   nval;
			int   copy;
			int   skip;

			x=0;
			xlt=(dxlt-dx+1);

	  err=schunk((flex_ptr)&spworkspace,(xlt+1)*4*3,SPWORKCHUNK);

			pww32=((int*)spworkspace)+1;
			pwwcopy=(int*)(spworkspace+(xlt+1)*4);
			pwwskip=(int*)(spworkspace+(xlt+1)*8);
			*pwwskip=0;
			*pwwcopy=0;
			lastw=-1;

   while(x++<xlt)
   {
    if(cw==0)
    {
     cr=*prw;
     cw=1;
    }

    if(cmf==0)
    {
     cm=*pmw;
     cmf=1;
    }

    if((cm>>shiftm)&0x1)
    {
					if(cr==lastw)
     {
      *pwwcopy=(*pwwcopy)+1;
					}
					else
   		{
						lastw=cr;
						*pww32++=cr;
						*++pwwcopy=1;
						*++pwwskip=0;
					}
    }
    else
    {
					lastw=-1;
					*pwwskip=(*pwwskip)+1;
    }

    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;
     prw+=inc1x;
     cw=0;

     shiftm+=inc1x;
     if(shiftm>=32)
     {
      pmw++;
      cmf=0;
      shiftm=0;
     }
     else
     if(shiftm<0)
     {
      pmw--;
      cmf=0;
      shiftm=31;
     }
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     prw=(int*)(((char*)prw)+incy);
     pmw=(int*)(((char*)pmw)+incmy);
     cw=0;
     cmf=0;
    }
			}



	 	nval=pww32-((int*)spworkspace);

//			cprintf("nval %d pww32 %x spwork %x",nval,pww32,spworkspace);

			if(prc->mode==PROCESSBLACK)
			{
				memset(spworkspace,0,nval*sizeof(int));
			}
			else
			if(prc->mode==PROCESSWHITE)
			{
				memset(spworkspace,0xFF,nval*sizeof(int));
			}
			else
			if(prc->mode==PROCESSFN)
			{
				if(sph->selector.type.type==SPT_CMYK)
				{
     prc->process(prc->handle,spworkspace,spworkspace,nval,PROCESST_RGB32,PROCESST_CMYK,NULL,NULL);
				}
				else
				{
     prc->process(prc->handle,spworkspace,spworkspace,nval,PROCESST_RGB32,PROCESST_RGB32,NULL,NULL);
				}
			}
			else
			if(sph->selector.type.type==SPT_CMYK)
			{
    prc->cmyk2rgb(prc->handle2,(int*)spworkspace,(int*)spworkspace,nval,NULL,NULL);
			}

			pww32=(int*)spworkspace;
			pwwcopy=(int*)(spworkspace+(xlt+1)*4);
			pwwskip=(int*)(spworkspace+(xlt+1)*8);

   while(pw<=pwrowlt)
   {
				cr=*pww32++;
				copy=*pwwcopy++;
				skip=*pwwskip++;
				while(copy--)
				{
     *pw++=(char)((cr & 0xFF0000)>>16);
     *pw++=(char)((cr & 0xFF00)>>8);
     *pw++=(char)((cr & 0xFF));
				}

				while(skip--)
				{
     *pw++=0;
     *pw++=0;
     *pw++=0;
				}
			}

		}
		else
		{
   while(pw<=pwrowlt)
   {
    if(cw==0)
    {
     cr=*prw;
     cw=1;
    }

    if(cmf==0)
    {
     cm=*pmw;
     cmf=1;
    }

    if((cm>>shiftm)&0x1)
    {
     *pw++=(char)((cr & 0xFF0000)>>16);
     *pw++=(char)((cr & 0xFF00)>>8);
     *pw++=(char)((cr & 0xFF));
    }
    else
    {
     *pw++=0;
     *pw++=0;
     *pw++=0;
    }


    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;
     prw+=inc1x;
     cw=0;

     shiftm+=inc1x;
     if(shiftm>=32)
     {
      pmw++;
      cmf=0;
      shiftm=0;
     }
     else
     if(shiftm<0)
     {
      pmw--;
      cmf=0;
      shiftm=31;
     }
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     prw=(int*)(((char*)prw)+incy);
     pmw=(int*)(((char*)pmw)+incmy);
     cw=0;
     cmf=0;
    }
			}
  }
 }
}








static int getspan(POINT * dpts,POINT * spts,int xpix,int ypix,int y,POINT * pts)
{
 int n;

 n=0;

 if(y==pts[0].y && pts[0].y==pts[1].y)
 {
  dpts[0].x=pts[0].x;
  dpts[1].x=pts[1].x;
  spts[0].x=0;
  spts[0].y=0;
  spts[1].x=xpix-1;
  spts[1].y=0;
  n=2;
 }
 else
 if(y==pts[1].y && pts[1].y==pts[2].y)
 {
  dpts[0].x=pts[1].x;
  dpts[1].x=pts[2].x;
  spts[0].x=xpix-1;
  spts[0].y=0;
  spts[1].x=xpix-1;
  spts[1].y=ypix-1;
  n=2;
 }
 else
 if(y==pts[2].y && pts[2].y==pts[3].y)
 {
  dpts[0].x=pts[2].x;
  dpts[1].x=pts[3].x;
  spts[0].x=xpix-1;
  spts[0].y=ypix-1;
  spts[1].x=0;
  spts[1].y=ypix-1;
  n=2;
 }
 else
 if(y==pts[3].y && pts[3].y==pts[0].y)
 {
  dpts[0].x=pts[3].x;
  dpts[1].x=pts[0].x;
  spts[0].x=0;
  spts[0].y=ypix-1;
  spts[1].x=0;
  spts[1].y=0;
  n=2;
 }
 else
 {
  if((y>=pts[0].y && y<pts[1].y) || (y<=pts[0].y && y>pts[1].y))
  {
   dpts[n].x=pts[0].x+scale(pts[1].x-pts[0].x,y-pts[0].y,pts[1].y-pts[0].y);
   spts[n].x=scale(xpix-1,y-pts[0].y,pts[1].y-pts[0].y);
   spts[n].y=0;
   n++;
  }

  if((y>=pts[1].y && y<pts[2].y) || (y<=pts[1].y && y>pts[2].y))
  {
   dpts[n].x=pts[1].x+scale(pts[2].x-pts[1].x,y-pts[1].y,pts[2].y-pts[1].y);
   spts[n].x=xpix-1;
   spts[n].y=scale(ypix-1,y-pts[1].y,pts[2].y-pts[1].y);
   n++;
  }

  if((y>=pts[2].y && y<pts[3].y) || (y<=pts[2].y && y>pts[3].y))
  {
   dpts[n].x=pts[3].x+scale(pts[2].x-pts[3].x,y-pts[3].y,pts[2].y-pts[3].y);
   spts[n].x=scale(xpix-1,y-pts[3].y,pts[2].y-pts[3].y);
   spts[n].y=ypix-1;
   n++;
  }

  if((y>=pts[3].y && y<pts[0].y) || (y<=pts[3].y && y>pts[0].y))
  {
   dpts[n].x=pts[0].x+scale(pts[3].x-pts[0].x,y-pts[0].y,pts[3].y-pts[0].y);
   spts[n].x=0;
   spts[n].y=scale(ypix-1,y-pts[0].y,pts[3].y-pts[0].y);
   n++;
  }
 }

 if(dpts[0].x>dpts[1].x)
 {
  SWOP(dpts[0].x,dpts[1].x);
  SWOP(spts[0].x,spts[1].x);
  SWOP(spts[0].y,spts[1].y);
//  dpts[1].x-=1;
 }

 return(n);
}


static void setmaskbitsscaled(POINT * pts,int sxpix,int sypix,int dbpp,int * dbits,int sx,int ex,int sy,int ey,processstr * prc)
{
 int   dx;
 int   dspanx;
 int   dy;
 POINT dpts[2];
 POINT spts[2];
 int   n;


// cprintf("setmaskbitsscaled");

 for(dy=sy;dy<ey;dy++)
 {
  n=getspan(dpts,spts,sxpix,sypix,dy,pts);
  if(n==2)
  {
   if(dpts[0].x>=ex || dpts[1].x<sx)
   {

   }
   else
   {
    dx=dpts[0].x;
    dspanx=dpts[1].x-dx;
    setmaskbitsscaledspan(dx,(ey-sy-1)-(dy-sy),dspanx,dbpp,dbits,sx,ex,prc);
   }
  }
 }
}




static void setspritebitsscaled4(POINT * pts,int * sbits,int sbpp,int sbytewidth,int sshift,int dbpp,int sxpix,int sypix,int * dbits,int sx,int sy,int ex,int ey,processstr * prc,newsprite_header * sph)
{
 int   dx;
 int   dspanx;
 int   dy;
 POINT dpts[2];
 POINT spts[2];
 int   n;

// cprintf("setspritesbitsscaled4");

// cprintf("x0 %d x1 %d x2 %d x3 %d",pts[0].x,pts[1].x,pts[2].x,pts[3].x);

 for(dy=sy;dy<ey;dy++)
 {
  n=getspan(dpts,spts,sxpix,sypix,dy,pts);
  if(n==2)
  {
   if(dpts[0].x>=ex || dpts[1].x<sx)
   {

   }
   else
   {
    dx=dpts[0].x;
    dspanx=dpts[1].x-dx;
    setspritebitsscaledspan(spts,sbits,sbpp,sbytewidth,sshift,dx,(ey-sy-1)-(dy-sy),dspanx,dbpp,dbits,sx,ex,prc,sph);
   }
  }
 }
}





static void setspritembitsscaled4(POINT * pts,int * sbits,int sbpp,int sbytewidth,int sshift,int * mbits,int dbpp,int sxpix,int sypix,int * dbits,int backindex,int sx,int sy,int ex,int ey,processstr * prc)
{
 int   dx;
 int   dspanx;
 int   dy;
 POINT dpts[2];
 POINT spts[2];
 int   n;

// cprintf("mask  XXXX setspritesmbitsscaled4");

// cprintf("x0 %d x1 %d x2 %d x3 %d",pts[0].x,pts[1].x,pts[2].x,pts[3].x);

 for(dy=sy;dy<ey;dy++)
 {
  n=getspan(dpts,spts,sxpix,sypix,dy,pts);
  if(n==2)
  {
   if(dpts[0].x>=ex || dpts[1].x<sx)
   {

   }
   else
   {
    dx=dpts[0].x;
    dspanx=dpts[1].x-dx;
    setspritembitsscaledspan(spts,sbits,sbpp,sbytewidth,sshift,mbits,dx,(ey-sy-1)-(dy-sy),dspanx,dbpp,dbits,backindex,sx,ex,prc);
   }
  }
 }
}




static void setspritembitsscaled14(POINT * pts,int * sbits,int sbpp,int sbytewidth,int sshift,int * mbits,int dbpp,int sxpix,int sypix,int * dbits,int backindex,int sx,int sy,int ex,int ey,processstr * prc,newsprite_header * sph)
{
 int   dx;
 int   dspanx;
 int   dy;
 POINT dpts[2];
 POINT spts[2];
 int   n;

// cprintf("mask setspritesmbitsscaled14 sbpp%d",sbpp);

// cprintf("x0 %d x1 %d x2 %d x3 %d",pts[0].x,pts[1].x,pts[2].x,pts[3].x);

 for(dy=sy;dy<ey;dy++)
 {
  n=getspan(dpts,spts,sxpix,sypix,dy,pts);
  if(n==2)
  {
   if(dpts[0].x>=ex || dpts[1].x<sx)
   {

   }
   else
   {
    dx=dpts[0].x;
    dspanx=dpts[1].x-dx;
    setspritembitsscaledspan1(spts,sbits,sbpp,sbytewidth,sxpix,sshift,mbits,dx,(ey-sy-1)-(dy-sy),dspanx,dbpp,dbits,backindex,sx,ex,prc,sph);
   }
  }
 }
}












os_error * sprite_renderbox(newsprite_header * sph,hdcstr * hdcptr,POINT * pts,int * pixtrans,processstr * prc)
{
 os_error   * err;
 int          i;
 int          xmin;
 int          xmax;
 int          ymin;
 int          ymax;
 int          xpix;
 int          ypix;
 int        * bits;
 int        * mbits;
 bmphdrstr    bmi;
 int          bitwidth;
 int          sbytewidth;
 int          mbytewidth;
 int          spbpp;
 int          mbpp;
 int        * dbits;
 HBITMAP      hbm;
// BITMAP       bitmap;
 HDC          hdcmem;
 int          dbpp;
 int          backindex;
 int          dxpix;
 int          dypix;
 int          sshift;
 int          x;
 int          y;
 graphicssavestr grsaves;
// void       * handle;
 RECT         clipbox;
 int          code;
 int          sx;
 int          ex;
 int          sy;
 int          ey;
 int          regular;



// cprintf("renderbox");

 err=NULL;

 code=GetClipBox(hdcptr->hdc,&clipbox);
 if(code!=NULLREGION)
 {
  regular=((pts[0].y==pts[1].y && pts[2].y==pts[3].y && pts[0].x==pts[3].x && pts[1].x==pts[2].x) ||
           (pts[0].x==pts[1].x && pts[2].x==pts[3].x && pts[0].y==pts[3].y && pts[1].y==pts[2].y)
          );


  xmin=xmax=pts[0].x;
  ymin=ymax=pts[0].y;

  for(i=1;i<4;i++)
  {
   xmin=MIN(xmin,pts[i].x);
   xmax=MAX(xmax,pts[i].x);
   ymin=MIN(ymin,pts[i].y);
   ymax=MAX(ymax,pts[i].y);
  }

  for(i=0;i<4;i++)
  {
   pts[i].x-=xmin;
   pts[i].y-=ymin;
  }

  x=xmin;
  y=ymin;

  dxpix=xmax-xmin+1;
  dypix=ymax-ymin+1;

  sy=clipbox.top-ymin;
  if(sy<0) sy=0;

  ey=clipbox.bottom-ymin;
  if(ey>dypix) ey=dypix;

  sx=clipbox.left-xmin;
  if(sx<0) sx=0;

  ex=clipbox.right-xmin;
  if(ex>dxpix) ex=dxpix;

// sx,sy are inclusive, i.e. fill pixels at sx,sy

  if(sx<ex && sy<ey)
  {
   getspriteinfo((newsprite_header*)sph,NULL,NULL,&spbpp);

   bits=(int*)(((char*)sph)+sph->image);
   mbits=(int*)(((char*)sph)+sph->mask);
   bitwidth=32*(sph->width+1)-sph->lbit-(31-sph->rbit);
   sbytewidth=4*(sph->width+1);
   sshift=sph->lbit;

   xpix=bitwidth/spbpp;
   ypix=sph->height+1;
  
   if(spbpp==2)  dbpp=4;
   else
   if(spbpp==32) dbpp=24;
   else          dbpp=spbpp;



   if(mbits==bits || (((newsprite_header*)sph)->selector.type.alpha))   // the no mask case
   {
    if(regular)
    {
    	graphicssave(hdcptr,&grsaves);
     graphicspixel(hdcptr);
//    cprintf("no mask case");
/*
			bbc_saveclippath(&handle);
   bbc_startpath();
   bbc_move(pts[0].x+xmin,pts[0].y+ymin);
   bbc_draw(pts[1].x+xmin,pts[1].y+ymin);
   bbc_draw(pts[2].x+xmin,pts[2].y+ymin);
   bbc_draw(pts[3].x+xmin,pts[3].y+ymin);
   bbc_draw(pts[0].x+xmin,pts[0].y+ymin);
   bbc_endpath();

   bbc_gclip(CLIP_AND);
*/

     memset(&bmi,0,sizeof(bmi));
   
     osgetpal(spbpp,(newsprite_header*)sph,&bmi.pal[0],NULL,pixtrans);

     bmi.bmi.biWidth=(ex-sx);
     bmi.bmi.biHeight=(ey-sy);
     bmi.bmi.biXPelsPerMeter=0;
     bmi.bmi.biYPelsPerMeter=0;
     bmi.bmi.biBitCount=(unsigned short)dbpp;
     bmi.bmi.biPlanes=1;
     bmi.bmi.biCompression=BI_RGB;
     bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
     bmi.bmi.biSizeImage=0;
     bmi.bmi.biClrImportant=0;
     if(spbpp==2) bmi.bmi.biClrUsed=(1<<spbpp);
     else         bmi.bmi.biClrUsed=0;

     err=XCreateDIBSection(hdcptr->hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
     if(!err)
     {
      setspritebitsscaled4(pts,bits,spbpp,sbytewidth,sshift,dbpp,xpix,ypix,dbits,sx,sy,ex,ey,prc,(newsprite_header*)sph);

      SetStretchBltMode(hdcptr->hdc,COLORONCOLOR);

//    GetObject(hbm,sizeof(BITMAP),&bitmap);

      hdcmem=CreateCompatibleDC(hdcptr->hdc);
      SelectObject(hdcmem,hbm);
      BitBlt(hdcptr->hdc,x+sx,y+sy,dxpix,dypix,hdcmem,0,0,SRCCOPY);
      DeleteDC(hdcmem);
  
      DeleteObject(hbm);
     }
//			bbc_loadclippath(handle);

	    graphicsload(hdcptr,&grsaves);
    }
    else
    {
    	graphicssave(hdcptr,&grsaves);
     graphicspixel(hdcptr);

     memset(&bmi,0,sizeof(bmi));

     osgetmpal(1,&bmi.pal[0]);

     bmi.bmi.biWidth=(ex-sx);
     bmi.bmi.biHeight=(ey-sy);
     bmi.bmi.biXPelsPerMeter=0;
     bmi.bmi.biYPelsPerMeter=0;
     bmi.bmi.biBitCount=(unsigned short)1;
     bmi.bmi.biPlanes=1;
     bmi.bmi.biCompression=BI_RGB;
     bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
     bmi.bmi.biSizeImage=0;
     bmi.bmi.biClrImportant=0;
     bmi.bmi.biClrUsed=0;

     err=XCreateDIBSection(hdcptr->hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
     if(!err)
     {
      setmaskbitsscaled(pts,xpix,ypix,1,dbits,sx,ex,sy,ey,prc);

//    GetObject(hbm,sizeof(BITMAP),&bitmap);
  
      hdcmem=CreateCompatibleDC(hdcptr->hdc);
      SelectObject(hdcmem,hbm);
      BitBlt(hdcptr->hdc,x+sx,y+sy,dxpix,dypix,hdcmem,0,0,SRCAND);  // and mask
      DeleteDC(hdcmem);
      DeleteObject(hbm);


      osgetpal(spbpp,(newsprite_header *)sph,&bmi.pal[0],NULL,pixtrans);
  
      bmi.bmi.biWidth=(ex-sx);
      bmi.bmi.biHeight=(ey-sy);
      bmi.bmi.biXPelsPerMeter=0;
      bmi.bmi.biYPelsPerMeter=0;
      bmi.bmi.biBitCount=(unsigned short)dbpp;
      bmi.bmi.biPlanes=1;
      bmi.bmi.biCompression=BI_RGB;
      bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
      bmi.bmi.biSizeImage=0;
      bmi.bmi.biClrImportant=0;
      if(spbpp==2) bmi.bmi.biClrUsed=(1<<spbpp);
      else         bmi.bmi.biClrUsed=0;

      err=XCreateDIBSection(hdcptr->hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
      if(!err)
      {
       setspritebitsscaled4(pts,bits,spbpp,sbytewidth,sshift,dbpp,xpix,ypix,dbits,sx,sy,ex,ey,prc,(newsprite_header*)sph);

       SetStretchBltMode(hdcptr->hdc,COLORONCOLOR);

//     GetObject(hbm,sizeof(BITMAP),&bitmap);

       hdcmem=CreateCompatibleDC(hdcptr->hdc);
       SelectObject(hdcmem,hbm);
       BitBlt(hdcptr->hdc,x+sx,y+sy,dxpix,dypix,hdcmem,0,0,SRCPAINT);
       DeleteDC(hdcmem);
  
       DeleteObject(hbm);
      }
     }

     graphicsload(hdcptr,&grsaves);
    }
   }
   else            // we have a mask
   {
  // there is a problem if the OS does not zero memory for the DIB sections
  // wonder if clipping speeds things up

   	graphicssave(hdcptr,&grsaves);
    graphicspixel(hdcptr);

/*
  bbc_saveclippath(&handle);  
  bbc_startpath();
  bbc_move(pts[0].x+xmin,pts[0].y+ymin);
  bbc_draw(pts[1].x+xmin,pts[1].y+ymin);
  bbc_draw(pts[2].x+xmin,pts[2].y+ymin);
  bbc_draw(pts[3].x+xmin,pts[3].y+ymin);
  bbc_draw(pts[0].x+xmin,pts[0].y+ymin);
  bbc_endpath();
  bbc_gclip(CLIP_AND);
*/ 
  
    if(((newsprite_header*)sph)->selector.type.type)
    {
     mbpp=dbpp=1;
     mbytewidth=((xpix+31)&(~31))/8;
    }
    else
    {
     mbpp=spbpp;
     if(spbpp==2) dbpp=4;
     else         dbpp=spbpp;
     mbytewidth=sbytewidth;
    }

    if(((newsprite_header*)sph)->selector.type.alpha)
    {
     mbpp=8;
     mbytewidth=((xpix+3)&(~3));
				}


    osgetmpal(mbpp,&bmi.pal[0]);

    bmi.bmi.biWidth=(ex-sx);
    bmi.bmi.biHeight=(ey-sy);
    bmi.bmi.biXPelsPerMeter=0;
    bmi.bmi.biYPelsPerMeter=0;
    bmi.bmi.biBitCount=(unsigned short)dbpp;
    bmi.bmi.biPlanes=1;
    bmi.bmi.biCompression=BI_RGB;
    bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
    bmi.bmi.biSizeImage=0;
    bmi.bmi.biClrImportant=0;
    if(mbpp==2) bmi.bmi.biClrUsed=(1<<mbpp);
    else        bmi.bmi.biClrUsed=0;

    err=XCreateDIBSection(hdcptr->hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
    if(!err)
    {
// we can assume bytewidth and start shift are the same for the mask because
// new format masks don't have left hand wastage
   
//   setspritebitsscaled(mbits,mbpp,mbytewidth,sshift,dbpp,xpix,ypix,dxpix,dypix,dbits);
     setspritebitsscaled4(pts,mbits,mbpp,mbytewidth,sshift,dbpp,xpix,ypix,dbits,sx,sy,ex,ey,prc,(newsprite_header*)sph);


//   GetObject(hbm,sizeof(BITMAP),&bitmap);

     hdcmem=CreateCompatibleDC(hdcptr->hdc);
     SelectObject(hdcmem,hbm);
     BitBlt(hdcptr->hdc,x+sx,y+sy,(ex-sx),(ey-sy),hdcmem,0,0,SRCAND);  // and mask
     DeleteDC(hdcmem);
     DeleteObject(hbm);

// Now start on main image

     if(spbpp==2)  dbpp=4;
     else
     if(spbpp==32) dbpp=24;
     else          dbpp=spbpp;
   
     osgetpal(spbpp,(newsprite_header *)sph,&bmi.pal[0],&backindex,pixtrans);

     bmi.bmi.biWidth=(ex-sx);
     bmi.bmi.biHeight=(ey-sy);
     bmi.bmi.biXPelsPerMeter=0;
     bmi.bmi.biYPelsPerMeter=0;
     bmi.bmi.biBitCount=(unsigned short)dbpp;
     bmi.bmi.biPlanes=1;
     bmi.bmi.biCompression=BI_RGB;
     bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
     bmi.bmi.biSizeImage=0;
     bmi.bmi.biClrImportant=0;
     if(spbpp==2) bmi.bmi.biClrUsed=(1<<spbpp);
     else         bmi.bmi.biClrUsed=0;

     err=XCreateDIBSection(hdcptr->hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
     if(!err)
     {
//    if(spbpp==mbpp) setspritembitsscaled(bits,spbpp,sbytewidth,sshift,mbits,dbpp,xpix,ypix,dxpix,dypix,dbits,backindex);
//    else            setspritembitsscaled1(bits,spbpp,sbytewidth,sshift,mbits,dbpp,xpix,ypix,dxpix,dypix,dbits,backindex);
//    cprintf("spbpp %d mbpp %d",spbpp,mbpp);

      if(spbpp==mbpp) setspritembitsscaled4(pts,bits,spbpp,sbytewidth,sshift,mbits,dbpp,xpix,ypix,dbits,backindex,sx,sy,ex,ey,prc);
      else            setspritembitsscaled14(pts,bits,spbpp,sbytewidth,sshift,mbits,dbpp,xpix,ypix,dbits,backindex,sx,sy,ex,ey,prc,(newsprite_header*)sph);
//    GetObject(hbm,sizeof(BITMAP),&bitmap);

      hdcmem=CreateCompatibleDC(hdcptr->hdc);
      SelectObject(hdcmem,hbm);
      BitBlt(hdcptr->hdc,x+sx,y+sy,ex-sx,ey-sy,hdcmem,0,0,SRCPAINT); // or data
      DeleteDC(hdcmem);
      DeleteObject(hbm);
     }
    }

//  bbc_loadclippath(handle);

  	 graphicsload(hdcptr,&grsaves);
   }
  }
 }
 return(err);
}


