/*->c.scrap */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "os.h"
#include "flex.h"

#include "err.h"

#include "fsx.h"
#include "bits.h"
#include "deb.h"

#include "scrap.h"






int programinstance(void)
{
 return(0);
}


#define SCRAPCHUNK (0x20*sizeof(int))


/* holds 1 if object is in use */

static unsigned int * scrapmap;
static int            scrapn;
static int            scrapinstance;
static char           scrapname[32];
static char           taskname[MAX_PATH];

/* scrap name is <Progname$Scrap>.Progname<instance>.<dir>.<dir>.file */
/* stored as 5 bits per filename and dir name */


char * sc_name(int sh)
{
 static char name[MAX_PATH];
 int    mask;
 int    shift;
 int    len;

 xsprintf(sizeof(name)/sizeof(TCHAR),NULL,name,"%s\\%s%d\\",taskname,scrapname,scrapinstance);
 len=strlen(name);

 mask=0x1F;
 shift=25;

 while(shift>=0)
 {
  if((sh>>shift)&mask) 
  {
   xsprintf(sizeof(name)/sizeof(TCHAR)-len,NULL,name+len,"%c%02d\\",shift?'D':'F',(sh>>shift)&mask);
   len+=4;
  }
  shift-=5;
 }

 if(name[len-1]=='\\') 
 {
  // strcpy(name+(len-1),".tmp");
  name[len-1]=0;
 }
 return(name);
}


char * sc_namex(int sh)
{
 char   buff[MAX_PATH];
 char * p;

 p=sc_name(sh);

 fs_expandpath(p,buff,sizeof(buff)/sizeof(TCHAR));

 strcpy(p,buff);

 return(p);
}



/* scan scrap map for unused handles */
/* else extend                       */

os_error * sc_create(int * shp)
{
 os_error * err;
 int        i;
 int        sh;

 err=NULL;

 sh=scrapn; /* compiler */

 for(i=0;i<scrapn;i++)
 {
  if((!(scrapmap[i>>5] & (1<<(i & 0x1F)))) && (i & 0x1F))
  {
   /* found an unused file handle */
   sh=i;
   break;
  }
 }

 if(i>=scrapn)
 {
  /* extend */

//  cprintf("extendmap=%x",*scrapmap);

  err=flex_chunk((flex_ptr)&scrapmap,((scrapn+2)/32+1)*sizeof(int),SCRAPCHUNK);
  if(!err)
  {
   sh=scrapn++;
   if(!(sh & 0x1F))
   {
    err=fs_cdir(sc_name(sh));
    if(!err)
    {
     scrapmap[sh>>5]|=(1<<(sh & 0x1F));
     sh=scrapn++;
    }
   }
  }
 }

 if(!err)
 {
  scrapmap[sh>>5]|=(1<<(sh & 0x1F));
  *shp=sh;
 }
 return(err);
}


/* kill handle, then scan end of map pruning empty directories */

os_error * sc_remove(int sh,os_error * err)
{

 if(scrapmap[sh>>5] & (1<<(sh & 0x1F))) scrapmap[sh>>5]^=(1<<(sh & 0x1F));

 if(sh==(scrapn-1))
 {
  while(scrapn)
  {
   sh=scrapn-1;

   if(!(sh & 0x1F))
   {
    /* remove directory */
    fs_delete(sc_name(sh));
    scrapn--;
   }
   else
   {
    if(scrapmap[sh>>5] & (1<<(sh & 0x1F))) break;
    else                                   scrapn--;
   }
  }
 }

 return(err);
}




static void startcase(void)
{
 fs_startscan();
}

static void endcase(void)
{
 fs_endscan();
}



static os_error * nextcase(fxstat * f,int * eof,int * n)
{
 os_error * err;
 char       wild[32];
 char     * p;
 char     * q;

 xsprintf(sizeof(wild)/sizeof(TCHAR),NULL,wild,"%s*",scrapname);

 while(1)
 {
  err=fs_nextitem(taskname,f,wild,eof);
  if(err)  return(err);
  if(*eof) return(err);

  q=p=f->name+strlen(scrapname);

  if(isdigit(*p++))
  {
   while(1)
   {
    if(!*p) 
    {
     *n=atoi(q);
     return(err);
    }
    if(!isdigit(*p++)) break;
   }
  }
 }

 return(err);
}



static int uniquecase(void)
{
 os_error * err;
 fxstat     f;
 int        eof;
 int        n;
 int        next;

 err=NULL;

 next=0;

 startcase();

 while(1)
 {
  err=nextcase(&f,&eof,&n);

  if(err) break;
  if(eof) break;

  if(n!=next) break;
  else        next++;
 }

 endcase();

 if(err) return(scrapinstance);
 else    return(next);
}



static int zapallcases(void)
{
 os_error * err;
 fxstat     f;
 int        eof;
 int        n;
 char       name[MAX_PATH];

 startcase();

 while(1)
 {
  err=nextcase(&f,&eof,&n);
  if(err || eof) break;

// used to use taskname%scrap
  xsprintf(sizeof(name)/sizeof(TCHAR),NULL,name,"%s\\%s",taskname,f.name);

  fs_wipe(name);
 }

 endcase();

 return(0);
}


/* scrpname is a short string to form into a scrap directory name */
/* tskname  is the name to appear in taskname$scrap               */
/* in general tskname may not be the same as wimpt_programname()  */


os_error * scrapinit(char * scrpname,char * tskname)
{
 os_error * err;
 int        len;

 scrapn=0;

 strcpy(scrapname,scrpname);

 if(tskname) strcpy(taskname,tskname);
 else
 {
  GetTempPath(sizeof(taskname)/sizeof(TCHAR),taskname);
  len=strlen(taskname);
  if(taskname[len-1]=='\\') taskname[len-1]=0;
 }

 err=flex_alloc((flex_ptr)&scrapmap,SCRAPCHUNK);
 scrapinstance=programinstance();

//cprintf("alloc map=%x",*scrapmap);

 /* fiddling - if scrapinstance is 0 then we are only case,
                  so can zap all scrap files.
               if scrapinstance is >0 then we better find a scrap
                  directory that does not exist */

 if(!scrapinstance) scrapinstance=zapallcases();
 else               scrapinstance=uniquecase();

 return(err);
}


/* attempt to trash all scrap files in sight */


os_error * scrapfinit(void)
{
 os_error * err;

 err=fs_wipe(sc_name(0));

 return(err);
}


